#ifndef COMMON_ARRAYS_H
#define COMMON_ARRAYS_H

#include "common_base.h"

//This struct holds information about a private array, it holds it's
//position in main memory, it's type, where it was created, a name
//to allow userspace fetching of the variable, and a size
typedef struct _privateArray{
    Name name; // scope
    int type; // IVAL, FVAL, DVAL
    int size; // size of array
    Addr64 base; // Address in main memory of the data
    //    char* uspaceName; // userspace name
    //    int no; // used to find the variable when more than one exists (i.e. arrays)
    Dma_tag* tag; // mfc-tag so we can wait for the variable to finish
    struct _privateArray* next; //for handling of the linked list in
				//the dataManagers
    struct _privateArray* prev; //for handling of the linked list in
				//the dataManagers
} PrivateArray;

//This struct holds information about a cyclic array, it holds it's
//position in main memory, it's type, where it was created, a name
//to allow userspace fetching of the variable, a blocksize to determine
//how to traverse the array and a total size
typedef struct _cyclicDistArray{
    Name name; //scope
    int type; //IVAL, FVAL, DVAL
    int size; // size of array
    Addr64 base; // Address in main memory of the data
    int blocksize; // size of block, data copied by by :
                   // for(i in (size/blocksize)%spu_#) memcpy(i*blocksize,...) 

    Dma_tag* tag; // mfc-tag so we can wait for the variable to finish
    struct _cyclicDistArray* next; //for handling of the linked list
				   //in the dataManagers
    struct _cyclicDistArray* prev; //for handling of the linked list
				   //in the dataManagers

} CyclicDistArray;

//This struct holds information about a block array, it holds it's
//position in main memory, it's type, where it was created, a name
//to allow userspace fetching of the variable, a localsize to determine
//how much of the array is in local memory, a lower and higher bound
//of the segment and a total size
typedef struct _blockDistArray{
    Name name; // scope
    int type; // IVAL, FVAL, DVAL
    int size; // size of array
    Addr64 base; // Address in main memory of the data
    int localsize; // size of segment available to spu
    int lower; // lower bound of segment
    int higher; // upper bound of segment
    Dma_tag* tag; // mfc-tag so we can wait for the variable to finish
    struct _blockDistArray* next; //for handling of the linked list in
				  //the dataManagers
    struct _blockDistArray* prev; //for handling of the linked list in
				  //the dataManagers
    int pad1;
    int pad2; // NOTE: necessary to get a multiple of 16byte,
              // can also be achieved by pragmas
} BlockDistArray;

//This struct holds information about a replicated array, it holds it's
//position in main memory, it's type, where it was created, a name
//to allow userspace fetching of the variable, a lower and higher bound
//of the segment that should be updated in the next combine and a total size
typedef struct _repArray{
    Name name; //scope
    int type; //IVAR, FVAR, DVAR
    int size; // size of array
    Addr64 base; // Address in main memory of the data
    int url; //update range lower
    int urh; //update range higher
    Dma_tag* tag; // mfc-tag so we can wait for the variable to finish
    struct _repArray* next; //for handling of the linked list in the dataManagers
    struct _repArray* prev; //for handling of the linked list in the dataManagers
    int pad1;
    int pad2; // NOTE: necessary to get a multiple of 16byte, 
              // can also be achieved by pragmas
} RepArray;

//NOTE: this is deprecated..
typedef struct _arrayHolder {
    BlockDistArray **block_arrays;
    int no_block_arrays;
    CyclicDistArray **cyclic_arrays;
    int no_cyclic_arrays;
} ArrayHolder;

//creators and destructors are in the ppulib with callers in spulib
//getters and setters are in the spulib

#endif //COMMON_ARRAYS_H
