#ifndef COMMON_BASE_H
#define COMMON_BASE_H

//variables pertaining to the groupsize
//TODO: should these be here or in common_ppu.h?
//#define MAX_SPE_THREADS 6 // to be able to run on ps3
#define MAX_SPE_THREADS 8 // when running in simulator and on cellblades
int GROUP_SIZE;

// possible variable types for shared variables and arrays
#define IVAR 1
#define FVAR 2
#define DVAR 3

//the different types of variables that can be in a datamanager
#define SHVAR 1
#define PVAR 2
#define PARR 3
#define CARR 4
#define BARR 5
#define RARR 6

//the different signals that can come to a ppu from a spu through a mailbox
#define NO_DATA -1
#define CREATE_VARIABLE 1
#define REMOVE_VARIABLE 2
#define DESTROY_THREAD 3
#define COMBINE 4
#define ATTACH_VARIABLE 5
#define MIRROR_VARIABLE 6
#define UPDATE_VARIABLE 7
#define REPARRAY_RANGE 8
#define NESTSTEP_SPU_FATAL_ERROR 9
#define GET_ALL_LS_ADDR 10
//if NULL isn't defined, define it.
#ifndef NULL
#define NULL 0
#endif
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

//defines for returning errors
#define OUT_OF_BOUNDS -1
//TODO: code removal
//#define OUT_OF_BOUNDS_H -2
//#define OUT_OF_BOUNDS_L -3


// NestStep defines
#define DEFAULT_BLOCKSIZE 128 // default blocksize for cyclic arrays in bytes
                              // the value is chosen in consideration to
                              // how the EIB on the cell works.

/* possible combine strategies for combineitems */
#define ADD 1
#define MULT 2
#define AND 4
#define OR 8
#define FUNC 16
#define LEADER 32
#define UPDATED 64
#define COMBINE_MAX 128
#define COMBINE_MIN 256

/* prefix tags */
#define NO_PREFIX 1
#define PREFIX 2
#define UPDATE 4

/* updated tags */
#define NOT_UPDATED 1
//#define UPDATED 2

// determines the maximum amount of combineitems in a message
//NOTE: the following line has changed behaviour from NestStep-RTS,
//      it now uses linked lists instead of static arrays.
//#define MAX_COMBINE_ITEMS 64

//max nesting depth
#define MAX_DEPTH 10

//an internal representation of the name, includes scope of variable.
typedef struct name {
	int procedure; // a number that is unique for this scope
	int relative; // a number that is unique for the variable in the given scope
	int path[MAX_DEPTH]; // a path to the variable
} Name; // 12 * 4 = 48byte = 3*16 byte, ok, can be put into an array

//CELL specific defines and structs/unions
//This union is used to represent a memory-address in the main-memory
typedef union _addr64
{
    unsigned long long ull;
    unsigned int ui[2];
    void *p;
    int* ia;
    float* fp;
    double* dp;
} Addr64; //8 byte, should not be put into an array

typedef struct _dma_tag{
    int tag[32];
    Addr64 addr[32];
} Dma_tag;

//TODO: remove this.
//this is the tree structure for handling combine-trees
typedef struct _tree{
    int parent;
    int no_children;
    int *children;
} Tree; // 12 byte, should not be put into an array

// this structure is used to communicate pointers to all spu's ls and ps for 
// spu to spu comunication
typedef struct _ls_pointers{
	volatile unsigned int ls;
	volatile unsigned int sig1;
	volatile unsigned int sig2;
	volatile unsigned int pad;
} Ls_pointers;

#endif //COMMON_BASE_H
