#ifndef COMMON_COMBINE_H
#define COMMON_COMBINE_H

#include "common_base.h"

//This union is used to hold a prefix-value.
typedef union _value{
    int i; /* IVAR */
    float f; /* FVAR */
    double d; /* DVAR */
} Value;

// A CombineItem is for returning prefix variables
// TODO: remove this?
typedef struct _combineItem{ 
    Name name; // this is the name of the shared var
    Value val; // value of the shared val
    int contributor; // process who the combineitem originates from
    int prefix_tag; //NO_PREFIX, PREFIX or UPDATE
    struct _combineItem* prev; //for the linked list
    struct _combineItem* next; //for the linked list
} CombineItem;

// CombineInfo is a structure that tells how to combine a ShVar
typedef struct _combineInfo {
    ShVar* var; // a pointer to the ShVar that should be combined
    int type; // CHVAR, IVAR, FVAR,DVAR
    int combine_strategy; // ADD, MULT, AND, OR, FUNC, LEADER, UPDATED
    void (*combineFn)(void*,void*,void*); // combine function
    int updated; // NOT_UPDATED, UPDATED
    int prefix_tag;
    void* prefix_base;
    struct _combineInfo* prev; //for the linked list
    struct _combineInfo* next; //for the linked list
} CombineInfo;

//TODO: what does this do?
//      remove this?
typedef struct _prefixVar {
    void* prefix;
    ShVar* shared;
    Value from_parent;
    Value my_value;
    Value* child_values;
    struct _prefixVar* prev; //for the linked list
    struct _prefixVar* next; //for the linked list
} PrefixVar;

// RepArrayInfo is a structure that tells how to combine a Replicated Array
typedef struct _repArrayInfo {
    RepArray *rep_array; // a pointer to the ReplicatedArray data
    int type; // IVAR, FVAR or DVAR
    int combine_strategy; // what combine-strategy to use
    void (*combineFn)(void*,void*,void*); // a custom function might be
                                          // used to combine values
    int prefix_tag; // how should the prefix be handled?
    //    void *old_base;
    void *prefix_base;
    void **child_arrays;
    int updated;
    struct _repArrayInfo* prev; //for the linked list
    struct _repArrayInfo* next; //for the linked list
} RepArrayInfo;

// BlockArrayInfo is a structure that holds information about a remote
// read/write (mirror/update)
typedef struct _blockArrayInfo {
    BlockDistArray* block_array; // not needed
    Name name;
    Addr64 buffer; // address in main memory of the buffer
    int lower; 
    int higher;
    int request; // MIRROR_VARIABLE / UPDATE_VARIABLE
    int request_id; // can be removed
    int no_reads; // can be removed
    struct _blockArrayInfo* prev; //for the linked list
    struct _blockArrayInfo* next; //for the linked list
} BlockArrayInfo;

// CyclicArrayInfo is a structure that holds information about a remote
// read/write (mirror/update)
typedef struct _cyclicArrayInfo {
    CyclicDistArray *cyclic_array; // not needed
    Name name;
    Addr64 buffer;
    int lower;
    int higher;
    int request; // MIRROR_VARIABLE / UPDATE_VARIABLE
    int request_id; // can be removed
    int no_reads; // can be removed
    struct _cyclicArrayInfo* prev; //for the linked list
    struct _cyclicArrayInfo* next; //for the linked list
} CyclicArrayInfo;

//A CombineList is an item that is used for each group to know how to
//combine it's variables at the end of a step 
//NOTE: BAI and CAI is here because of mirror/update, the number of
//items are combined so each node can know how many requests it has to
//serve, don't know if it's necessary for me..
typedef struct _combineList{
    CombineItem *combine_items; // prefix vars
    CombineInfo *combine_info; // ShVars
    int no_combine_items; // can be removed

    PrefixVar *prefix_vars;
    int no_prefix_vars; // can be removed .. can PrefixVars be removed entirely?

    RepArrayInfo *rep_info;
    int no_rep_arrays; // can be removed
    //    int largest_rep_array;

    BlockArrayInfo *block_info;
    int no_block_arrays; // can be removed

    CyclicArrayInfo *cyclic_info;
    int no_cyclic_arrays; // can be removed

    int request_id_counter;
    int *NDAmsgs;
    
} CombineList;


#endif // COMMON_COMBINE_H

