#include "common_neststep.h"

//copy a name between two initialize containers
void namecpy(Name* _to, Name* _from){
    int i;
    _to->procedure = _from->procedure;
    _to->relative = _from->relative;
    for(i=0; i<MAX_DEPTH;i++){
	_to->path[i] = _from->path[i];
    }
    
    return;
}

//returns the size of an element (IVAR, FVAR, DVAR)
int get_element_size(int _type){
    switch(_type){
    case IVAR:
	return sizeof(int);
    case FVAR:
	return sizeof(float);
    case DVAR:
	return sizeof(double);
    default:
	return -1;
    }
}

//compare two names, return true if they match
//TODO: i think it is sufficient to check procedure and relative here
int compareName(Name* n1, Name* n2){
    if(n1->procedure != n2->procedure){
	return 0;
    }
    if(n1->relative != n2->relative){
	return 0;
    }
    return 1;
}

//add to datamanager
void add_PVar(PVar* var, DataManager* _dm){    
    PVar* tmp_pvar;
    if(_dm->no_pVars == 0){
	_dm->privateVars = var;
	_dm->no_pVars++;
	return;
    }
    tmp_pvar = _dm->privateVars;
    while(tmp_pvar->next != NULL){
	tmp_pvar = tmp_pvar->next;
    }
    tmp_pvar->next = var;
    var->prev = tmp_pvar;
    _dm->no_pVars++;

    return;
}
// remove variable from datamanager
void remove_PVar(PVar* var, DataManager* _dm){
    PVar* tmp_pvar;
    if(_dm->no_pVars == 0){
	return; //no found vars
    }
    tmp_pvar = _dm->privateVars;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_pvar->name), &(var->name))){
	tmp_pvar = tmp_pvar->next;
	if(tmp_pvar == NULL){
	    return; // no allegible vars
	}
    }
    if(tmp_pvar->prev == NULL){
	_dm->privateVars = tmp_pvar->next;
	if(_dm->privateVars)
	    _dm->privateVars->prev = NULL;
	_dm->no_pVars--;
	return;
    }
    // tmp_pvar is now the correct var
    tmp_pvar = tmp_pvar->prev; //go back one step
    tmp_pvar->next = tmp_pvar->next->next; //remove variable
    if(tmp_pvar->next != NULL){
	tmp_pvar->next->prev = tmp_pvar; //remove from backwards-list if existing
    }
    _dm->no_pVars--;
    return;
}
PVar* locate_PVar(Name _name, DataManager* _dm){
    //find the variable inside the dataManager
    PVar* tmp_pvar;
    if(_dm->no_pVars == 0){
	return NULL; //no found vars
    }
    tmp_pvar = _dm->privateVars;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_pvar->name), &(_name))){
	tmp_pvar = tmp_pvar->next;
	if(tmp_pvar == NULL){
	    return NULL; // no allegible vars
	}
    }
    // tmp_pvar is now the correct var
    return tmp_pvar;
    //    return NULL;
}


void add_ShVar(ShVar* var, DataManager* _dm){    
    ShVar* tmp_shvar;
    if(_dm->no_shVars == 0){
	_dm->shVars = var;
	_dm->no_shVars++;
	return;
    }
    tmp_shvar = _dm->shVars;
    while(tmp_shvar->next != NULL){
	tmp_shvar = tmp_shvar->next;
    }
    tmp_shvar->next = var;
    var->prev = tmp_shvar;
    _dm->no_shVars++;

    return;
}
// remove variable from datamanager
void remove_ShVar(ShVar* var, DataManager* _dm){
    ShVar* tmp_shvar;
    if(_dm->no_shVars == 0){
	return; //no found vars
    }
    tmp_shvar = _dm->shVars;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_shvar->name), &(var->name))){
	tmp_shvar = tmp_shvar->next;
	if(tmp_shvar == NULL){
	    return; // no allegible vars
	}
    }
    if(tmp_shvar->prev == NULL){
	_dm->shVars = tmp_shvar->next;
	if(_dm->shVars)
	    _dm->shVars->prev = NULL;
	_dm->no_shVars--;
	return;
    }
    // tmp_pvar is now the correct var
    tmp_shvar = tmp_shvar->prev; //go back one step
    tmp_shvar->next = tmp_shvar->next->next; //remove variable
    if(tmp_shvar->next != NULL){
	tmp_shvar->next->prev = tmp_shvar; //remove from backwards-list if existing
    }
    _dm->no_shVars--;
    return;
}

ShVar* locate_ShVar(Name _name, DataManager* _dm){
    ShVar* tmp_shvar;
    if(_dm->no_shVars == 0){
	return NULL; //no found vars
    }
    tmp_shvar = _dm->shVars;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_shvar->name), &(_name))){
	tmp_shvar = tmp_shvar->next;
	if(tmp_shvar == NULL){
	    return NULL; // no allegible vars
	}
    }
    // tmp_shvar is now the correct var
    return tmp_shvar;
    //    return NULL;
}

void add_PrivateArray(PrivateArray* var, DataManager* _dm){    
    PrivateArray* tmp_parr;
    if(_dm->no_pArrays == 0){
	_dm->privateArrays = var;
	_dm->no_pArrays++;
	return;
    }
    tmp_parr = _dm->privateArrays;
    while(tmp_parr->next != NULL){
	tmp_parr = tmp_parr->next;
    }
    tmp_parr->next = var;
    var->prev = tmp_parr;
    _dm->no_pArrays++;

    return;
}
// remove variable from datamanager
void remove_PrivateArray(PrivateArray* var, DataManager* _dm){
    //    printf("a\n");
    PrivateArray* tmp_parr;
    if(_dm->no_pArrays == 0){
	return; //no found vars
    }
    tmp_parr = _dm->privateArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_parr->name), &(var->name))){
	tmp_parr = tmp_parr->next;
	if(tmp_parr == NULL){
	    return; // no allegible vars
	}
    }
    //    printf("b\n");
    if(tmp_parr->prev == NULL){
	_dm->privateArrays = tmp_parr->next;
	if(_dm->privateArrays)
	    _dm->privateArrays->prev = NULL;
	_dm->no_pArrays--;
	return;
    }
    // tmp_parr is now the correct var
    //    printf("prev: %x, next: %x\n",tmp_parr->prev, tmp_parr->next);

    tmp_parr = tmp_parr->prev; //go back one step
    //    printf("ba\n");
    tmp_parr->next = tmp_parr->next->next; //remove variable
    //    printf("bb\n");
    if(tmp_parr->next != NULL){
	tmp_parr->next->prev = tmp_parr; //remove from backwards-list if existing
	//    printf("bc\n");
    }
    _dm->no_pArrays--;
    //    printf("c\n");
    return;
}

PrivateArray* locate_PrivateArray(Name _name, DataManager* _dm){
    PrivateArray* tmp_parr;
    if(_dm->no_pArrays == 0){
	return NULL; //no found vars
    }
    tmp_parr = _dm->privateArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_parr->name), &(_name))){
	tmp_parr = tmp_parr->next;
	if(tmp_parr == NULL){
	    return NULL; // no allegible vars
	}
    }
    // tmp_pvar is now the correct var
    return tmp_parr;
    //    return NULL;
}

void add_RepArray(RepArray* var, DataManager* _dm){    
    RepArray* tmp_rarr;
    if(_dm->no_repArrays == 0){
	_dm->repArrays = var;
	_dm->no_repArrays++;
	return;
    }
    tmp_rarr = _dm->repArrays;
    while(tmp_rarr->next != NULL){
	tmp_rarr = tmp_rarr->next;
    }
    tmp_rarr->next = var;
    var->prev = tmp_rarr;
    _dm->no_repArrays++;

    return;
}
// remove variable from datamanager
void remove_RepArray(RepArray* var, DataManager* _dm){
    RepArray* tmp_rarr;
    if(_dm->no_repArrays == 0){
	return; //no found vars
    }
    tmp_rarr = _dm->repArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_rarr->name), &(var->name))){
	tmp_rarr = tmp_rarr->next;
	if(tmp_rarr == NULL){
	    return; // no allegible vars
	}
    }
    if(tmp_rarr->prev == NULL){
	_dm->repArrays = tmp_rarr->next;
	if(_dm->repArrays)
	    _dm->repArrays->prev = NULL;
	_dm->no_repArrays--;
	return;
    }
    // tmp_pvar is now the correct var
    tmp_rarr = tmp_rarr->prev; //go back one step
    tmp_rarr->next = tmp_rarr->next->next; //remove variable
    if(tmp_rarr->next != NULL){
	tmp_rarr->next->prev = tmp_rarr; //remove from backwards-list if existing
    }
    _dm->no_repArrays--;
    return;
}

RepArray* locate_RepArray(Name _name, DataManager* _dm){
    RepArray* tmp_rarr;
    if(_dm->no_repArrays == 0){
	return NULL; //no found vars
    }
    tmp_rarr = _dm->repArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_rarr->name), &(_name))){
	tmp_rarr = tmp_rarr->next;
	if(tmp_rarr == NULL){
	    return NULL; // no allegible vars
	}
    }
    // tmp_pvar is now the correct var
    return tmp_rarr;
    //    return NULL;
}

void add_BlockDistArray(BlockDistArray* var, DataManager* _dm){    
    BlockDistArray* tmp_barr;
    if(_dm->no_bArrays == 0){
	_dm->blockDistArrays = var;
	_dm->no_bArrays++;
	return;
    }
    tmp_barr = _dm->blockDistArrays;
    while(tmp_barr->next != NULL){
	tmp_barr = tmp_barr->next;
    }
    tmp_barr->next = var;
    var->prev = tmp_barr;
    _dm->no_bArrays++;

    return;
}
// remove variable from datamanager
void remove_BlockDistArray(BlockDistArray* var, DataManager* _dm){
    BlockDistArray* tmp_barr;
    if(_dm->no_bArrays == 0){
	return; //no found vars
    }
    tmp_barr = _dm->blockDistArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_barr->name), &(var->name))){
	tmp_barr = tmp_barr->next;
	if(tmp_barr == NULL){
	    return; // no allegible vars
	}
    }
    // tmp_pvar is now the correct var
    //    printf("prev: %x, next: %x\n",tmp_barr->prev, tmp_barr->next);
    if(tmp_barr->prev == NULL){
	_dm->blockDistArrays = tmp_barr->next;
	if(_dm->blockDistArrays)
	    _dm->blockDistArrays->prev = NULL;
	_dm->no_bArrays--;
	return;
    }
    tmp_barr = tmp_barr->prev; //go back one step
    tmp_barr->next = tmp_barr->next->next; //remove variable
    if(tmp_barr->next != NULL){
	tmp_barr->next->prev = tmp_barr; //remove from backwards-list if existing
    }
    _dm->no_bArrays--;
    return;
}

BlockDistArray* locate_BlockDistArray(Name _name, DataManager* _dm){
    BlockDistArray* tmp_barr;
    if(_dm->no_bArrays == 0){
	return NULL; //no found vars
    }
    tmp_barr = _dm->blockDistArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_barr->name), &(_name))){
	tmp_barr = tmp_barr->next;
	if(tmp_barr == NULL){
	    return NULL; // no allegible vars
	}
    }
    // tmp_pvar is now the correct var
    return tmp_barr;
    //    return NULL;
}

void add_CyclicDistArray(CyclicDistArray* var, DataManager* _dm){    
    CyclicDistArray* tmp_carr;
    if(_dm->no_cArrays == 0){
	_dm->cyclicDistArrays = var;
	_dm->no_cArrays++;
	return;
    }
    tmp_carr = _dm->cyclicDistArrays;
    while(tmp_carr->next != NULL){
	tmp_carr = tmp_carr->next;
    }
    tmp_carr->next = var;
    var->prev = tmp_carr;
    _dm->no_cArrays++;

    return;
}
// remove variable from datamanager
void remove_CyclicDistArray(CyclicDistArray* var, DataManager* _dm){
    CyclicDistArray* tmp_carr;
    if(_dm->no_cArrays == 0){
	return; //no found vars
    }
    tmp_carr = _dm->cyclicDistArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_carr->name), &(var->name))){
	tmp_carr = tmp_carr->next;
	if(tmp_carr == NULL){
	    return; // no allegible vars
	}
    }
    if(tmp_carr->prev == NULL){
	_dm->cyclicDistArrays = tmp_carr->next;
	if(_dm->cyclicDistArrays)
	    _dm->cyclicDistArrays->prev = NULL;
	_dm->no_cArrays--;
	return;
    }
    // tmp_pvar is now the correct var
    tmp_carr = tmp_carr->prev; //go back one step
    tmp_carr->next = tmp_carr->next->next; //remove variable
    if(tmp_carr->next != NULL){
	tmp_carr->next->prev = tmp_carr; //remove from backwards-list if existing
    }
    _dm->no_cArrays--;
    return;
}

CyclicDistArray* locate_CyclicDistArray(Name _name, DataManager* _dm){
    CyclicDistArray* tmp_carr;
    if(_dm->no_cArrays == 0){
	return NULL; //no found vars
    }
    tmp_carr = _dm->cyclicDistArrays;

    //loop as long as we don't find a matching name
    while(!compareName(&(tmp_carr->name), &(_name))){
	tmp_carr = tmp_carr->next;
	if(tmp_carr == NULL){
	    return NULL; // no allegible vars
	}
    }
    // tmp_pvar is now the correct var
    return tmp_carr;
    //    return NULL;
}

