#ifndef COMMON_NESTSTEP_H
#define COMMON_NESTSTEP_H

#include "common_arrays.h"
#include "common_var.h"
#include "common_combine.h"

//an info-structure for managing the internal structure of the system
typedef struct _groupInfo{
    char* name;
    int size;
    int my_rank;
    int my_rank_local;
    int first_rank;
    int *my_group;
    int my_group_number;
    int no_of_groups;
    int **all_groups;
    int *group_leaders;
    struct _groupInfo *parent;
} GroupInfo; // 11*4 = 44 byte, non power of two, mustn't be put in array

typedef struct _group{
    GroupInfo groupInfo;
    struct _group *parent;
    Tree *tree;
    int path[MAX_DEPTH];
    CombineList *clist;
} Group; // 24 * 4 = 96 = 6*16, ok, can be put in array

//This is used to send data between the spu and ppu, namely when
//the spu wants to create or remove a variable
typedef struct _mboxExtraData{
    unsigned int type; // ShVar, PVar, RepArray
    unsigned int datatype; // INT, DOUBLE
    unsigned int size; // number of elements
    unsigned int blocksize; // size of each block of elements
    Addr64 addr; //address in main memory, sent back to the spu
    Name name; // 48 byte
    Name prefixname; //48 byte
    unsigned int lower;
    unsigned int higher; 
} MboxExtraData; // 128 byte, optimal.

//construct for managing the data used by the SPU:s, allowing them
//to use larger data-segments than 256K and to have persistant data
//between segments of SPU code stored in multiple files
typedef struct dataManager{
    //control handling
    Group* thisgroup; // the root group, holds all data for managing
                     // the connection between nodes.
    //extra-data sent from the spu
    MboxExtraData mboxExtraData __attribute__ ((aligned(16)));
    Addr64 mboxAddr;
    //data handling
    ShVar* shVars;
    int no_shVars;
    PVar* privateVars;
    int no_pVars;
    PrivateArray* privateArrays;
    int no_pArrays;
    CyclicDistArray* cyclicDistArrays;
    int no_cArrays;
    BlockDistArray* blockDistArrays;
    int no_bArrays;
    RepArray* repArrays;
    int no_repArrays;
    int tags[32];
} DataManager; //144+128byte struct = 17*16, OK! Can be put in array, must
	       //be transfered in parts though since it's larger than
	       //128byte

//common functions

//function for copying a name
void namecpy(Name* _to, Name* _from);

//function for comparing two names
int compareName(Name* n1, Name* n2);

//returns the size of a specific type
int get_element_size(int _type);

//adds the variable to the end of the datamanagers linked list
void add_PVar(PVar* var, DataManager* _dm); 
void remove_PVar(PVar* var, DataManager* _dm);
// locates a specific variable in the datamanager
PVar* locate_PVar(Name _name, DataManager* _dm); 


//adds the variable to the end of the datamanagers linked list
void add_ShVar(ShVar* var, DataManager* _dm); 
void remove_ShVar(ShVar* var, DataManager* _dm);
// locates a specific variable in the datamanager
ShVar* locate_ShVar(Name _name, DataManager* _dm);

//adds the variable to the end of the datamanagers linked list
void add_PrivateArray(PrivateArray* var, DataManager* _dm); 
void remove_PrivateArray(PrivateArray* var, DataManager* _dm);
// locates a specific variable in the datamanager
PrivateArray* locate_PrivateArray(Name _name, DataManager* _dm); 

//adds the variable to the end of the datamanagers linked list
void add_RepArray(RepArray* var, DataManager* _dm); 
void remove_RepArray(RepArray* var, DataManager* _dm);
// locates a specific variable in the datamanager
RepArray* locate_RepArray(Name _name, DataManager* _dm); 

//adds the variable to the end of the datamanagers linked list
void add_BlockDistArray(BlockDistArray* var, DataManager* _dm); 
void remove_BlockDistArray(BlockDistArray* var, DataManager* _dm);
// locates a specific variable in the datamanager
BlockDistArray* locate_BlockDistArray(Name _name, DataManager* _dm); 

//adds the variable to the end of the datamanagers linked list
void add_CyclicDistArray(CyclicDistArray* var, DataManager* _dm);
void remove_CyclicDistArray(CyclicDistArray* var, DataManager* _dm);
// locates a specific variable in the datamanager
CyclicDistArray* locate_CyclicDistArray(Name _name, DataManager* _dm); 
#endif // COMMON_NESTSTEP_H
