#ifndef COMBINE_PPU_H
#define COMBINE_PPU_H

#include "../NestStep_common/common_neststep.h"
#include "base_ppu.h"

void NestStep_combine(DataManager* dataManagers);
void clear_CombineList(DataManager* dataManagers);

//functions for managing combinelists
CombineList* new_CombineList();
void free_CombineList(CombineList* _clist);

//attach a combinefunction and bind a name to it so that it can be referenced from the spu
//when running combine
//NOTE: We don't allow combine-functions on this architecture
//void NestStep_attach_combineFn(char* _functionName, 
//                               void* _combineFn);

//This functions attaches a shared variable to the combine-list that is used
//to get a consistent state after a step-statement
void NestStep_ShVar_attach(ShVar *_var, 
                           int _combine_strategy,
                           void (*_combineFn)(void*, void*, void*),
                           int _prefix_tag,
                           Addr64 _prefix_base,
                           int _updated,
                           CombineList* _clist,
			   int spu_id);
//This function attaches a replicated array to the combine-list that is used
//to get a consistent state after a step-statement, inside the RepArray a
//lower and higher bound should be set to indicate what parts of the array
//should be combined
void NestStep_RepArray_attach(RepArray *_array,
                              int _combine_strategy,
                              void (*_combineFn)(void*, void*, void*),
                              int _prefix_tag,
                              Addr64 _prefix_base,
                              int updated,
                              CombineList *_clist,
			      int spu_id);

// sets the range for which the combine-function should affect the
// array, will stay in effect until after the next combine
int NestStep_RepArray_range(RepArray *_array, 
			    int _lower, 
			    int _higher,
			    int spu_id);

//these functions transform a local index to a global index and
//from global to local respectively.
int global_local_index_b(BlockDistArray *array, int index, int spu_no);
int local_global_index_b(BlockDistArray *array, int index, int spu_no);
int global_local_index_c(CyclicDistArray *array, int index, int spu_no);
int local_global_index_c(CyclicDistArray *array, int index, int spu_no);


#endif // COMBINE_PPU_H
