/* ************************************************************** *
 * NestStep library for the IBM CELL. This is the PPU-part,       *
 * there also exist an SPU-part. The library is structured        *
 * to run controls such as managing memory and what SPU should    *
 * do what on the PPU and all calculations should reside in the   *
 * step-statements, which should run on the SPU, please read the  *
 * intstructions on http://www.ida.liu.se/chrke/NestStep/ for     *
 * more help on how to write NestStep programs.                   *
 * ************************************************************** */

#include "neststep_ppu.h"

//TODO: this file needs to be cleaned up
//NOTE: remember, in 32bit mode variables on the ppu are 8byte aligned
//      thus, if we allocate memory we must make sure that it is 16byte
//      aligned, if we're in 64bit mode we get 16byte alignment.
//      on the spu we are 16byte aligned by default, this has to do with
//      registers on the ppu beeing 64bit and registers on the spu beeing
//      128bit, the vector-registers on the ppu requires 16byte alignment
//      but this hasn't been accounted for in 32bit version of the ppu-libc
//      thus you have to use memalign or malloc_align on the ppu in 32bit
//      mode to get the 16byte alignment. Also, if you run in 64bit mode you
//      will get 8byte pointers on the ppu while the spu will still run 
//      in 32bit mode and thus have a 4byte pointer, this will make the common
//      structures have different sizes on the spu and the ppu.

//this function is sent to the pthread responsible for handling the spe, it is
//put to sleep as soon as the spe starts running.
/* void pthread_wrapper_for_spe(thread_args* arg){ */
/*     unsigned int runflags = 0; */
/*     unsigned int entry = SPE_DEFAULT_ENTRY; */
/*     // run SPE context */
/*     spe_context_run(arg->spe, 
       &entry, 
       runflags, 
       (unsigned long long*)arg->argp, 
       (unsigned long long*)arg->envp, NULL); */
/*     // done - now exit thread */
/*     pthread_exit(NULL); */
/* } */

void debug_datastructures(){
    int* i;
    printf("\n\n########################################\n");
    printf("PPU debug-section for datastructures\n\n");
    printf("Pointer size: %i\n",sizeof(i));
    printf("DataManager size: %i (16*%i = %i)\n", 
	   sizeof(DataManager), sizeof(DataManager)/16, 
	   (sizeof(DataManager)/16)*16);
    printf("DataManagers[0],[1] offset: %lx, %lx\n\n", 
	   (long unsigned int)&dataManagers[0], 
	   (long unsigned int)&dataManagers[1]);
    printf("Group size: %i (16*%i = %i)\n", 
	   sizeof(Group), sizeof(Group)/16, 
	   (sizeof(Group)/16)*16);
    printf("Group pos [0], [1]: %lx, %lx\n\n", 
	   (long unsigned int)&dataManagers[0].thisgroup, 
	   (long unsigned int)&dataManagers[1].thisgroup);

    printf("ShVar size: %i\n",sizeof(ShVar));
    printf("PVar size: %i\n",sizeof(PVar));
    printf("CARR size: %i\n",sizeof(CyclicDistArray));
    printf("BARR size: %i\n",sizeof(BlockDistArray));
    printf("RARRr size: %i\n",sizeof(RepArray));
    printf("PARRr size: %i\n",sizeof(PrivateArray));
    printf("########################################\n");
    return;
}

//NestStep functions, these functions are the actual NestStep_init and 
//NestStep_finalize, the ones on the spu are only for managing contexts
//for the programs.
void NestStep_PPU_init(int* pargc, char*** pargv){
#ifdef DEBUG_PPU
    printf("     ###  ###  #  #   # #   # # #####     \n");
    printf("     #  # #  # #  #   # ##  # #   #       \n");
    printf("     ###  ###  #  #   # # # # #   #       \n");
    printf("     #    #    #  #   # #  ## #   #       \n");
    printf("     #    #     ##    # #   # #   #       \n");
#endif //DEBUG_PPU
    //    printf("size of dataManager: %i\n", sizeof(DataManager));
    //get size of group, create DataManagers, create a base group
    int i;
    spe_gid = spe_create_group (SCHED_OTHER, 0, 1);

    if (spe_gid == NULL) {
	fprintf(stderr, "Failed spe_create_group(errno=%d)\n", errno);
	exit(1);
    }
    //    GROUP_SIZE = 8; 
    GROUP_SIZE = spe_group_max(spe_gid); 
    if (GROUP_SIZE < 1) {
        fprintf(stderr, "System doesn't have a working SPE.  I'm leaving.\n");
        exit(1);
    }
    if(*pargc > 1){
	GROUP_SIZE = (*pargv)[1][0] - '0';
	printf("GROUP_SIZE: %i\n",GROUP_SIZE);
    }
    //NOTE: GROUP_SIZE will default to 16, but there are only 8 spe:s
    //this will force the GROUP_SIZE to be smaller than the defined
    //MAX_SPE_THREADS which is set to be 8 at the moment.
    //this is because the kernel does not allow scheduling of spu-threads
    if(GROUP_SIZE > MAX_SPE_THREADS){
	printf("group_size specified exceeds max_spe_threads, setting groupsize to: %i\n",MAX_SPE_THREADS);
	GROUP_SIZE = MAX_SPE_THREADS;
    }

    Group *thisgroup = new_Group();

    thisgroup->parent = NULL;
    thisgroup->tree = NULL;
    thisgroup->path[0] = 0;
    thisgroup->path[1] = -1;
    thisgroup->clist = new_CombineList();
    //thisgroup->groupInfo.name = (char *) malloc (sizeof(char) * 2);
    thisgroup->groupInfo.name = (char *) memalign (16,sizeof(char) * 2);
    //    printf("original addr of name: %x\n",thisgroup->groupInfo.name);
    thisgroup->groupInfo.name[0] = '0'; // name of first group is "0"
    thisgroup->groupInfo.name[1] = '\0';
    thisgroup->groupInfo.size = GROUP_SIZE;
    thisgroup->groupInfo.my_rank = 0;
    thisgroup->groupInfo.my_rank_local = 0;
    thisgroup->groupInfo.first_rank = 0;
    thisgroup->groupInfo.no_of_groups = 1;
    //thisgroup->groupInfo.all_groups = (int**) malloc(sizeof(int*)*1);
    thisgroup->groupInfo.all_groups = (int**) memalign(16,sizeof(int*)*1);
    //thisgroup->groupInfo.all_groups[0] = (int *) malloc (sizeof(int) * 
    //							 GROUP_SIZE);
    thisgroup->groupInfo.all_groups[0] = (int *) memalign (16,sizeof(int) * 
    							 GROUP_SIZE);
    thisgroup->groupInfo.my_group = thisgroup->groupInfo.all_groups[0];
    thisgroup->groupInfo.my_group_number = 0;

    for(i=0; i<GROUP_SIZE;i++){
	thisgroup->groupInfo.my_group[i] = i;
    }

    //thisgroup->groupInfo.group_leaders = (int*)malloc(sizeof(int)*1);
    thisgroup->groupInfo.group_leaders = (int*)memalign(16,sizeof(int)*1);
    thisgroup->groupInfo.group_leaders[0] = 0;
    thisgroup->groupInfo.parent = NULL;

    //    thisgroup->groupInfo.sizes = (int *) malloc (sizeof(int) * 1);
    //    thisgroup->groupInfo.sizes[0] = GROUP_SIZE;

    for(i=0;i<GROUP_SIZE;i++){
	dataManagers[i].thisgroup = new_Group(); //malloc(sizeof(Group));
	//printf("#PPU group_addr[%i]: %x\n",i,dataManagers[i].thisgroup);
	//	printf("#PPU dataManager_addr[%i]: %x\n",i,&(dataManagers[i]));
	//	printf("#PPU size of ll: %i\n",sizeof(long long*));
	groupcpy(dataManagers[i].thisgroup, thisgroup);
        dataManagers[i].thisgroup->groupInfo.my_rank = i;
        dataManagers[i].thisgroup->groupInfo.my_rank_local = i;
        dataManagers[i].thisgroup->groupInfo.my_group = NULL;
	dataManagers[i].thisgroup->clist = new_CombineList();
	dataManagers[i].mboxAddr.p = (void*)(&(dataManagers[i].mboxExtraData));
    }
    
    free_Group(thisgroup);

#ifdef DEBUG_PPU
    printf("SPE group max: %i\n", GROUP_SIZE);
    debug_datastructures();
#endif // DEBUG_PPU
    return;
}

void NestStep_PPU_finalize(){
    int i;
#ifdef DEBUG_PPU
    printf("       NestStep finalize             \n");
    printf(" ###  ###  #  #    #### #   # ###    \n");
    printf(" #  # #  # #  #    #    ##  # #  #   \n");
    printf(" ###  ###  #  #    ##   # # # #  #   \n");
    printf(" #    #    #  #    #    #  ## #  #   \n");
    printf(" #    #     ##     #### #   # ###    \n");
#endif // DEBUG_PPU
    for(i=0;i<GROUP_SIZE;i++){
	//	free_CombineList(dataManagers[i].thisgroup->clist);
	//	free_Group(dataManagers[i].thisgroup);
	//	printf("free datamanager: %i\n", i);
	free_DataManager(&dataManagers[i]);
	//	printf("after free datamanager: %i\n",i);
    }
    //clear DataManagers, clear base group

    return;
}

int NestStep_get_size(){

    return GROUP_SIZE;
}

Group* new_Group(){

    Group* tmp;
    //    tmp = (Group*)malloc(sizeof(Group));

    //needs to be 16 byte aligned so that we can transfer it to the
    //spu, should only be an issue when running in 32bit mode
    tmp = (Group*)memalign(16,sizeof(Group)); 
    tmp->tree = NULL;
    tmp->parent = NULL;
    tmp->clist = NULL;
    tmp->groupInfo.name = NULL;
    tmp->groupInfo.my_group = NULL;
    tmp->groupInfo.all_groups = NULL;
    tmp->groupInfo.group_leaders = NULL;
    tmp->groupInfo.parent = NULL;

    return tmp;

}
void free_Group(Group* _g){
    free(_g->groupInfo.name);
    free(_g->groupInfo.my_group);
    free(_g->groupInfo.all_groups);
    free(_g->groupInfo.group_leaders);
    // NOTE: we don't have parents/NestStep_neststep .. remove from group?
    //free(_g->groupInfo.parent);

    // NOTE: we don't have parents/NestStep_neststep ..    
    //free(_g->parent); 


    free(_g->tree);
    free_CombineList(_g->clist);
   //free(_g->clist); // included in free_CombineList
    free(_g);

    return;
}
//this should make an identical copy, all references should reference
//new objects, no shared references should come out of this function
void groupcpy(Group* _out, Group* _in){
    int i;
    int namelen = strlen(_in->groupInfo.name);

    _out->groupInfo.name = (char *) memalign(16,sizeof(char)*(namelen+1));
    //_out->groupInfo.name = (char *) malloc(sizeof(char) * (namelen +1));
    for(i=0;i<namelen;i++){
	_out->groupInfo.name[i] = _in->groupInfo.name[i];
    }
    //    printf("new addr of name: %x\n",_out->groupInfo.name);
    _out->groupInfo.size = _in->groupInfo.size;
    _out->groupInfo.my_rank = _in->groupInfo.my_rank;

    _out->groupInfo.my_rank_local = _in->groupInfo.my_rank_local;
    _out->groupInfo.first_rank = _in->groupInfo.first_rank;

    //NOTE: we have no groups .. remove the following 7?
    _out->groupInfo.my_group = NULL; 
    //    for(i=0; i < sizeof(_in->groupInfo.my_group)/sizeof(int);i++){
    //	_out->groupInfo.my_group[i] = _in->groupInfo.my_group[i];
    //    }
    _out->groupInfo.my_group_number = _in->groupInfo.my_group_number;
    _out->groupInfo.no_of_groups = _in->groupInfo.no_of_groups;
    //    _out->groupInfo.all_groups = malloc(no_of_groups*sizeof(int*));
    //    for(i=0;i<no_of_groups;i++){
    //	_out->groupInfo.all_groups[i] = malloc(sizeof(int)
    _out->groupInfo.all_groups = NULL;
    _out->groupInfo.group_leaders = NULL;

    _out->groupInfo.parent = _in->groupInfo.parent;
    _out->parent = _in->parent;

    // NOTE: this is only called in the beginning so no need to care about making
    //complete copies of these
    _out->tree = NULL; 
    _out->clist = NULL; 
    for(i=0;i<MAX_DEPTH;i++){
	_out->path[i] = _in->path[i];
    }

    return;
}
void free_DataManager(DataManager* _dm){
    //    int i;
    //    printf("free group\n");
    free_Group(_dm->thisgroup);

    //NOTE: would be nice to get rid of some ugly things from NestStep-RTS
    //but that will have to wait until things work..

    ShVar* shvar = _dm->shVars;
    PVar* pvar = _dm->privateVars;
    RepArray* rarr = _dm->repArrays;
    PrivateArray* parr = _dm->privateArrays;
    CyclicDistArray* carr = _dm->cyclicDistArrays;
    BlockDistArray* barr = _dm->blockDistArrays;


    while(shvar != NULL){
	_dm->shVars = _dm->shVars->next;
	free_ShVar(shvar);
	shvar = _dm->shVars;
    }
    while(pvar != NULL){
	_dm->privateVars = _dm->privateVars->next;
	free_PVar(pvar);
	pvar = _dm->privateVars;
    }    
    while(rarr != NULL){
	_dm->repArrays = _dm->repArrays->next;
	free_RepArray(rarr);
	rarr = _dm->repArrays;
    }    
    while(parr != NULL){
	_dm->privateArrays = _dm->privateArrays->next;
	free_PrivateArray(parr);
	parr = _dm->privateArrays;
    }    
    while(barr != NULL){
	_dm->blockDistArrays = _dm->blockDistArrays->next;
	free_BlockDistArray(barr);
	barr = _dm->blockDistArrays;
    }
    while(carr != NULL){
	_dm->cyclicDistArrays = _dm->cyclicDistArrays->next;
	free_CyclicDistArray(carr);
	carr = _dm->cyclicDistArrays;
    }

    return;
}

void handle_attach_variable(int spu_no){
    MboxExtraData *mb_data = &dataManagers[spu_no].mboxExtraData;
    ShVar* tmp_shvar = NULL;
    RepArray* tmp_rarr = NULL;
    PVar* tmp_pvar = NULL;
    PrivateArray* tmp_parr = NULL;
    int combine_strategy = mb_data->datatype;
    int prefix_tag = mb_data->size;
    Addr64 prefix_base;
    int updated = mb_data->blocksize;
    
    //    printf("handle_attach_variable(%i)\n",spu_no);
    if(mb_data->type == SHVAR){
	tmp_shvar = locate_ShVar(mb_data->name, &dataManagers[spu_no]);
	if(prefix_tag == PREFIX){
	    tmp_pvar = locate_PVar(mb_data->prefixname, &dataManagers[spu_no]);
	    prefix_base = tmp_pvar->base;
	}
	//	dataManagers[spu_no].thisgroup->clist
	if(!tmp_shvar){
	    printf("can't attach non-existing variable!\n");
	    printf("combine_strategy: %i\n",combine_strategy);
	    printf("PPU:name: %i, %i\n",mb_data->name.procedure,mb_data->name.relative);
	}
	NestStep_ShVar_attach(tmp_shvar,
			      combine_strategy,
			      NULL,
			      prefix_tag,
			      prefix_base,
			      updated,
			      dataManagers[spu_no].thisgroup->clist,
			      spu_no);
	return;
    }
    if(mb_data->type == RARR){
	tmp_rarr = locate_RepArray(mb_data->name, &dataManagers[spu_no]);
	if(prefix_tag == PREFIX){
	    tmp_parr = locate_PrivateArray(mb_data->prefixname, &dataManagers[spu_no]);
	    prefix_base = tmp_parr->base;
	}
	if(!tmp_rarr){
	    printf("can't attach non-existing variable!\n");
	    printf("combine_strategy: %i\n",combine_strategy);
	    printf("PPU:name: %i, %i\n",mb_data->name.procedure,mb_data->name.relative);
	}
	NestStep_RepArray_attach(tmp_rarr,
				 combine_strategy,
				 NULL,
				 prefix_tag,
				 prefix_base,
				 updated,
				 dataManagers[spu_no].thisgroup->clist,
				 spu_no);
	return;
    }
}

void handle_reparray_range(int spu_no){
    MboxExtraData *mb_data = &dataManagers[spu_no].mboxExtraData;
    RepArray* tmp_rarr = locate_RepArray(mb_data->name, &dataManagers[spu_no]);
    NestStep_RepArray_range(tmp_rarr,mb_data->size, mb_data->blocksize,spu_no);
    return;
}

void handle_create_variable(int spu_no){
    ShVar* tmp_shvar = NULL;
    PVar* tmp_pvar = NULL;
    BlockDistArray* tmp_barr = NULL;
    CyclicDistArray* tmp_carr = NULL;
    PrivateArray* tmp_parr = NULL;
    RepArray* tmp_rarr = NULL;

    Addr64 tmp_int;
    Addr64 tmp_float;
    Addr64 tmp_double;

    MboxExtraData *mb_data = &dataManagers[spu_no].mboxExtraData;

    if(mb_data->type == PVAR){
      //create the pvar
      if(mb_data->datatype == IVAR){
	tmp_int.p = (void*)memalign(16,sizeof(int));
	tmp_pvar = new_PVar(mb_data->name,
			    mb_data->datatype,
			    tmp_int);
      }
      if(mb_data->datatype == FVAR){
	tmp_float.p = (void*)memalign(16,sizeof(float));
	tmp_pvar = new_PVar(mb_data->name,
			    mb_data->datatype,
			    tmp_float);
      }
      if(mb_data->datatype == DVAR){
	tmp_double.p = (void*)memalign(16,sizeof(double));
	tmp_pvar = new_PVar(mb_data->name,
			    mb_data->datatype,
			    tmp_double);
      }
      //insert tmp_pvar into the datamanager
      add_PVar(tmp_pvar, &dataManagers[spu_no]);
      //store the address of the allocation in mboxExtraData
      mb_data->addr = tmp_pvar->base;
    }
    if(mb_data->type == SHVAR){
      //create the shvar
      if(mb_data->datatype == IVAR){
	tmp_int.p = (void*)memalign(16,sizeof(int));
	tmp_shvar = new_ShVar(mb_data->name, mb_data->datatype, tmp_int);
      }
      if(mb_data->datatype == FVAR){
	tmp_float.p = (void*)memalign(16,sizeof(float));
	tmp_shvar = new_ShVar(mb_data->name, mb_data->datatype, tmp_float);
      }
      if(mb_data->datatype == DVAR){
	tmp_double.p = (void*)memalign(16,sizeof(double));
	tmp_shvar = new_ShVar(mb_data->name, mb_data->datatype, tmp_double);
      }
      //insert tmp_shvar into the datamanager
      add_ShVar(tmp_shvar, &dataManagers[spu_no]);
      //store the address of the allocation in mboxExtraData
      mb_data->addr = tmp_shvar->base;
    }
    if(mb_data->type == PARR){
      //create the parr
      if(mb_data->datatype == IVAR){
	tmp_int.p = (void*)memalign(128,sizeof(int)*mb_data->size);
	//	printf("cbuffer[%i]: %x\n",spu_no,tmp_int.p);
	tmp_parr = new_PrivateArray(mb_data->name, mb_data->datatype,
				    mb_data->size, tmp_int);
      }
      if(mb_data->datatype == FVAR){
	tmp_float.p = (void*)memalign(128,sizeof(float)*mb_data->size);
	tmp_parr = new_PrivateArray(mb_data->name, mb_data->datatype,
				    mb_data->size, tmp_float);
      }
      if(mb_data->datatype == DVAR){
	tmp_double.p = (void*)memalign(128,sizeof(double)*mb_data->size);
	tmp_parr = new_PrivateArray(mb_data->name, mb_data->datatype,
				    mb_data->size,tmp_double);
      }
      //insert tmp_parr into the datamanager
      add_PrivateArray(tmp_parr, &dataManagers[spu_no]);
      //store the address of the allocation in mboxExtraData
      mb_data->addr = tmp_parr->base;
    }
    if(mb_data->type == BARR){
      //create the barr
	//NOTE: these three can be combined into one
      if(mb_data->datatype == IVAR){
	tmp_barr = new_BlockDistArray(mb_data->name, mb_data->datatype,
				      mb_data->size, spu_no);
      }
      else if(mb_data->datatype == FVAR){
	tmp_barr = new_BlockDistArray(mb_data->name, mb_data->datatype,
				      mb_data->size, spu_no);
      }
      else if(mb_data->datatype == DVAR){
	tmp_barr = new_BlockDistArray(mb_data->name, mb_data->datatype,
				      mb_data->size, spu_no);
      } else {
	printf("Error when creating BARR, unknown datatype: %i\n",mb_data->datatype);
	exit(1);
      }
      //insert tmp_barr into the datamanager
      add_BlockDistArray(tmp_barr, &dataManagers[spu_no]);
      //store the address of the allocation in mboxExtraData
      mb_data->addr = tmp_barr->base;
      mb_data->lower = tmp_barr->lower;
      mb_data->higher = tmp_barr->higher;
      mb_data->blocksize = tmp_barr->localsize;
    }
    if(mb_data->type == CARR){
      //create the carr
	//NOTE: these three can be combined into one.
      if(mb_data->datatype == IVAR){
	tmp_carr = new_CyclicDistArray(mb_data->name, mb_data->datatype,
				       mb_data->size, mb_data->blocksize,
				       spu_no);
      }
      if(mb_data->datatype == FVAR){
	tmp_carr = new_CyclicDistArray(mb_data->name, mb_data->datatype,
				       mb_data->size, mb_data->blocksize,
				       spu_no);
      }
      if(mb_data->datatype == DVAR){
	tmp_carr = new_CyclicDistArray(mb_data->name, mb_data->datatype,
				       mb_data->size, mb_data->blocksize,
				       spu_no);
      }
      //insert tmp_carr into the datamanager
      add_CyclicDistArray(tmp_carr, &dataManagers[spu_no]);
      //store the address of the allocation in mboxExtraData
      mb_data->addr = tmp_carr->base;
    }
    if(mb_data->type == RARR){
      //create the rarr
      if(mb_data->datatype == IVAR){
	tmp_int.p = (void*)memalign(16,sizeof(int)*mb_data->size);
	tmp_rarr = new_RepArray(mb_data->name, mb_data->datatype,
				mb_data->size, tmp_int);
      }
      if(mb_data->datatype == FVAR){
		tmp_float.p = (void*)memalign(16,sizeof(float)*mb_data->size);
	tmp_rarr = new_RepArray(mb_data->name, mb_data->datatype,
				mb_data->size, tmp_float);
      }
      if(mb_data->datatype == DVAR){
	tmp_double.p = (void*)memalign(16,sizeof(double)*mb_data->size);
	tmp_rarr = new_RepArray(mb_data->name, mb_data->datatype,
				mb_data->size, tmp_double);
      }
      //insert tmp_rarr into the datamanager
      add_RepArray(tmp_rarr, &dataManagers[spu_no]);
      //store the address of the allocation in mboxExtraData
      mb_data->addr = tmp_rarr->base;
    }
    return;
}

void handle_remove_variable(int spu_no){
    ShVar* tmp_shvar = NULL;
    PVar* tmp_pvar = NULL;
    BlockDistArray* tmp_barr = NULL;
    CyclicDistArray* tmp_carr = NULL;
    PrivateArray* tmp_parr = NULL;
    RepArray* tmp_rarr = NULL;

    MboxExtraData *mb_data = &dataManagers[spu_no].mboxExtraData;

    if(mb_data->type == PVAR){
      //find the variable
      tmp_pvar = locate_PVar(mb_data->name, &dataManagers[spu_no]);
      //remove the variable from datamanager
      remove_PVar(tmp_pvar, &dataManagers[spu_no]);
      //free the variable
      free_PVar(tmp_pvar);
    }
    if(mb_data->type == SHVAR){
      //find the variable
      tmp_shvar = locate_ShVar(mb_data->name, &dataManagers[spu_no]);
      //remove the variable from datamanager
      remove_ShVar(tmp_shvar, &dataManagers[spu_no]);
      //free the variable
      free_ShVar(tmp_shvar);
    }
    if(mb_data->type == PARR){
      //find the variable
      tmp_parr = locate_PrivateArray(mb_data->name, &dataManagers[spu_no]);
      //remove the variable from datamanager
      remove_PrivateArray(tmp_parr, &dataManagers[spu_no]);
      //free the variable
      free_PrivateArray(tmp_parr);
    }
    if(mb_data->type == BARR){
      //find the variable
      tmp_barr = locate_BlockDistArray(mb_data->name, &dataManagers[spu_no]);
      //remove the variable from datamanager
      remove_BlockDistArray(tmp_barr, &dataManagers[spu_no]);
      //free the variable
      free_BlockDistArray(tmp_barr);
    }
    if(mb_data->type == CARR){
      //find the variable
      tmp_carr = locate_CyclicDistArray(mb_data->name, &dataManagers[spu_no]);
      //remove the variable from datamanager
      remove_CyclicDistArray(tmp_carr, &dataManagers[spu_no]);
      //free the variable
      free_CyclicDistArray(tmp_carr);
    }		
    if(mb_data->type == RARR){
      //find the variable
      tmp_rarr = locate_RepArray(mb_data->name, &dataManagers[spu_no]);
      //remove the variable from datamanager
      remove_RepArray(tmp_rarr, &dataManagers[spu_no]);
      //free the variable
      free_RepArray(tmp_rarr);
    }
    return;
}

void attach_update_or_mirror_variable(int spu_no, int request_type){
    MboxExtraData *mb_data = &dataManagers[spu_no].mboxExtraData;
    CombineList* tmp_clist = dataManagers[spu_no].thisgroup->clist;
    PrivateArray* tmp_parr;
    if(mb_data->type == BARR){
	//	printf("BARR req: %i\n",request_type);
	BlockArrayInfo* barrinfo = malloc(sizeof(BlockArrayInfo));

	tmp_parr = locate_PrivateArray(mb_data->prefixname, &dataManagers[spu_no]);
	//	printf("aspu_no:%i\n",spu_no);
	namecpy(&(barrinfo->name), &(mb_data->name));
	barrinfo->buffer = tmp_parr->base;
		//printf("abuffer: %x\n",barrinfo->buffer.p);
	barrinfo->lower = mb_data->lower;
	barrinfo->higher = mb_data->higher;
	barrinfo->request = request_type;
	barrinfo->next = NULL;
	barrinfo->prev = NULL;

	if(!tmp_clist->block_info){
	    tmp_clist->block_info = barrinfo;
	} else {
	    BlockArrayInfo* tmp_bai = tmp_clist->block_info;
	    while(tmp_bai->next){
		tmp_bai = tmp_bai->next;
	    }
	    tmp_bai->next = barrinfo;
	    barrinfo->prev = tmp_bai;
	}
    }
    if(mb_data->type == CARR){
	printf("CARR req: %i\n",request_type);
	CyclicArrayInfo* carrinfo = malloc(sizeof(CyclicArrayInfo));

	tmp_parr = locate_PrivateArray(mb_data->prefixname, &dataManagers[spu_no]);
	
	namecpy(&(carrinfo->name),&(mb_data->name));
	carrinfo->buffer = tmp_parr->base;
	carrinfo->lower = mb_data->lower;
	carrinfo->higher = mb_data->higher;
	carrinfo->request = request_type;
	carrinfo->next = NULL;
	carrinfo->prev = NULL;

	if(!tmp_clist->cyclic_info){
	    tmp_clist->cyclic_info = carrinfo;
	} else {
	    CyclicArrayInfo* tmp_cai = tmp_clist->cyclic_info;
	    while(tmp_cai->next){
		tmp_cai = tmp_cai->next;
	    }
	    tmp_cai->next = carrinfo;
	    carrinfo->prev = tmp_cai;
	}	
    }
}

//run a spu-program and manage all calls to create variables, finishes
//when program is done
void NestStep_run_spu_thread(spe_program_handle_t thread){
    int i;
    int status;
    int data;
	Ls_pointers ls[MAX_SPE_THREADS]__attribute__ ((aligned(16)));
    /* allocate the SPE task */
    for(i=0;i<GROUP_SIZE;i++){
	//printf("size of pointer, %i\n", sizeof(int*));
#ifdef DEBUG_PPU
	printf("PPU: addr of datamanager[%i]: %x\n",i,(int)&dataManagers[i]);
	printf("PPU: addr of group[%i]: %x\n",i,dataManagers[i].thisgroup);
#endif // DEBUG_PPU
	spe_id[i] = spe_create_thread (spe_gid, &thread,
				       (unsigned long long *) (&(dataManagers[i])),
				       NULL, -1, SPE_MAP_PS|SPE_CFG_SIGNOTIFY1_OR); 
	
        if (spe_id[i] == NULL) {
            fprintf (stderr, "FAILED: spe_create_thread(num=%d, errno=%d)\n", 0, errno);
            exit(1);
		}
		
		ls[i].ls = (unsigned int)spe_get_ls(spe_id[i]);
		//spe_sig_notify_1_area_t* sig1 = (spe_sig_notify_1_area_t*)spe_get_ps_area(spe_id[i],SPE_SIG_NOTIFY_1_AREA);
		//ls[i].sig1 = (unsigned int)(&sig1->SPU_Sig_Notify_1);
		ls[i].sig1 = (unsigned int)(&((spe_sig_notify_1_area_t*)spe_get_ps_area(spe_id[i],SPE_SIG_NOTIFY_1_AREA))->SPU_Sig_Notify_1);
		ls[i].sig2 = (unsigned int)(&((spe_sig_notify_2_area_t*)spe_get_ps_area(spe_id[i],SPE_SIG_NOTIFY_2_AREA))->SPU_Sig_Notify_2);
		//ls[i] = (unsigned int)spe_get_ps_area(spe_id[i],SPE_SIG_NOTIFY_1_AREA);
		//ls[i] = (unsigned int)(&sig1->SPU_Sig_Notify_1);
		//printf("diff: 0x%x\n", -(ls[i] - (unsigned int)spe_get_ls(spe_id[i])));
    }

#ifdef DEBUG_PPU
    printf("All SPE:s initialized, proceeding to manage requests\n");
#endif //DEBUG_PPU

    //Poll the spu:s for events and handle them
    // possible events are: CREATE_VARIABLE, REMOVE_VARIABLE and DESTROY_THREAD
    //NOTE: debug counters
    for(i=0;i<GROUP_SIZE;i++){
	no_shvars[i]=0;
	no_barrs[i]=0;
	no_calls[i]=0;
	no_create_variable[i]=0;
	no_destroy_thread[i]=0;
    }

    while(TRUE){
	for(i=0;i<GROUP_SIZE ;i++){
	    status = NO_DATA;
	    data = NO_DATA;
	    errno = 0;
	    status = spe_stat_out_mbox(spe_id[i]);

	    if(status > 1){
		printf("Error, too many posts in mbox, this isn't possible\n");
		exit(1);
	    }
	    if(status > 0){
		data = spe_read_out_mbox(spe_id[i]); // this is non-blocking
		int j;
		for(j=0;j<100;j++){
		    //NOTE: spe_read_out_mbox doesnt clear the fucking flag
		    // perhaps solving it like this can come back later and bite me in the ass
		    // be wary!
		    if(spe_stat_out_mbox(spe_id[i])){
			int k;
			k = spe_read_out_mbox(spe_id[i]);
			if(k != data){
			    printf("double post in mbox and they differ\n");
			    exit(1);
			}
			//#ifdef DEBUG_PPU_MESSAGE_HANDLING
			printf("error: !\n");
			printf("stat_mbox:%i\n",spe_stat_out_mbox(spe_id[i]));
			printf("read_out_mbox:%i\n",k);
			printf("spu_id:%lx\n",(unsigned long int)&dataManagers[i]);
			printf("stat_mbox:%i\n",spe_stat_out_mbox(spe_id[i]));
			//#endif // DEBUG_PPU_MESSAGE_HANDLING
			//		    exit(1);
		    } 
		}

		// non-blocking,
		// read one msg
		no_calls[i]++;
		if(data == 1){
		    no_create_variable[i]++;
		}
		if(data == 3){
		    no_destroy_thread[i]++;
		}
	    } else if(status == -1){
		//		printf("errno was: %i\n", errno);
		//		printf("errno EIO: %i\n", EIO);
		//		printf("errno ESRCH: %i\n", ESRCH); //spe-id is invalid
		                                    //(couldnt be found)
		//		printf("errno EBADF: %i\n",EBADF); // NOTE: bad file-number ..? why? -> spufs.
		//NOTE: if the context we're polling is already destroyed,
		//then these two errors will pop out
		//		errno = 0;
		//		continue;
		// ESRCH, EBADF, ESPIPE can occur here
		if(errno == ESRCH || errno == EBADF || errno == ESPIPE){
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		    printf("Error when handling messages, spe has already terminated\n");
		    printf("Error: %s\n", strerror(errno));
#endif // DEBUG_PPU_MESSAGE_HANDLING
		    errno = 0;
		    continue; // stop current iteration and go to the next
		}
		else{
		    printf("Error, this shouldnt happen");
		    perror("Error in messagehandling, now aborting");
		    printf("errno was: %i\n", errno);
		    //		    printf("errno EIO: %i\n", EIO);
		    //		    printf("errno ESRCH: %i\n", ESRCH); //spe-id is invalid
		    //(couldnt be found)
		    //		    printf("errno EBADF: %i\n",EBADF); // NOTE: bad file-number ..? why? -> spufs

		    exit(1);
		}
	    } else {
		data = NO_DATA;
		continue;
	    }

	    //usleep(1); //NOTE problem fixed, if spu-thread finishes while it has
	    //things in it's mbox things will be bad. this is due to mboxes not
	    //beeing cleared as they should so we will get out of synch.
	    //problem: why in the hell does the program crash
	    // here if i don't have this? does the spe_read_out_mbox
	    // return before the value is written to data so that this
	    // if-clause is executed before the correct value is in it?
	    // and then the message is lost? or what?
	    // Solved? if the context we're working with is already destroyed
	    // then we will have errno set and things will break, code is now
	    // changed to handle this. the sleep would ensure that all nodes have
	    // completed up to a certain point before polling for events
	    // this would lead to all nodes beeing ready for polling when they were
	    // polled and thus they would not try to fetch data from a node 
	    //that wasn't ready
	    if(data == CREATE_VARIABLE){
		handle_create_variable(i);
		//signal the spe
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		printf("sending CREATE_VARIABLE to %lx\n",
		       (unsigned long int)&dataManagers[i]);
#endif // DEBUG_PPU_MESSAGE_HANDLING
		spe_write_in_mbox(spe_id[i],CREATE_VARIABLE);
	    }
	    if(data == REMOVE_VARIABLE){
		handle_remove_variable(i);
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		printf("sending REMOVE_VARIABLE to %lx\n",
		       (unsigned long int)&dataManagers[i]);
#endif // DEBUG_PPU_MESSAGE_HANDLING
		spe_write_in_mbox(spe_id[i],REMOVE_VARIABLE);
	    }

	    if(data == ATTACH_VARIABLE){
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		printf("ATTACH_VARIABLE\n");
#endif // DEBUG_PPU_MESSAGE_HANDLING
		handle_attach_variable(i);
		spe_write_in_mbox(spe_id[i],ATTACH_VARIABLE);
	    }
	    if(data == REPARRAY_RANGE){
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		printf("REPARRAY_RANGE\n");
#endif // DEBUG_PPU_MESSAGE_HANDLING
		handle_reparray_range(i);
		spe_write_in_mbox(spe_id[i],REPARRAY_RANGE);
	    }	    
	    if(data == COMBINE){
		combine_threads += 1;
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		printf("COMBINE, no_threads = %i\n",combine_threads);
#endif // DEBUG_PPU_MESSAGE_HANDLING
		if(combine_threads == GROUP_SIZE){
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		    printf("GROUP_SIZE:%i, no_combine: %i\n",
			   GROUP_SIZE,
			   combine_threads);
#endif // DEBUG_PPU_MESSAGE_HANDLING
		    combine_threads = 0;
		    NestStep_combine(dataManagers);
		    int j;
		    for(j=0;j<GROUP_SIZE;j++){
			spe_write_in_mbox(spe_id[j],COMBINE);
		    }
		}
		
	    }
		if(data == GET_ALL_LS_ADDR){
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		printf("GET_ALL_LS_ADDR\n");
#endif // DEBUG_PPU_MESSAGE_HANDLING
			dataManagers[i].mboxExtraData.addr.p = ls;
		//	printf("lsptr: 0x%x\n", ls);
			spe_write_in_mbox(spe_id[i],GET_ALL_LS_ADDR);
	    }
	    //NOTE: these need a PArr so the spu has a buffer in main
	    //memory to store things in. these are handled last in the
	    //combine-phase
	    if(data == MIRROR_VARIABLE){
		//attach mirror variable to combinelist
		attach_update_or_mirror_variable(i, data);
		spe_write_in_mbox(spe_id[i],MIRROR_VARIABLE);
	    }
	    if(data == UPDATE_VARIABLE){
		//attach update variable to combinelist
		attach_update_or_mirror_variable(i, data);
		spe_write_in_mbox(spe_id[i],UPDATE_VARIABLE);
	    }
	    if(data == DESTROY_THREAD){
#ifdef DEBUG_PPU_MESSAGE_HANDLING
		printf("debug counters:\n");
		//		int j;
		//		for(j=0;j<1;j++){ //GROUP_SIZE;j++){
		printf("Thread: %lx\n",(unsigned long int)&dataManagers[i]);
		printf("no_calls: %i\n",no_calls[i]);
		printf("no_create_varaible: %i\n",no_create_variable[i]);
		printf("no_destroy_thread: %i\n",no_destroy_thread[i]);
		//		}

		printf("PPU Destroy Thread id:%lx!\n",(unsigned long int)&dataManagers[i]);
#endif // DEBUG_PPU_MESSAGE_HANDLING
		spe_write_in_mbox(spe_id[i],DESTROY_THREAD);
		// blocking, will wait until thread is ready to be
		// removed 
		spe_wait(spe_id[i], &status, 0);
		/* returns the SPE status */
		if (WIFEXITED(status)) {
		    if (WEXITSTATUS(status)) {
			fprintf(stderr, "FAILED: SPE(%lx) returned a non-zero exit status\n",(long unsigned int)&dataManagers[i]);
			fprintf(stderr, "status was: %i\n", status);
			exit(1);
		    }
		} else {
		    fprintf(stderr, "FAILED: SPE(%lx) abnormally terminated\n",(long unsigned int)&dataManagers[i]);
		    exit(1);
		}
		finished_threads += 1;
		spe_id[i] = 0;
		if(finished_threads == GROUP_SIZE){
		    finished_threads = 0;
		    return;
		}
	    }
	    if(data == NESTSTEP_SPU_FATAL_ERROR){
		fprintf(stderr,"NestStep encountered a fatal error when running on the SPU, now exiting\n");
		exit(1);
	    }
	}
	if(spe_stat_out_mbox(spe_id[i]) > 0){
	    printf("Duplicate post in mbox, erronous spulib?!\n");
	    printf("data:%i\n",spe_read_out_mbox(spe_id[i]));
	    exit(1);
	}
    }

    return;
}


//ShVar, PVar, RepArray, PArray, BlockDistArray and CyclicDistArray 
//Constructors and Destructors, these will add the variable
//to the DataManager object(s)

//create a new private variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
PVar* new_PVar(Name _name, int _type, Addr64 _base){
    PVar* tmp_pvar;
    tmp_pvar = (PVar*)memalign(16, sizeof(PVar));
    namecpy(&(tmp_pvar->name), &_name);
    tmp_pvar->type = _type;
    tmp_pvar->base = _base;
    tmp_pvar->next = tmp_pvar->prev = NULL;
    return tmp_pvar;
}

void free_PVar(PVar* var){
    free(var->base.p);
    free(var);
    return;
}


//create a new shared variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
ShVar* new_ShVar(Name _name, int _type, Addr64 _base){
    ShVar* tmp_shvar;
    tmp_shvar = (ShVar*)memalign(16, sizeof(ShVar));
    namecpy(&(tmp_shvar->name), &_name);
    tmp_shvar->type = _type;
    tmp_shvar->base = _base;
    tmp_shvar->next = tmp_shvar->prev = NULL;
    return tmp_shvar;
}

void free_ShVar(ShVar* var){
    free(var->base.p);
    free(var);
    return;
}

//create a new private array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
PrivateArray* new_PrivateArray(Name _name, int _type, int _size, Addr64 _base){
    PrivateArray* tmp_parr;
//    tmp_parr = (PrivateArray*)memalign(16,sizeof(PrivateArray));
    tmp_parr = (PrivateArray*)memalign(16,sizeof(PrivateArray));
    namecpy(&(tmp_parr->name),&_name);
    tmp_parr->type = _type;
    tmp_parr->size = _size;
    tmp_parr->base = _base;
    tmp_parr->next = tmp_parr->prev = NULL;
    return tmp_parr;
}

void free_PrivateArray(PrivateArray* var){
    free(var->base.p);
    free(var);
    return;
}

//create a new replicated array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
RepArray* new_RepArray(Name _name, int _type, int _size, Addr64 _base){
    RepArray* tmp_rarr;
    tmp_rarr = (RepArray*)memalign(16,sizeof(RepArray));
    namecpy(&(tmp_rarr->name),&_name);
    tmp_rarr->type = _type;
    tmp_rarr->size = _size;
    tmp_rarr->base = _base;
    tmp_rarr->next = tmp_rarr->prev = NULL;
    tmp_rarr->url = 0;
    tmp_rarr->urh = _size-1;
    return tmp_rarr;
}

void free_RepArray(RepArray* var){
    free(var->base.p);
    free(var);
    return;
}

//create a new block array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
BlockDistArray* new_BlockDistArray(Name _name, int _type, int _size, int _spu){
    BlockDistArray* tmp_barr;
    int element_size;
    int no_elements;
    int ratio;
    int remainder;

    ratio = _size / NestStep_get_size();
    remainder = _size % NestStep_get_size();
    element_size = get_element_size(_type);
    no_elements = ratio + ((remainder > _spu) ? 1 : 0);

    tmp_barr = (BlockDistArray*)memalign(16,sizeof(BlockDistArray));
    namecpy(&(tmp_barr->name),&_name);
    tmp_barr->type = _type;
    tmp_barr->size = _size;
    tmp_barr->localsize = no_elements;

    tmp_barr->base.p = memalign(128, element_size*no_elements);
    //    tmp_barr->base.ull = (unsigned long)memalign(128, element_size*no_elements);
    //    printf("ppu addr: %llx\n",(unsigned long)tmp_barr->base.p);
//    printf("ratio: %i, remainder: %i, element_size: %i, no_elements: %i\n",
//	   ratio,remainder,element_size,no_elements);
//    printf("no_elements: %i\n",no_elements);
    tmp_barr->next = tmp_barr->prev = NULL;

    tmp_barr->lower = (_spu * ratio) + ((remainder > _spu) ? _spu : remainder); 
    tmp_barr->higher = tmp_barr->lower + no_elements - 1;
    if(tmp_barr->higher < 0){
	tmp_barr->higher = 0;
    }
    
    return tmp_barr;
}

void free_BlockDistArray(BlockDistArray* var){
    free(var->base.p);
    //free(var->base.ull);
    free(var);
    return;
}

//create a new cyclic array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, size is the total size of the array, blocksize
//is the size of each segment that is chosen modulo the number of threads
//allocated, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
CyclicDistArray* new_CyclicDistArray(Name _name, 
                                     int _type, 
                                     int _size, 
                                     int _blocksize,
				     int _spu){
    CyclicDistArray* tmp_carr;
    int element_size;
    int ratio;
    int remainder;
    int no_elements;

    tmp_carr = (CyclicDistArray*)memalign(16,sizeof(CyclicDistArray));
    namecpy(&(tmp_carr->name),&_name);
    tmp_carr->type = _type;
    tmp_carr->size = _size;
    tmp_carr->blocksize = _blocksize;
    element_size = get_element_size(_type);

    ratio = _size / NestStep_get_size();
    remainder = _size % NestStep_get_size();
    no_elements = ratio + ((remainder > _spu)?1:0);
    
    
    tmp_carr->base.p = memalign(128,element_size*no_elements);
    tmp_carr->next = tmp_carr->prev = NULL;

    return tmp_carr;
}

void free_CyclicDistArray(CyclicDistArray* var){
    free(var->base.p);
    free(var);
    return;
}
