/* ************************************************************** *
 * Library for the ppu, manages sending and receiving of messages *
 * to the spu as well as reading/writing to memory                *
 * ************************************************************** */

#ifndef NESTSTEP_PPU_H
#define NESTSTEP_PPU_H

//ppu specific functions and variables
#include "combine_ppu.h"
#include "import_export_ppu.h"
#include "tree_builders_ppu.h"
//common functions and variables for ppu and spu
//#include "../NestStep_common/common_arrays.h"
//#include "../NestStep_common/common_var.h"
//#include "../NestStep_common/common_constants.h"

#include "../NestStep_common/common_neststep.h"
#include "base_ppu.h"
//CELL and system specific includes
//#include<libspe2.h> // spe-handling
#include<libspe.h> // spe-handling

#include<sys/wait.h> // WIFEXITED && WEXITSTATUS



//CELL and system specific defines

typedef struct _thread_args{
    struct spe_context* spe;
    void* argp;
    void* envp;
} thread_args;

int finished_threads;
int combine_threads;
spe_gid_t spe_gid; // group id for the spes
//spe_gang_context_ptr_t spe_gid; // group id for the spes
speid_t spe_id[MAX_SPE_THREADS]; // id for communication with the spe
//spe_context_ptr_t spe_id[MAX_SPE_THREADS]; // id for communication with the spe
//pthread_t pts[MAX_SPE_THREADS];
//thread_args t_args[MAX_SPE_THREADS];

//this function is sent to the pthread responsible for handling the spe, it is
//put to sleep as soon as the spe starts running.
//void pthread_wrapper_for_spe(thread_args* arg);

//NOTE: counters for debugging
int no_shvars[MAX_SPE_THREADS];
int no_barrs[MAX_SPE_THREADS];
int no_calls[MAX_SPE_THREADS];
int no_create_variable[MAX_SPE_THREADS];
int no_destroy_thread[MAX_SPE_THREADS];

//NestStep functions, these functions are the actual NestStep_init and 
//NestStep_finalize, the ones on the spu are only for managing contexts
//for the programs.
void NestStep_PPU_init(int* pargc, char*** pargv); // NestStep Initialization
void NestStep_PPU_finalize(); // NestStep Finalize
int NestStep_get_size();

//functions for managing groups
Group* new_Group();
void free_Group(Group* _g);
void groupcpy(Group* _out, Group* _in);

//function for freeing up a datamanager
void free_DataManager(DataManager* _dm);

//ShVar, PVar, RepArray, PArray, BlockDistArray and CyclicDistArray 
//Constructors and Destructors, these will add the variable
//to the DataManager object(s)

//create a new private variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
PVar* new_PVar(Name _name, int _type, Addr64 _base);
void free_PVar(PVar* var);

//create a new shared variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
ShVar* new_ShVar(Name _name, int _type, Addr64 _base);
void free_ShVar(ShVar* var);

//create a new private array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
PrivateArray* new_PrivateArray(Name _name, int _type, int _size, Addr64 _base);
void free_PrivateArray(PrivateArray* var);

//create a new replicated array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
RepArray* new_RepArray(Name _name, int _type, int _size, Addr64 _base);
void free_RepArray(RepArray* var);

//create a new block array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
BlockDistArray* new_BlockDistArray(Name _name, int _type, int _size, int _spu);
void free_BlockDistArray(BlockDistArray* var);

//create a new cyclic array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, size is the total size of the array, blocksize
//is the size of each segment that is chosen modulo the number of threads
//allocated, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array
CyclicDistArray* new_CyclicDistArray(Name _name, int _type, int _size, int _blocksize, int _spu);
void free_CyclicDistArray(CyclicDistArray* var);


//NestStep Variables

//manager for handling data inside SPU threads, allows for allocating 
//datastructures that are larger than 256K and allows for persistent 
//data between different SPU code-segments split between several program-files
DataManager dataManagers[MAX_SPE_THREADS] __attribute__ ((aligned (128))); //NOTE: this doesn't give too large overhead so we allocate statically

//run a spu-program and manage all calls to create variables, finishes when program is done
void NestStep_run_spu_thread(spe_program_handle_t thread);

#endif //NESTSTEP_PPU_H
