#include "base_spu.h"

void write_mboxExtraData(){
/*     printf("write_mboxExtraData from: %llx, to: %llx\n", */
/* 	   (unsigned long long)&dataManager.mboxExtraData, */
/* 	   dataManager.mboxAddr.p); */

    mfc_put(&dataManager.mboxExtraData, dataManager.mboxAddr.p, 
	    sizeof(MboxExtraData), NESTSTEP_DMA_TAG, 0,0);

    mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
    mfc_read_tag_status_all();
    //    printf("SPU write_mboxExtraData\n");
    return;
}

void read_mboxExtraData(){
/*     printf("read_mboxExtraData from: %llx, to: %llx\n", */
/* 	   dataManager.mboxAddr.ull, */
/* 	   (unsigned long long)&dataManager.mboxExtraData */
/* 	   ); */
    
//    printf("size is: %i\n",sizeof(MboxExtraData));
   
    mfc_get(&dataManager.mboxExtraData,dataManager.mboxAddr.p,
	    sizeof(MboxExtraData), NESTSTEP_DMA_TAG,0,0);

    mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
    mfc_read_tag_status_all();
    //    printf("SPU read_mboxExtraData\n");
    return;
}

int get_available_mfc_chan(){
    int i;
    //return first available dma-channel
    for(i=2;i<31;i++){
	if(dataManager.tags[i] == 0){
	    dataManager.tags[i] = 1;
	    return i;
	}
    }
    // no available channels
    //NOTE: should i wait here with read_tag_status_all on some channel?
    //or return an error?
#ifdef DEBUG_MFC
    printf("No available mfc channels, waiting for one..\n");
#endif
    while(TRUE){
	for(i=2;i<31;i++){
	    if(dataManager.tags[i] == 0){
		dataManager.tags[i] = 1;
#ifdef DEBUG_MFC
		printf("Found channel: %i, returning\n",i);
#endif DEBUG_MFC
		return i;
	    }
	}
    }
    return 0;
}

void check_dma(char* fn, int lbound, int ubound, 
	       void* lsaddr, void* mmaddr, int type, int test){
    if(test == TEST_BOUNDS){
	if(lbound == OUT_OF_BOUNDS || ubound == OUT_OF_BOUNDS){
	    fprintf(stderr,"Error: %s, trying to get array from outside local scope\n",fn);
	    spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
	    exit(1);
	}
    }
    switch(type){
    case IVAR:
	switch(test){
	case TEST_SIZE_OFFSET:
	    if((ubound-lbound + 1)*sizeof(int) % 16 != 0){
		fprintf(stderr,"Error: %s with size %lu not a multiple of 16 byte\n",
			fn, (ubound-lbound+1)*sizeof(int)); // the +1 comes from transfer(0,1) = size2
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_SIZE:
	    if((ubound-lbound + 1)*sizeof(int) > 16384){
		fprintf(stderr,"Error: %s with size %lu larger than 16KB (16384B)\n",
		       fn, (ubound-lbound+1)*sizeof(int));
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR);  // the +1 comes from transfer(0,1) = size2
		exit(1);
	    }
	    break;
	case TEST_MMADDR:
	    if(0x0000000f & (int)((int*)mmaddr + lbound)){
		fprintf(stderr,"Error: %s with transfer from lbound %i starting at",fn, lbound);
		fprintf(stderr,"0x%x, not 16 byte aligned.\n",(unsigned int)((int*)mmaddr + lbound));
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_LSADDR:
	    if(0x0000000f & (int)((int*)lsaddr + lbound)){
		fprintf(stderr,"Error: %s with transfer from lbound %i starting at",fn, lbound);
		fprintf(stderr,"0x%x, not 16 byte aligned.\n",(unsigned int)((int*)lsaddr + lbound));
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	}
	break;
    case FVAR:
	switch(test){
	case TEST_SIZE_OFFSET:
	    if((ubound-lbound + 1)*sizeof(float) % 16 != 0){
		fprintf(stderr,"Error: %s with size %lu not a multiple of 16 byte\n",
		       fn, (ubound-lbound + 1)*sizeof(float)); // the +1 comes from transfer(0,1) = size2
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_SIZE:
	    if((ubound-lbound + 1)*sizeof(float) > 16384){
		fprintf(stderr,"Error: %s with size %lu larger than 16KB (16384B)\n",
		       fn, (ubound-lbound + 1)*sizeof(float)); // the +1 comes from transfer(0,1) = size2
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_MMADDR:
	    if(0x0000000f & (int)((float*)mmaddr + lbound)){
		fprintf(stderr,"Error: %s with transfer from lbound %i starting at",fn, lbound);
		fprintf(stderr,"0x%x, not 16 byte aligned.\n",(unsigned int)((float*)mmaddr + lbound));
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_LSADDR:
	    if(0x0000000f & (int)((float*)lsaddr + lbound)){
		fprintf(stderr,"Error: %s with transfer from lbound %i starting at",fn, lbound);
		fprintf(stderr,"0x%x, not 16 byte aligned.\n",(unsigned int)((float*)lsaddr + lbound));
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	}
	break;
    case DVAR:
	switch(test){
	case TEST_SIZE_OFFSET:
	    if((ubound-lbound + 1)*sizeof(double) % 16 != 0){
		fprintf(stderr,"Error: %s with size %lu not a multiple of 16 byte\n",
		       fn, (ubound-lbound + 1)*sizeof(double)); // the +1 comes from transfer(0,1) = size2
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_SIZE:
	    if((ubound-lbound + 1)*sizeof(double) > 16384){
		fprintf(stderr,"Error: %s with size %lu larger than 16KB (16384B)\n",
		       fn, (ubound-lbound + 1)*sizeof(double)); // the +1 comes from transfer(0,1) = size2
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_MMADDR:
	    if(0x0000000f & (int)((double*)mmaddr + lbound)){
		fprintf(stderr,"Error: %s with transfer from lbound %i starting at",fn, lbound);
		fprintf(stderr,"0x%x, not 16 byte aligned.\n",(unsigned int)((double*)mmaddr + lbound));
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	case TEST_LSADDR:
	    if(0x0000000f & (int)((double*)lsaddr + lbound)){
		fprintf(stderr,"Error: %s with transfer from lbound %i starting at",fn, lbound);
		fprintf(stderr,"0x%x, not 16 byte aligned.\n",(unsigned int)((double*)lsaddr + lbound));
		spu_write_out_mbox(NESTSTEP_SPU_FATAL_ERROR); 
		exit(1);
	    }
	    break;
	}
	break;
    }
}

//set tag
void set_tag(Dma_tag* tag, int tagnumber, Addr64 address){
    int i;
#ifdef DEBUG_TAG
    printf("set_tag: %i, %x\n",tagnumber,address.p);
#endif
    for(i=0;i<32;i++){
	if(!tag->tag[i]){
#ifdef DEBUG_TAG
	    printf("number: %i\n",i);
#endif
	    tag->tag[i] = tagnumber;
	    tag->addr[i] = address;
	    return;
	}
    }
    printf("Error, Too many tags in variable, try waiting for completion..\n");
}
//remove tag
void remove_tag(Dma_tag* tag, Addr64 address){
    int i;
    for(i=0;i<32;i++){
	if(tag->addr[i].p == address.p){
#ifdef DEBUG_TAG
	    printf("remove_tag: %i, %x\n",tag->tag[i],address.p);
#endif
	    tag->tag[i] = 0;
	    tag->addr[i] = (Addr64)NULL;
	    return;
	}
    }
}
//find tag
int find_tag(Dma_tag* tag, Addr64 address){
    int i;
    for(i=0;i<32;i++){
	if(tag->addr[i].p == address.p){
#ifdef DEBUG_TAG
	    printf("find_tag: %i, %x\n",tag->tag[i],address.p);
#endif
	    return tag->tag[i];
	}
    }
    return 0;
}

Dma_tag* new_tag(){
    Dma_tag* tmp_tag = malloc(sizeof(Dma_tag));
    int i=0;
    for(;i<32;i++){
	tmp_tag->tag[i]=0;
	tmp_tag->addr[i]=(Addr64)NULL;
    }
    return tmp_tag;
}
void free_tag(Dma_tag* tmp){
    free(tmp);
}
