#include "combine_spu.h"

//This function is called after a step-statement, it should be called
//with both it's parameters set to NULL. the variables are there if
//we want to run combine on another group
void NestStep_combine(CombineList *_clist, ArrayHolder *_holder){
    int data;    
    //send COMBINE to ppu
    spu_write_out_mbox(COMBINE);
    //wait COMBINE from ppu
    while(TRUE){
	//	printf("combine\n");
	data = spu_read_in_mbox();
	if(data == NO_DATA){
	    continue;
	}
	if(data == COMBINE){
	    break;
	}
    }
    return;
}

//This functions attaches a shared variable to the combine-list that is used
//to get a consistent state after a step-statement
//NOTE:
//_prefix_tag = PREFIX if prefix-summation should occur
//_prefix_base has to reside in main-memory, i.e. it has to be inside
//a PrivateVar.
//void NestStep_ShVar_attach(ShVar *_var, 
//                           int _combine_strategy,
//                           char* _combineFnName, //void (*_combineFn)(void*, void*, void*),
//                           int _prefix_tag,
//                           Addr64 _prefix_base,
//                           int _updated,
//                           CombineList* _clist){
void NestStep_ShVar_attach(ShVar *_var, 
                           int _combine_strategy,
                           char* _combineFnName, //void (*_combineFn)(void*, void*, void*),
                           int _prefix_tag,
                           PVar* _prefix_var,
                           int _updated,
                           CombineList* _clist){
    int data = NO_DATA;
    //put data in mbox
#ifdef DEBUG_SPU_ATTACH
    printf("SPU:name_var: %i, %i\n",_var->name.procedure, _var->name.relative);
#endif //DEBUG_SPU
    namecpy(&(dataManager.mboxExtraData.name), &(_var->name));
    if(_prefix_var)
	namecpy(&(dataManager.mboxExtraData.prefixname), &(_prefix_var->name));

    dataManager.mboxExtraData.type = SHVAR;
    dataManager.mboxExtraData.datatype = _combine_strategy;
    //NOTE: how do i handle combineFn:s ? i dont.

    dataManager.mboxExtraData.size = _prefix_tag;
    //    dataManager.mboxExtraData.addr = _prefix_base;
    
    dataManager.mboxExtraData.blocksize = _updated;

    write_mboxExtraData();
    //send ATTACH_VARIABLE to ppu
    spu_write_out_mbox(ATTACH_VARIABLE);
    //wait for ATTACH_VARIABLE from ppu
    
    while(TRUE){
	data = spu_read_in_mbox();
	if(data == NO_DATA){
	    continue;
	}
	if(data == ATTACH_VARIABLE){
	    break;
	}

    }


    return;
}
//This function attaches a replicated array to the combine-list that is used
//to get a consistent state after a step-statement, inside the RepArray a
//lower and higher bound should be set to indicate what parts of the array
//should be combined
//NOTE:
//_prefix_tag = PREFIX if prefix-summation should occur
//_prefix_base has to reside in main-memory, i.e. it has to be inside
//a PrivateArray.
//void NestStep_RepArray_attach(RepArray *_array,
//                              int _combine_strategy,
//                              char* _combineFnName, //void (*_combineFn)(void*, void*, void*),
//                              int _prefix_tag,
//                              Addr64 _prefix_base,
//                              int _updated,
//                              CombineList *_clist){
void NestStep_RepArray_attach(RepArray *_array,
                              int _combine_strategy,
                              char* _combineFnName, //void (*_combineFn)(void*, void*, void*),
                              int _prefix_tag,
                              PrivateArray* _prefix_array,
                              int _updated,
                              CombineList *_clist){
    int data = NO_DATA;
    //put data in mbox
    namecpy(&(dataManager.mboxExtraData.name), &(_array->name));
    if(_prefix_array)
	namecpy(&(dataManager.mboxExtraData.prefixname), &(_prefix_array->name));
    dataManager.mboxExtraData.type = RARR;
    dataManager.mboxExtraData.datatype =_combine_strategy;
    //NOTE: how do i handle combineFn:s ? i dont.

    dataManager.mboxExtraData.size = _prefix_tag;
    //    dataManager.mboxExtraData.addr = _prefix_base;
    
    dataManager.mboxExtraData.blocksize = _updated;

    write_mboxExtraData();
    //send ATTACH_VARIABLE to ppu
    spu_write_out_mbox(ATTACH_VARIABLE);
    //wait for ATTACH_VARIABLE from ppu
    
    while(TRUE){
	data = spu_read_in_mbox();
	if(data == NO_DATA){
	    continue;
	}
	if(data == ATTACH_VARIABLE){
	    break;
	}

    }

    return;
}

// sets the range for which the combine-function should affect the
// array, will stay in effect until after the next combine
int NestStep_RepArray_range(RepArray *_array, int _lower, int _higher){
    int data = NO_DATA;
    namecpy(&(dataManager.mboxExtraData.name), &(_array->name));
    dataManager.mboxExtraData.size = _lower;
    dataManager.mboxExtraData.blocksize = _higher;
    write_mboxExtraData();

    //send ATTACH_VARIABLE to ppu
    spu_write_out_mbox(REPARRAY_RANGE);
    //wait for ATTACH_VARIABLE from ppu
    
    while(TRUE){
	data = spu_read_in_mbox();
	if(data == NO_DATA){
	    continue;
	}
	if(data == REPARRAY_RANGE){
	    break;
	}

    }

    return 0;
}
