/* ************************************************************** *
 * NestStep library for the IBM CELL. This is the SPU-part,       *
 * there also exist an PPU-part. The library is structured        *
 * to run controls such as managing memory and what SPU should    *
 * do what on the PPU and all calculations should reside in the   *
 * step-statements, which should run on the SPU, please read the  *
 * intstructions on http://www.ida.liu.se/chrke/NestStep/ for     *
 * more help on how to write NestStep programs.                   *
 * ************************************************************** */


#include "neststep_spu.h"

//TODO: this file needs to be cleaned up

//(void) mfc_put(volatile void *ls, uint64_t ea, uint32_t size, uint32_t tag,
//     uint32_t tid, uint32_t rid)

//(void) mfc_get(volatile void *ls, uint64_t ea, uint32_t size, uint32_t tag,
//     uint32_t tid, uint32_t rid)
//set tid, rid = 0 and tag to something unique..
//NestStep init initializes the datastructures so userspace programs
//can access created variables and Neststep finalize clears up allocated memory

void debug_datastructures(){
    int* i;
    printf("\n\n########################################\n");
    printf("SPU debug-section for datastructures\n\n");
    printf("Pointer size: %li\n",sizeof(i));
    printf("DataManager size: %li (16*%li = %li)\n", 
	   sizeof(DataManager), 
	   sizeof(DataManager)/16, 
	   (sizeof(DataManager)/16)*16);
    printf("DataManager offset: %lx\n\n", (long unsigned int)&dataManager);
    printf("Group size: %li (16*%li = %li)\n", 
	   sizeof(Group), 
	   sizeof(Group)/16, 
	   (sizeof(Group)/16)*16);
    printf("Group pos: %lx\n\n", (long unsigned int)&dataManager.thisgroup);
    printf("########################################\n");
    return;
}

void NestStep_SPU_init(Addr64 argp){
    //    printf("sizeof ll: %i\n", sizeof(long long*));
#ifdef DEBUG_SPU
    printf("spu_init: %llx\n", argp.ull);
    //    debug_datastructures();
    char debugthreadname[20];
    sprintf(debugthreadname,"%llx",argp.ull);
    debugfile = fopen(debugthreadname,"wa");
#endif //DEBUG_SPU
    dataManagerAddress = argp;
    //DMA in the dataManager from dataManagerAddress
    //    printf("before get\n");
    mfc_get(&dataManager,dataManagerAddress.ull,sizeof(DataManager),
    	    NESTSTEP_DMA_TAG,0,0);
    //    mfc_get(&dataManager,dataManagerAddress.p,sizeof(DataManager),
    //	    NESTSTEP_DMA_TAG,0,0);
    //    printf(p"Size of DataManager: %i\n", sizeof(DataManager));
    //wait for the DMA to complete
    //    printf("before mask\n");
    mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
    //    printf("before status\n");
    mfc_read_tag_status_all();
    //    printf("after status\n");
    //dataManager.thisgroup = NULL;
    //return;
    //    printf("SPU group address: %x\n",dataManager.thisgroup);
    //DMA in the group, shvar, pvar, parr, carr, barr, rarr
    //    Group* tmp = (Group*)malloc_align(sizeof(Group),16);
    Group* tmp = (Group*)malloc(sizeof(Group));
    //    printf("SPU group_addr: %llx\n",dataManager.thisgroup);
    mfc_get(tmp,(unsigned long long*)(dataManager.thisgroup),sizeof(Group),
	    NESTSTEP_DMA_TAG,0,0);
    //wait for the DMA to complete
    mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
    mfc_read_tag_status_all();
    //    printf("fetched the group\n");
    dataManager.thisgroup = tmp;
    //    printf("SPE: fetched group address is on: %x\n",dataManager.thisgroup);
    //    printf("SPU init complete: %llx\n", argp.ull);

    // we also need to import all shVars, pVars, pArrays, cArrays,
    // bArrays and repArrays
    import_shVar();
    import_pVar();
    import_pArray();
    import_cArray();
    import_bArray();
    import_repArray();
#ifdef DEBUG_SPU
    printf("SPU init complete: %llx\n", argp.p);
#endif //DEBUG_SPU
    return;
}
void NestStep_SPU_finalize(void){
    int data;
    //free memory allocated in the dataManager
#ifdef DEBUG_SPU
    printf("SPU finalize: %llx\n",dataManagerAddress.p);
    //printf("SPU finalize: %llx\n",dataManagerAddress.ull);
#endif //DEBUG_SPU
    // tell the ppu to destroy this thread
    //    int i;
//    for(i=0;i<1000;i++){
//	if(spu_stat_in_mbox()){
//	    printf("THE MBOX SHOULD BE EMPTY FFS!!!\n");
//	    spu_write_out_mbox(DESTROY_THREAD); 
//	    exit(1);
//	}
//    }
    spu_write_out_mbox(DESTROY_THREAD); 

    // wait for the ppu to be ready for handling the removal of this thread.
//    data = spu_read_in_mbox();
//    while(data != DESTROY_THREAD){
//	spu_read_in_mbox();
//    }
    data = spu_read_in_mbox();
#ifdef DEBUG_SPU
    printf("waiting for DESTROY_THREAD, got %i\n",data);
#endif //DEBUG_SPU
    if(data != DESTROY_THREAD){
	printf("SPU Error, didn't get DESTROY_THREAD in NestStep_SPU_finalize(), id:%llx\n",
	       dataManagerAddress.ull);
	printf("data: %i\n",data);
	exit(1);
    }

    if(spu_stat_in_mbox()){
	data = spu_read_in_mbox();
	printf("SPU got extra %i message\n",data);
    }
    //spu_write_out_mbox(NO_DATA); 
#ifdef DEBUG_SPU
    //NOTE: close the filedescriptor
    fclose(debugfile);
#endif //DEBUG_SPU

    //wait for all transfers to finnish
    mfc_write_tag_mask(0xffffffff); // wait for all 32 channels to complete
    mfc_read_tag_status_all();
    if(errno){
	printf("errno: %i\n",errno);
    }
    //    data = spu_read_in_mbox();

    return;
}

// NestStep begin step, it is only allowed to update shared variables inside
// a step-statement, this also sets the scope of the variables declared
// this is stored inside thisgroup->path
void NestStep_step(){
    int i;
    for(i=0;dataManager.thisgroup->path[i] != -1; i++)
	;
    dataManager.thisgroup->path[i] = NestStep_get_group_id();
    dataManager.thisgroup->path[i + 1] = -1;

    return;
}
// NestStep end step, reduces the thisgroup->path so that it is on the
// correct level.
void NestStep_end_step(){
    int i;
    for(i=0;dataManager.thisgroup->path[i] != -1; i++)
	;
    dataManager.thisgroup->path[i - 1] = -1;

    return;
}

//retreives the rank of the SPU in the current group
int NestStep_get_rank(void){
    return dataManager.thisgroup->groupInfo.my_rank;
}

//retreives the group size of the current group
int NestStep_get_size(void){
    return dataManager.thisgroup->groupInfo.size;
}

//get the group number
int NestStep_get_group_id(void){
    return dataManager.thisgroup->groupInfo.my_group_number;
}

//return a pointer to the parent group
Group* NestStep_get_parent_group(void){
    return dataManager.thisgroup->parent;
}

void import_shVar(){
    ShVar *addr, *data, *last;
    addr =  dataManager.shVars;
    last = NULL;
    dataManager.shVars = last;
    while(addr != NULL){
	data = malloc(sizeof(ShVar));
/* 	printf("import_shVar from: %llx, to: %llx\n", */
/* 	       (unsigned long long)addr, */
/* 	       (unsigned long long)data */
/* 	       ); */
	    
	mfc_get(data, addr, sizeof(ShVar), NESTSTEP_DMA_TAG,0,0);
	
	mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
	//	printf("import_ShVar complete\n");

	data->tag = new_tag();
	//insert the variable into the dataManager
	if(dataManager.shVars == NULL){
	    dataManager.shVars = data;
	} else {
	    last->next = data;
	    data->prev = last;
	    last = data;
	}
       
	addr = data->next;

    }
    return;
}
void import_pVar(){
    PVar *addr, *data, *last;
    addr =  dataManager.privateVars;
    last = NULL;
    dataManager.privateVars = last;
    while(addr != NULL){
	data = malloc(sizeof(PVar));
/* 	printf("import_pVar from: %llx, to: %llx\n", */
/* 	       (unsigned long long)addr, */
/* 	       (unsigned long long)data */
/* 	       ); */
   
	mfc_get(data,addr, sizeof(PVar), NESTSTEP_DMA_TAG,0,0);
	
	mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
	//	printf("import_pVar complete\n");
	data->tag = new_tag();
	//insert the variable into the dataManager
	if(dataManager.privateVars == NULL){
	    dataManager.privateVars = data;
	} else {
	    last->next = data;
	    data->prev = last;
	    last = data;
	}
       
	addr = data->next;

    }
    return;
}
void import_pArray(){
    PrivateArray *addr, *data, *last;
    addr =  dataManager.privateArrays;
    last = NULL;
    dataManager.privateArrays = last;
    while(addr != NULL){
	data = malloc(sizeof(PrivateArray));
/* 	printf("import_pArray from: %llx, to: %llx\n", */
/* 	       (unsigned long long)addr, */
/* 	       (unsigned long long)data */
/* 	       ); */
   
	mfc_get(data,addr, sizeof(PrivateArray), NESTSTEP_DMA_TAG,0,0);
	
	mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
	//	printf("import_pArray complete\n");
	data->tag = new_tag();
	//insert the variable into the dataManager
	if(dataManager.privateArrays == NULL){
	    dataManager.privateArrays = data;
	} else {
	    last->next = data;
	    data->prev = last;
	    last = data;
	}
       
	addr = data->next;

    }
    return;
}
void import_bArray(){
    BlockDistArray *addr, *data, *last;
    addr =  dataManager.blockDistArrays;
    last = NULL;
    dataManager.blockDistArrays = last;
    while(addr != NULL){
	data = malloc(sizeof(BlockDistArray));
/* 	printf("import_bArray from: %llx, to: %llx\n", */
/* 	       (unsigned long long)addr, */
/* 	       (unsigned long long)data */
/* 	       ); */
   
	mfc_get(data,addr, sizeof(BlockDistArray), NESTSTEP_DMA_TAG,0,0);
	
	mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
	//	printf("import_bArray complete\n");
	data->tag = new_tag();
	//insert the variable into the dataManager
	if(dataManager.blockDistArrays == NULL){
	    dataManager.blockDistArrays = data;
	} else {
	    last->next = data;
	    data->prev = last;
	    last = data;
	}
       
	addr = data->next;

    }
    return;
}
void import_cArray(){
    CyclicDistArray *addr, *data, *last;
    addr =  dataManager.cyclicDistArrays;
    last = NULL;
    dataManager.cyclicDistArrays = last;
    while(addr != NULL){
	data = malloc(sizeof(CyclicDistArray));
/* 	printf("import_cArray from: %llx, to: %llx\n", */
/* 	       (unsigned long long)addr, */
/* 	       (unsigned long long)data */
/* 	       ); */
   
	mfc_get(data,addr, sizeof(CyclicDistArray), NESTSTEP_DMA_TAG,0,0);
	
	mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
	//	printf("import_cArray complete\n");
	data->tag = new_tag();
	//insert the variable into the dataManager
	if(dataManager.cyclicDistArrays == NULL){
	    dataManager.cyclicDistArrays = data;
	} else {
	    last->next = data;
	    data->prev = last;
	    last = data;
	}
       
	addr = data->next;

    }
    return;
}
void import_repArray(){
    RepArray *addr, *data, *last;
    addr =  dataManager.repArrays;
    last = NULL;
    dataManager.repArrays = last;
    while(addr != NULL){
	data = malloc(sizeof(RepArray));
/* 	printf("import_repArray from: %llx, to: %llx\n", */
/* 	       (unsigned long long)addr, */
/* 	       (unsigned long long)data */
/* 	       ); */
   
	mfc_get(data,addr, sizeof(RepArray), NESTSTEP_DMA_TAG,0,0);
	
	mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
	//	printf("import_repArray complete\n");
	data->tag = new_tag();
	//insert the variable into the dataManager
	if(dataManager.repArrays == NULL){
	    dataManager.repArrays = data;
	} else {
	    last->next = data;
	    data->prev = last;
	    last = data;
	}
       
	addr = data->next;

    }
    return;
}

//create a new private variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace, _no
//is the variable number in case of duplicates (i.e. arrays)
//TODO: base is deprecated, send NULL.. remove from here?
PVar* new_PVar(Name _name, 
               int _type, 
               Addr64 _base){
    //printf("SPU new PVar create data\n");
    //    return NULL;
    int data;
    PVar* tmp_pvar;
    dataManager.mboxExtraData.type = PVAR;
    dataManager.mboxExtraData.datatype = _type;
    namecpy(&(dataManager.mboxExtraData.name),&_name);
    //    printf("SPU new pvar transfer data\n");

    write_mboxExtraData();
#ifdef DEBUG_SPU
    printf("send CREATE_VARIABLE to ppu\n");
#endif //DEBUG_SPU
    spu_write_out_mbox(CREATE_VARIABLE); //non blocking
    //create our own version of the variable while we wait for the ppu to complete
    //this will do almost the same as the ppu-version but it wont allocate memory
    //    tmp_pvar = malloc_align(sizeof(PVar),16);
    tmp_pvar = (PVar*)malloc(sizeof(PVar)); // no memory alignment is
                                            // necessary here.. i think
    tmp_pvar->tag = new_tag();
    namecpy(&(tmp_pvar->name),&_name);
    tmp_pvar->type = _type;
    tmp_pvar->base.p = NULL;
    while(TRUE){
	data = spu_read_in_mbox(); // this is a blocking operation
	//	printf(".");
	//i guess checking for -1 isn't necessary since spu_read_in_box is blocking
	if(data == NO_DATA){ 
	    continue;
	}
	if(data == CREATE_VARIABLE){
	    //	    printf("SPU got PVar CREATE_VARIABLE, store address and return\n");
	    // ppu has created the variable, get the pointer to the
	    // content of the variable
	    read_mboxExtraData();
	    tmp_pvar->base = dataManager.mboxExtraData.addr;
	    break;
	}
	//TODO: throw exception? what if something other than CREATE_VARIABLE
	//gets passed here..?

    }
    //    printf("SPU PVar creation complete\n");
    //add the PVar to the datamanager and return it
    add_PVar(tmp_pvar, &dataManager);
    //spu_write_out_mbox(NO_DATA); 
    return tmp_pvar;
}

void free_PVar(PVar* var){
    //    return;
    int data;
    namecpy(&(dataManager.mboxExtraData.name),&(var->name));
    dataManager.mboxExtraData.type = PVAR;
    write_mboxExtraData();
    spu_write_out_mbox(REMOVE_VARIABLE); // non-blocking
    while(TRUE){
	data = spu_read_in_mbox();
	if(data == NO_DATA){
	    continue;
	}
	if(data == REMOVE_VARIABLE){
	    //	    printf("SPU got REMOVE_VARIABLE\n");
	    break;
	}
    }
    free_tag(var->tag);
    //remove from dataManager
    remove_PVar(var, &dataManager);
    //remove 
    //    free_align(var);
    free(var);
    //spu_write_out_mbox(NO_DATA); 
    return;
}

//create a new shared variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace, _no
//is the variable number in case of duplicates (i.e. arrays)

ShVar* new_ShVar(Name _name, 
                 int _type, 
                 Addr64 _base){
#ifdef DEBUG_SPU
    printf("SPU new ShVar create data\n");
#endif //DEBUG_SPU
    //    return NULL;
    int data;
    ShVar* tmp_shvar;
    dataManager.mboxExtraData.type = SHVAR;
    dataManager.mboxExtraData.datatype = _type;
    namecpy(&(dataManager.mboxExtraData.name),&_name);
    //    printf("SPU new ShVar transfer data\n");


    write_mboxExtraData();
#ifdef DEBUG_SPU
    printf("send CREATE_VARIABLE to ppu\n");
#endif //DEBUG_SPU
    spu_write_out_mbox(CREATE_VARIABLE); //non blocking
    //create our own version of the variable while we wait for the ppu to complete
    //this will do almost the same as the ppu-version but it wont allocate memory
    //    tmp_shvar = (ShVar*)malloc_align(sizeof(ShVar),16);
    tmp_shvar = (ShVar*)malloc(sizeof(ShVar)); // no memory alignment is
				     // necessary here.. SPU is 16byte aligned
    tmp_shvar->tag = new_tag();
    namecpy(&(tmp_shvar->name),&_name);
    tmp_shvar->type = _type;
    tmp_shvar->base.p = NULL;
    while(TRUE){
#ifdef DEBUG_SPU
	printf("wait for CREATE_VARIABLE from ppu\n");
#endif //DEBUG_SPU
	//	printf("new_ShVar before read_in_mbox, waiting for CREATE_VARIABLE\n");
	//	fprintf(debugfile,"stat_in_mbox before read, got: %i\n",spu_stat_in_mbox());
	data = spu_read_in_mbox(); // this is a blocking operation
#ifdef DEBUG_SPU
	printf("stat_in_mbox after read, got: %i\n",spu_stat_in_mbox());
	printf("wanted CREATE_VARIABLE(1), got: %i\n",data);
#endif //DEBUG_SPU
	//	printf("new_ShVar after read_in_mbox\n");
	//	printf(".");
	//i guess checking for -1 isn't necessary since spu_read_in_box is blocking

	if(data == NO_DATA){ 
	    printf("ERROR: data = -1\n");
	    continue;
	}
	if(data == CREATE_VARIABLE){
	    //      	    printf("SPU got ShVar CREATE_VARIABLE, store address and return\n");
	    // ppu has created the variable, get the pointer to the
	    // content of the variable
	    read_mboxExtraData();
	    tmp_shvar->base = dataManager.mboxExtraData.addr;
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	    exit(1);
	}

    }
    //    printf("SPU ShVar creation complete\n");
    return tmp_shvar;
}
void free_ShVar(ShVar* var){
    //    return;
    int data;
    namecpy(&(dataManager.mboxExtraData.name),&(var->name));
    dataManager.mboxExtraData.type = SHVAR;
    write_mboxExtraData();
    spu_write_out_mbox(REMOVE_VARIABLE); // non-blocking
    while(TRUE){
	data = spu_read_in_mbox(); //blocking
	if(data == NO_DATA){
	    continue;
	}
	if(data == REMOVE_VARIABLE){
	    //	    printf("SPU got REMOVE_VARIABLE\n");
	    break;
	}
    }
    free_tag(var->tag);
    //remove from dataManager
    remove_ShVar(var, &dataManager);
    //remove 
    //    free_align(var);
    free(var);
    //spu_write_out_mbox(NO_DATA); 
    return;
}

void NestStep_get_all_ls(Ls_pointers *ls)
{
    spu_write_out_mbox(GET_ALL_LS_ADDR); //non blocking
	int data;
    while(TRUE){
	//	printf("new_PrivateArray before read_in_mbox, waiting for CREATE_VARIABLE\n");
	data = spu_read_in_mbox(); // this is a blocking operation
#ifdef DEBUG_SPU
	printf("stat_in_mbox after read, got: %d\n",data);
#endif //DEBUG_SPU
	//	printf("new_PrivateArray after read_in_mbox\n");
	//	printf(".");
	//i guess checking for -1 isn't necessary since spu_read_in_box is blocking
	if(data == NO_DATA){ 
	    printf("ERROR: data = -1\n");
	    continue;
	}
	if(data == GET_ALL_LS_ADDR){
	    // ppu has created the variable, get the pointer to the
	    // content of the variable
	    read_mboxExtraData();
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	    exit(1);
	}
    }
//	printf("data: %d, pointer: 0x%x\n",data, dataManager.mboxExtraData.addr.p);
//	read_mboxExtraData();
//	mfc_get(ls, (Addr64)(dataManager.mboxExtraData.addr), sizeof(unsigned int)*MAX_SPE_THREADS, NESTSTEP_DMA_TAG,0,0);
	mfc_get(ls, (unsigned long)dataManager.mboxExtraData.addr.p, sizeof(Ls_pointers)*MAX_SPE_THREADS, NESTSTEP_DMA_TAG,0,0);
	
	mfc_write_tag_mask(1 << NESTSTEP_DMA_TAG);
	mfc_read_tag_status_all();
}

//create a new private array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
//NOTE: base is deprecated, right? send NULL
PrivateArray* new_PrivateArray(Name _name, 
                               int _type, 
                               int _size, 
                               Addr64 _base){

    int data;
    PrivateArray* tmp_parr;
    //data to send to the ppus version of the datamanager
    dataManager.mboxExtraData.type = PARR;
    dataManager.mboxExtraData.datatype = _type;
    dataManager.mboxExtraData.size = _size;

    namecpy(&(dataManager.mboxExtraData.name),&_name);
    
    //    printf("SPU new PrivateArray transfer data\n");


    write_mboxExtraData();
#ifdef DEBUG_SPU
    printf("send CREATE_VARIABLE to ppu\n");
#endif //DEBUG_SPU

    spu_write_out_mbox(CREATE_VARIABLE); //non blocking
    //create our own version of the variable while we wait for the ppu to complete
    //this will do almost the same as the ppu-version but it wont allocate memory
    //    tmp_parr = (PrivateArray*)malloc_align(sizeof(PrivateArray),16);
    tmp_parr = (PrivateArray*)malloc(sizeof(PrivateArray)); // no memory alignment is
				     // necessary here.. SPU is 16byte aligned
    tmp_parr->tag = new_tag();
    //data to keep in this datamanager
    namecpy(&(tmp_parr->name),&_name);
    tmp_parr->type = _type;
    tmp_parr->base.p = NULL;
    tmp_parr->size = _size;
    while(TRUE){
	//	printf("new_PrivateArray before read_in_mbox, waiting for CREATE_VARIABLE\n");
	data = spu_read_in_mbox(); // this is a blocking operation
#ifdef DEBUG_SPU
	printf("stat_in_mbox after read, got: %i\n",spu_stat_in_mbox());
#endif //DEBUG_SPU
	//	printf("new_PrivateArray after read_in_mbox\n");
	//	printf(".");
	//i guess checking for -1 isn't necessary since spu_read_in_box is blocking
	if(data == NO_DATA){ 
	    printf("ERROR: data = -1\n");
	    continue;
	}
	if(data == CREATE_VARIABLE){
	    //      	    printf("SPU got PrivateArray CREATE_VARIABLE, store address and return\n");
	    // ppu has created the variable, get the pointer to the
	    // content of the variable
	    read_mboxExtraData();
	    tmp_parr->base = dataManager.mboxExtraData.addr;
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	    exit(1);
	}
    }
    //    printf("SPU PrivateArray creation complete\n");
    //spu_write_out_mbox(NO_DATA); 
    return tmp_parr;
    //    return NULL;
}
void free_PrivateArray(PrivateArray* var){
    int data;
    namecpy(&(dataManager.mboxExtraData.name),&(var->name));
    dataManager.mboxExtraData.type = PARR;

    write_mboxExtraData();
    //    printf("#\n");
    spu_write_out_mbox(REMOVE_VARIABLE); // non-blocking
    while(TRUE){
	data = spu_read_in_mbox(); //blocking
	if(data == NO_DATA){
	    continue;
	}
	if(data == REMOVE_VARIABLE){
	    //	    printf("SPU got REMOVE_VARIABLE\n");
	    break;
	}
    }
    //    printf(".\n");
    free_tag(var->tag);
    //remove from dataManager
    //    printf("..\n");
    remove_PrivateArray(var, &dataManager);
    //remove 
    //    free_align(var);
    //    printf("...\n");
    free(var);
    //    printf("....\n");
    //spu_write_out_mbox(NO_DATA); 
    return;
}

//create a new replicated array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
//NOTE: base is deprecated, right? send NULL
RepArray* new_RepArray(Name _name, 
                       int _type, 
                       int _size, 
                       Addr64 _base){
    int data;
    RepArray* tmp_rarr;
    //Data that will be sent to the ppu-version of the datamanager
    dataManager.mboxExtraData.type = RARR;
    dataManager.mboxExtraData.datatype = _type;
    dataManager.mboxExtraData.size = _size;
    namecpy(&(dataManager.mboxExtraData.name),&_name);
    //    printf("SPU new RepArray transfer data\n");


    write_mboxExtraData();
#ifdef DEBUG_SPU
    printf("send CREATE_VARIABLE to ppu\n");
#endif //DEBUG_SPU
    spu_write_out_mbox(CREATE_VARIABLE); //non blocking
    //create our own version of the variable while we wait for the ppu to complete
    //this will do almost the same as the ppu-version but it wont allocate memory
    //    tmp_rarr = (RepArray*)malloc_align(sizeof(RepArray),16);
    tmp_rarr = (RepArray*)malloc(sizeof(RepArray)); // no memory alignment is
				     // necessary here.. SPU is 16byte aligned
    tmp_rarr->tag = new_tag();
    namecpy(&(tmp_rarr->name),&_name);
    tmp_rarr->type = _type;
    tmp_rarr->base.p = NULL;
    tmp_rarr->size = _size;
    while(TRUE){
	//	printf("new_RepArray before read_in_mbox, waiting for CREATE_VARIABLE\n");
	data = spu_read_in_mbox(); // this is a blocking operation
#ifdef DEBUG_SPU
	printf("stat_in_mbox after read, got: %i\n",spu_stat_in_mbox());
#endif //DEBUG_SPU
	//	printf("new_RepArray after read_in_mbox\n");
	//	printf(".");
	//i guess checking for -1 isn't necessary since spu_read_in_box is blocking
	if(data == NO_DATA){ 
	    printf("ERROR: data = -1\n");
	    continue;
	}
	if(data == CREATE_VARIABLE){
	    //      	    printf("SPU got RepArray CREATE_VARIABLE, store address and return\n");
	    // ppu has created the variable, get the pointer to the
	    // content of the variable
	    read_mboxExtraData();
	    tmp_rarr->base = dataManager.mboxExtraData.addr;
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	    exit(1);
	}
	//TODO: throw exception? what if something other than CREATE_VARIABLE
	//gets passed here..?

    }
    //    printf("SPU RepArray creation complete\n");
    return tmp_rarr;
}

void free_RepArray(RepArray* var){
    int data;
    namecpy(&(dataManager.mboxExtraData.name),&(var->name));
    dataManager.mboxExtraData.type = RARR;
    write_mboxExtraData();
    spu_write_out_mbox(REMOVE_VARIABLE); // non-blocking
    while(TRUE){
	data = spu_read_in_mbox(); //blocking
	if(data == NO_DATA){
	    continue;
	}
	if(data == REMOVE_VARIABLE){
	    //	    printf("SPU got REMOVE_VARIABLE\n");
	    break;
	}
    }
    free_tag(var->tag);
    //remove from dataManager
    remove_RepArray(var, &dataManager);
    //remove 
    //    free_align(var);
    free(var);
    //spu_write_out_mbox(NO_DATA); 
    return;
}

//create a new block array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
BlockDistArray* new_BlockDistArray(Name _name, 
                                   int _type, 
                                   int _size){
    int data;
    BlockDistArray* tmp_barr;
    // data that will be sent to the ppu-version of the datamanager
    dataManager.mboxExtraData.type = BARR;
    dataManager.mboxExtraData.datatype = _type;
    dataManager.mboxExtraData.size = _size;
    namecpy(&(dataManager.mboxExtraData.name),&_name);
    //    printf("SPU new BlockDistArray transfer data\n");


    write_mboxExtraData();
#ifdef DEBUG_SPU
    printf("send CREATE_VARIABLE to ppu\n");
#endif //DEBUG_SPU
    spu_write_out_mbox(CREATE_VARIABLE); //non blocking
    //create our own version of the variable while we wait for the ppu to complete
    //this will do almost the same as the ppu-version but it wont allocate memory
    //    tmp_barr = (BlockDistArray*)malloc_align(sizeof(BlockDistArray),16);
    tmp_barr = (BlockDistArray*)malloc(sizeof(BlockDistArray)); // no memory alignment is
				     // necessary here.. SPU is 16byte aligned
    tmp_barr->tag = new_tag();
    //data is also stored on the spu-part of the data-manager
    namecpy(&(tmp_barr->name),&_name);
    tmp_barr->type = _type;
    tmp_barr->base.p = NULL;
	//tmp_barr->base.ull = NULL;
    tmp_barr->size = _size;

    while(TRUE){
	//	printf("new_BlockDistArray before read_in_mbox, waiting for CREATE_VARIABLE\n");
	data = spu_read_in_mbox(); // this is a blocking operation
#ifdef DEBUG_SPU
	printf("stat_in_mbox after read, got: %i\n",spu_stat_in_mbox());
#endif //DEBUG_SPU
	//	printf("new_BlockDistArray after read_in_mbox\n");
	//	printf(".");
	//i guess checking for -1 isn't necessary since spu_read_in_box is blocking
	if(data == NO_DATA){ 
	    printf("ERROR: data = -1\n");
	    continue;
	}
	if(data == CREATE_VARIABLE){
	    //      	    printf("SPU got BlockDistArray CREATE_VARIABLE, store address and return\n");
	    // ppu has created the variable, get the pointer to the
	    // content of the variable
	    read_mboxExtraData();
	    tmp_barr->base = dataManager.mboxExtraData.addr;
	    tmp_barr->lower = dataManager.mboxExtraData.lower;
	    tmp_barr->higher = dataManager.mboxExtraData.higher;
	    tmp_barr->localsize = dataManager.mboxExtraData.blocksize;
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	    exit(1);
	}
	//TODO: throw exception? what if something other than CREATE_VARIABLE
	//gets passed here..?

    }
    //    printf("SPU BlockDistArray creation complete\n");
    //spu_write_out_mbox(NO_DATA); 
    return tmp_barr;

    //    return NULL;
}

void free_BlockDistArray(BlockDistArray* var){
    int data;
    namecpy(&(dataManager.mboxExtraData.name),&(var->name));
    dataManager.mboxExtraData.type = BARR;
    write_mboxExtraData();
    spu_write_out_mbox(REMOVE_VARIABLE); // non-blocking
    while(TRUE){
	data = spu_read_in_mbox(); //blocking
	if(data == NO_DATA){
	    continue;
	}
	if(data == REMOVE_VARIABLE){
	    //	    printf("SPU got REMOVE_VARIABLE\n");
	    break;
	}
    }
    free_tag(var->tag);
    //remove from dataManager
    remove_BlockDistArray(var, &dataManager);
    //remove 
    //    free_align(var);
    free(var);
    //spu_write_out_mbox(NO_DATA); 
    return;
}

//create a new cyclic array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, size is the total size of the array, blocksize
//is the size of each segment that is chosen modulo the number of threads
//allocated, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
CyclicDistArray* new_CyclicDistArray(Name _name, 
                                     int _type, 
                                     int _size, 
                                     int _blocksize){
    int data;
    CyclicDistArray* tmp_carr;
    dataManager.mboxExtraData.type = CARR;
    dataManager.mboxExtraData.datatype = _type;
    dataManager.mboxExtraData.size = _size;
    dataManager.mboxExtraData.blocksize = _blocksize;
    namecpy(&(dataManager.mboxExtraData.name),&_name);
    //    printf("SPU new CyclicDistArray transfer data\n");


    write_mboxExtraData();
#ifdef DEBUG_SPU
    printf("send CREATE_VARIABLE to ppu\n");
#endif //DEBUG_SPU
    spu_write_out_mbox(CREATE_VARIABLE); //non blocking
    //create our own version of the variable while we wait for the ppu to complete
    //this will do almost the same as the ppu-version but it wont allocate memory
    //    tmp_carr = (CyclicDistArray*)malloc_align(sizeof(CyclicDistArray),16);
    tmp_carr = (CyclicDistArray*)malloc(sizeof(CyclicDistArray)); // no memory alignment is
				     // necessary here.. SPU is 16byte aligned
    tmp_carr->tag = new_tag();
    namecpy(&(tmp_carr->name),&_name);
    tmp_carr->type = _type;
    tmp_carr->base.p = NULL;
    //    tmp_carr->base.ull = NULL;
    tmp_carr->blocksize = _blocksize;
    tmp_carr->size = _size;
    while(TRUE){
	//	printf("new_CyclicDistArray before read_in_mbox, waiting for CREATE_VARIABLE\n");
	data = spu_read_in_mbox(); // this is a blocking operation
#ifdef DEBUG_SPU
	printf("stat_in_mbox after read, got: %i\n",spu_stat_in_mbox());
#endif // DEBUG_SPU

	if(data == CREATE_VARIABLE){
	    // ppu has created the variable, get the pointer to the
	    // content of the variable
	    read_mboxExtraData();
	    tmp_carr->base = dataManager.mboxExtraData.addr;
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	    //exit(1);
	}
	//NOTE: throw exception? what if something other than CREATE_VARIABLE
	//gets passed here..? this can happen, in that case we should throw it away
	//and wait for the real value to come.. i don't know why this is happening
	//could be some bug with the simulator..

    }
    return tmp_carr;


}

void free_CyclicDistArray(CyclicDistArray* var){
    int data;
    namecpy(&(dataManager.mboxExtraData.name),&(var->name));
    dataManager.mboxExtraData.type = CARR;
    write_mboxExtraData();
    spu_write_out_mbox(REMOVE_VARIABLE); // non-blocking
    while(TRUE){
	data = spu_read_in_mbox(); //blocking
	if(data == NO_DATA){
	    continue;
	}
	if(data == REMOVE_VARIABLE){
	    //	    printf("SPU got REMOVE_VARIABLE\n");
	    break;
	}
    }
    free_tag(var->tag);
    //remove from dataManager
    remove_CyclicDistArray(var, &dataManager);
    //remove 
    //    free_align(var);
    free(var);
    //spu_write_out_mbox(NO_DATA); 
    return;
}

/* ************************************************************** *
 * ShVar, PVar, PArr, RArr, CArr, BArr getters, setters           *
 * and barriers. The getters and setters have separate functions  *
 * for storing/retreiving integers, floats and doubles. these are *
 * named get_***I, store_***I and so on with F and D as suffixes  *
 * respectively.                                                  *
 * ************************************************************** */

//get/set a shared variable from/to the DataManager
void get_ShVarI(ShVar* _shvar,Addr64 lsaddr){
    //    lsaddr.p = malloc(sizeof(int));
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("addr of tmp:%lx\n",tmp);
#endif // DEBUG_SPU_VAR_TRANSFER
    int tag;
    //get avaiable tag
    tag = get_available_mfc_chan();
    //    printf("tag: %i\n",tag);
    //store tag in variable
    set_tag(_shvar->tag, tag,lsaddr);
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("get_ShVarI transfer from: %lx, to: %lx\n",(long unsigned int)_shvar->base.p,(long unsigned int)tmp);
    printf("sizeof(int): %i\n",sizeof(int));
#endif // DEBUG_SPU_VAR_TRANSFER
    mfc_get(lsaddr.p,(unsigned long)_shvar->base.p, sizeof(int), tag,0,0);

    //    printf("return from get_ShVarI\n");
}
void get_ShVarF(ShVar* _shvar,Addr64 lsaddr){
    //    lsaddr.p = malloc(sizeof(float));
    int tag;
    //get avaiable tag
    tag = get_available_mfc_chan();
    //store tag in variable
    set_tag(_shvar->tag, tag,lsaddr);
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("get_ShVarF transfer from: %lx, to: %lx\n",(long unsigned int)_shvar->base.p,(long unsigned int)tmp);
#endif // DEBUG_SPU_VAR_TRANSFER
    mfc_get(lsaddr.p,(unsigned long)_shvar->base.p,
	    sizeof(float), tag,0,0);
}
void get_ShVarD(ShVar* _shvar,Addr64 lsaddr){
    //    lsaddr.p = malloc(sizeof(double));;
    int tag;
    //get avaiable tag
    tag = get_available_mfc_chan();
    //store tag in variable
    set_tag(_shvar->tag, tag,lsaddr);
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("get_ShVarD transfer from: %lx, to: %lx\n",(long unsigned int)_shvar->base.p,(long unsigned int)tmp);
#endif // DEBUG_SPU_VAR_TRANSFER
    mfc_get(lsaddr.p,(unsigned long)_shvar->base.p,
	    sizeof(double), tag,0,0);
}
//_var will leave scope here.. not good. will tag be copied?
void store_ShVarI(ShVar* _shvar,Addr64 lsaddr){
    int tag;
    check_dma("store_ShVarI",0,0,lsaddr.p,NULL,IVAR,TEST_LSADDR);

    //get available tag
    tag = get_available_mfc_chan();
    //    printf("store_ShVarI, to: %lx, from: %lx\n",(long unsigned int)_shvar->base.p,(long unsigned int)_var);
    set_tag(_shvar->tag, tag,lsaddr);
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("store_ShVarI from addr: %lx, to: %lx\n",_var,_shvar->base.p);
#endif // DEBUG_SPU_VAR_TRANSFER
    mfc_put(lsaddr.p,(unsigned long)_shvar->base.p, 
	    sizeof(int), tag, 0,0);
    return;
}
void store_ShVarF(ShVar* _shvar,Addr64 lsaddr){
    int tag;
    check_dma("store_ShVarF",0,0,lsaddr.p,NULL,FVAR,TEST_LSADDR);

    //get available tag
    tag = get_available_mfc_chan();
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("store_ShVarF, to: %lx, from: %lx\n",(long unsigned int)_shvar->base.p,(long unsigned int)lsaddr.p);
#endif // DEBUG_SPU_VAR_TRANSFER
    set_tag(_shvar->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)_shvar->base.p, 
	    sizeof(float), tag, 0,0);
    return;
}
void store_ShVarD(ShVar* _shvar,Addr64 lsaddr){
    int tag;
    check_dma("store_ShVarD",0,0,lsaddr.p,NULL,DVAR,TEST_LSADDR);
    //get available tag
    tag = get_available_mfc_chan();
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("store_ShVarD, to: %lx, from: %lx\n",(long unsigned int)_shvar->base.p,(long unsigned int)lsaddr.p);
#endif // DEBUG_SPU_VAR_TRANSFER
    set_tag(_shvar->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)_shvar->base.p, 
	    sizeof(double), tag, 0,0);
    return;
}


//get/set a private variable from/to the DataManager
void get_PVarI(PVar* _pvar,Addr64 lsaddr){
    //    lsaddr.p = malloc(sizeof(int));
    int tag;
    //get avaiable tag
    tag = get_available_mfc_chan();
    //store tag in variable
    set_tag(_pvar->tag, tag,lsaddr);
/*     mfc_get(tmp,_pvar->base.ull, */
/* 	    sizeof(int), tag,0,0); */
    mfc_get(lsaddr.p,(unsigned long)_pvar->base.p,
	    sizeof(int), tag,0,0);
}
void get_PVarF(PVar* _pvar,Addr64 lsaddr){
    //    lsaddr.p = malloc(sizeof(float));
    int tag;
    //get avaiable tag
    tag = get_available_mfc_chan();
    //store tag in variable
    set_tag(_pvar->tag, tag,lsaddr);
    mfc_get(lsaddr.p,(unsigned long)_pvar->base.p,
	    sizeof(float), tag,0,0);
}
void get_PVarD(PVar* _pvar,Addr64 lsaddr){
    //    lsaddr.p = malloc(sizeof(double));;
    int tag;
    //get avaiable tag
    tag = get_available_mfc_chan();
    //store tag in variable
    set_tag(_pvar->tag, tag,lsaddr);
    mfc_get(lsaddr.p,(unsigned long)_pvar->base.p,
	    sizeof(double), tag,0,0);
}
void store_PVarI(PVar* _pvar,Addr64 lsaddr){
    int tag;
    check_dma("store_PVarI",0,0,lsaddr.p,NULL,IVAR,TEST_LSADDR);
    //get available tag
    tag = get_available_mfc_chan();
    set_tag(_pvar->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)_pvar->base.p, 
	    sizeof(int), tag, 0,0);
    return;
}
void store_PVarF(PVar* _pvar,Addr64 lsaddr){
    int tag;
    check_dma("store_PVarF",0,0,lsaddr.p,NULL,FVAR,TEST_LSADDR);

    //get available tag
    tag = get_available_mfc_chan();
    set_tag(_pvar->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)_pvar->base.p, 
	    sizeof(float), tag, 0,0);

    return;
}
void store_PVarD(PVar* _pvar,Addr64 lsaddr){
    int tag;
    check_dma("store_PVarD",0,0,lsaddr.p,NULL,DVAR,TEST_LSADDR);
    //get available tag
    tag = get_available_mfc_chan();
    set_tag(_pvar->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)_pvar->base.p, 
	    sizeof(double), tag, 0,0);
    return;
}


//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from a private array with the userspace-name name from/to the DataManager
void get_PArrI(PrivateArray* _parr,Addr64 lsaddr, int lbound, int ubound){
    if(ubound-lbound){
	check_dma("get_PArrI",lbound,ubound,NULL,_parr->base.p,IVAR,TEST_SIZE_OFFSET);
	check_dma("get_PArrI",lbound,ubound,NULL,_parr->base.p,IVAR,TEST_SIZE);
	check_dma("get_PArrI",lbound,ubound,NULL,_parr->base.p,IVAR,TEST_MMADDR);
    }
    //    lsaddr.p = malloc((ubound-lbound)*sizeof(int));
    //NOTE: start DMA-transfer from _parr to tmp
    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    int* addr = (int*)_parr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_parr->tag, tag,lsaddr);
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void get_PArrF(PrivateArray* _parr,Addr64 lsaddr, int lbound, int ubound){
    if(ubound-lbound){
	check_dma("get_PArrF",lbound,ubound,NULL,_parr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("get_PArrF",lbound,ubound,NULL,_parr->base.p,FVAR,TEST_SIZE);
	check_dma("get_PArrF",lbound,ubound,NULL,_parr->base.p,FVAR,TEST_MMADDR);
    }
    //   lsaddr.p = malloc((ubound-lbound+1)*sizeof(float));
    //    printf("malloc: %x, end: %x\n",tmp,tmp+(ubound-lbound+1));
    //NOTE: start DMA-transfer from _parr to tmp
    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    float* addr = (float*)_parr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_parr->tag, tag,lsaddr);
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void get_PArrD(PrivateArray* _parr,Addr64 lsaddr, int lbound, int ubound){
    if(ubound-lbound){
	check_dma("get_PArrD",lbound,ubound,NULL,_parr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("get_PArrD",lbound,ubound,NULL,_parr->base.p,DVAR,TEST_SIZE);
	check_dma("get_PArrD",lbound,ubound,NULL,_parr->base.p,DVAR,TEST_MMADDR);
    }
    //    lsaddr.p = malloc((ubound-lbound)*sizeof(double));
    //NOTE: start DMA-transfer from _parr to tmp
    int size = (ubound-lbound + 1)*sizeof(double); // the +1 comes from transfer(0,1) = size2
    double* addr = (double*)_parr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_parr->tag, tag,lsaddr);
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}

void store_PArrI(PrivateArray* _parr, Addr64 lsaddr,int lbound,int ubound){
    if(ubound-lbound){
	check_dma("store_PArrI",lbound,ubound,lsaddr.p,_parr->base.p,IVAR,TEST_SIZE_OFFSET);
	check_dma("store_PArrI",lbound,ubound,lsaddr.p,_parr->base.p,IVAR,TEST_SIZE);
	check_dma("store_PArrI",lbound,ubound,lsaddr.p,_parr->base.p,IVAR,TEST_MMADDR);
	check_dma("store_PArrI",0,0,lsaddr.p,_parr->base.p,IVAR,TEST_LSADDR);
    }
    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    int* addr = (int*)_parr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_parr->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)addr, size, tag,0,0);
    return;
}
void store_PArrF(PrivateArray* _parr, Addr64 lsaddr,int lbound,int ubound){
    //    printf("base: %x\n",_parr->base.p);
    if(ubound-lbound){
	check_dma("store_PArrF",lbound,ubound,lsaddr.p,_parr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("store_PArrF",lbound,ubound,lsaddr.p,_parr->base.p,FVAR,TEST_SIZE);
	check_dma("store_PArrF",lbound,ubound,lsaddr.p,_parr->base.p,FVAR,TEST_MMADDR);
	check_dma("store_PArrF",0,0,lsaddr.p,_parr->base.p,FVAR,TEST_LSADDR);
    }
    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    float* addr = (float*)_parr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_parr->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)addr, size, tag,0,0);
    return;
}
void store_PArrD(PrivateArray* _parr, Addr64 lsaddr,int lbound,int ubound){
    if(ubound-lbound){
	check_dma("store_PArrD",lbound,ubound,lsaddr.p,_parr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("store_PArrD",lbound,ubound,lsaddr.p,_parr->base.p,DVAR,TEST_SIZE);
	check_dma("store_PArrD",lbound,ubound,lsaddr.p,_parr->base.p,DVAR,TEST_MMADDR);
	check_dma("store_PArrD",0,0,lsaddr.p,_parr->base.p,DVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(double); // the +1 comes from transfer(0,1) = size2
    double* addr = (double*)_parr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_parr->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)addr, size, tag,0,0);
    return;
}


//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from the segment that this node owns from a cyclic array with the 
//userspace-name name from/to the DataManager
//NOTE: the request is with local addresses
//NOTE: remember to ask for aligned and properly sized data!
//NOTE: when doing the request, remember that it is called with global indexes, and it is assumed that
//      the range will fit inside one blocksize
void get_CArrI(CyclicDistArray* _carr,Addr64 lsaddr, int lbound, int ubound){
    int local_lbound = global_local_index_c(_carr, lbound);
    int local_ubound = global_local_index_c(_carr, ubound);

    check_dma("get_CArrI",local_lbound,local_ubound,NULL,_carr->base.p,IVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("get_CArrI",local_lbound,local_ubound,NULL,_carr->base.p,IVAR,TEST_SIZE_OFFSET);
	check_dma("get_CArrI",local_lbound,local_ubound,NULL,_carr->base.p,IVAR,TEST_SIZE);
	check_dma("get_CArrI",local_lbound,local_ubound,NULL,_carr->base.p,IVAR,TEST_MMADDR);
    }
    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    //    int* tmp = (int*)malloc(size);
    int* addr = (int*)_carr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_carr->tag, tag,lsaddr);
    //NOTE: start DMA-transfer from _carr to tmp
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void get_CArrF(CyclicDistArray* _carr,Addr64 lsaddr, int lbound, int ubound){
    int local_lbound = global_local_index_c(_carr, lbound);
    int local_ubound = global_local_index_c(_carr, ubound);

    check_dma("get_CArrF",local_lbound,local_ubound,NULL,_carr->base.p,FVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("get_CArrF",local_lbound,local_ubound,NULL,_carr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("get_CArrF",local_lbound,local_ubound,NULL,_carr->base.p,FVAR,TEST_SIZE);
	check_dma("get_CArrF",local_lbound,local_ubound,NULL,_carr->base.p,FVAR,TEST_MMADDR);
    }
    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    //    float* tmp = (float*)malloc(size);
    float* addr = (float*)_carr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_carr->tag, tag,lsaddr);
    //NOTE: start DMA-transfer from _carr to tmp
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void get_CArrD(CyclicDistArray* _carr,Addr64 lsaddr, int lbound, int ubound){
    int local_lbound = global_local_index_c(_carr, lbound);
    int local_ubound = global_local_index_c(_carr, ubound);
    check_dma("get_CArrD",local_lbound,local_ubound,NULL,_carr->base.p,DVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("get_CArrD",local_lbound,local_ubound,NULL,_carr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("get_CArrD",local_lbound,local_ubound,NULL,_carr->base.p,DVAR,TEST_SIZE);
	check_dma("get_CArrD",local_lbound,local_ubound,NULL,_carr->base.p,DVAR,TEST_MMADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(double);// the +1 comes from transfer(0,1) = size2
    //    double* tmp = (double*)malloc(size);
    double* addr = (double*)_carr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_carr->tag, tag,lsaddr);
    //NOTE: start DMA-transfer from _carr to tmp
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void store_CArrI(CyclicDistArray* _carr, Addr64 lsaddr,int lbound,int ubound){
    int local_lbound = global_local_index_c(_carr, lbound);
    int local_ubound = global_local_index_c(_carr, ubound);
    check_dma("store_CArrI",local_lbound,local_ubound,lsaddr.p,_carr->base.p,IVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("store_CArrI",local_lbound,local_ubound,lsaddr.p,_carr->base.p,IVAR,TEST_SIZE_OFFSET);
	check_dma("store_CArrI",local_lbound,local_ubound,lsaddr.p,_carr->base.p,IVAR,TEST_SIZE);
	check_dma("store_CArrI",local_lbound,local_ubound,lsaddr.p,_carr->base.p,IVAR,TEST_MMADDR);
	check_dma("store_CArrI",0,0,lsaddr.p,_carr->base.p,IVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    int* addr = (int*)_carr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_carr->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)addr, size, tag,0,0);
    return;
}
void store_CArrF(CyclicDistArray* _carr, Addr64 lsaddr,int lbound,int ubound){
    int local_lbound = global_local_index_c(_carr, lbound);
    int local_ubound = global_local_index_c(_carr, ubound);
    check_dma("store_CArrF",local_lbound,local_ubound,lsaddr.p,_carr->base.p,FVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("store_CArrF",local_lbound,local_ubound,lsaddr.p,_carr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("store_CArrF",local_lbound,local_ubound,lsaddr.p,_carr->base.p,FVAR,TEST_SIZE);
	check_dma("store_CArrF",local_lbound,local_ubound,lsaddr.p,_carr->base.p,FVAR,TEST_MMADDR);
	check_dma("store_CArrF",0,0,lsaddr.p,_carr->base.p,FVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    float* addr = (float*)_carr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_carr->tag, tag,lsaddr);
    mfc_put(lsaddr.p,(unsigned long)addr, size, tag,0,0);
    return;
}
void store_CArrD(CyclicDistArray* _carr, Addr64 lsaddr,int lbound,int ubound){
    int local_lbound = global_local_index_c(_carr, lbound);
    int local_ubound = global_local_index_c(_carr, ubound);
    check_dma("store_CArrD",local_lbound,local_ubound,lsaddr.p,_carr->base.p,DVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("store_CArrD",local_lbound,local_ubound,lsaddr.p,_carr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("store_CArrD",local_lbound,local_ubound,lsaddr.p,_carr->base.p,DVAR,TEST_SIZE);
	check_dma("store_CArrD",local_lbound,local_ubound,lsaddr.p,_carr->base.p,DVAR,TEST_MMADDR);
	check_dma("store_CArrD",0,0,lsaddr.p,_carr->base.p,DVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(double); // the +1 comes from transfer(0,1) = size2
    double* addr = (double*)_carr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_carr->tag, tag,lsaddr);
    //    printf("lsaddr: %x, addr: %x, size: %i\n",(unsigned long)lsaddr.p,(unsigned long)addr,size);
    mfc_put(lsaddr.p,(unsigned long)addr, size, tag,0,0);
    return;
}

//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from the segment that this node owns from a block array with the 
//userspace-name name from/to the DataManager
//NOTE: these load/store data in global-address-space
void get_BArrI(BlockDistArray* _barr,Addr64 lsaddr, int lbound, int ubound){
    int local_lbound = global_local_index_b(_barr, lbound);
    int local_ubound = global_local_index_b(_barr, ubound);
//    printf("lower: %i, higher: %i, local_lower: %i, local_higher: %i\n",lbound,ubound,
//	   local_lbound,local_ubound);
    check_dma("get_BArrI",local_lbound,local_ubound,NULL,_barr->base.p,IVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("get_BArrI",local_lbound,local_ubound,NULL,_barr->base.p,IVAR,TEST_SIZE_OFFSET);
	check_dma("get_BArrI",local_lbound,local_ubound,NULL,_barr->base.p,IVAR,TEST_SIZE);
	check_dma("get_BArrI",local_lbound,local_ubound,NULL,_barr->base.p,IVAR,TEST_MMADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    //    int* tmp = (int*)malloc(size);
    int* addr = (int*)_barr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_barr->tag, tag,lsaddr);
    //NOTE: start DMA-transfer from _barr to tmp
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void get_BArrF(BlockDistArray* _barr,Addr64 lsaddr, int lbound, int ubound){
    int local_lbound = global_local_index_b(_barr, lbound);
    int local_ubound = global_local_index_b(_barr, ubound);
    check_dma("get_BArrF",local_lbound,local_ubound,NULL,_barr->base.p,FVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("get_BArrF",local_lbound,local_ubound,NULL,_barr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("get_BArrF",local_lbound,local_ubound,NULL,_barr->base.p,FVAR,TEST_SIZE);
	check_dma("get_BArrF",local_lbound,local_ubound,NULL,_barr->base.p,FVAR,TEST_MMADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    //    float* tmp = (float*)malloc(size);
    float* addr = (float*)_barr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_barr->tag, tag,lsaddr);
    //NOTE: start DMA-transfer from _barr to tmp
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void get_BArrD(BlockDistArray* _barr,Addr64 lsaddr, int lbound, int ubound){
    int local_lbound = global_local_index_b(_barr, lbound);
    int local_ubound = global_local_index_b(_barr, ubound);

    check_dma("get_BArrD",local_lbound,local_ubound,NULL,_barr->base.p,DVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("get_BArrD",local_lbound,local_ubound,NULL,_barr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("get_BArrD",local_lbound,local_ubound,NULL,_barr->base.p,DVAR,TEST_SIZE);
	check_dma("get_BArrD",local_lbound,local_ubound,NULL,_barr->base.p,DVAR,TEST_MMADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(double); // the +1 comes from transfer(0,1) = size2
    //    double* tmp = (double*)malloc(size);
    double* addr = (double*)_barr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_barr->tag, tag,lsaddr);
    //NOTE: start DMA-transfer from _barr to tmp
    mfc_get(lsaddr.p,(unsigned long)addr, size, tag,0,0);
}
void store_BArrI(BlockDistArray* _barr, Addr64 lsaddr,int lbound,int ubound){
    int local_lbound = global_local_index_b(_barr, lbound);
    int local_ubound = global_local_index_b(_barr, ubound);

    check_dma("store_BArrI",local_lbound,local_ubound,lsaddr.p,_barr->base.ull,IVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("store_BArrI",local_lbound,local_ubound,lsaddr.p,_barr->base.ull,IVAR,TEST_SIZE_OFFSET);
	check_dma("store_BArrI",local_lbound,local_ubound,lsaddr.p,_barr->base.ull,IVAR,TEST_SIZE);
	check_dma("store_BArrI",local_lbound,local_ubound,lsaddr.p,_barr->base.ull,IVAR,TEST_MMADDR);
	check_dma("store_BArrI",0,0,lsaddr.p,_barr->base.ull,IVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    int* addr = (int*)_barr->base.p + local_lbound;

    int tag = get_available_mfc_chan();
    set_tag(_barr->tag, tag,lsaddr);
    mfc_put(lsaddr.p, (unsigned long)addr, size, tag,0,0);
    return;
}
void store_BArrF(BlockDistArray* _barr, Addr64 lsaddr,int lbound,int ubound){
    int local_lbound = global_local_index_b(_barr, lbound);
    int local_ubound = global_local_index_b(_barr, ubound);

    check_dma("store_BArrF",local_lbound,local_ubound,lsaddr.p,_barr->base.p,FVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("store_BArrF",local_lbound,local_ubound,lsaddr.p,_barr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("store_BArrF",local_lbound,local_ubound,lsaddr.p,_barr->base.p,FVAR,TEST_SIZE);
	check_dma("store_BArrF",local_lbound,local_ubound,lsaddr.p,_barr->base.p,FVAR,TEST_MMADDR);
	check_dma("store_BArrF",0,0,lsaddr.p,_barr->base.p,FVAR,TEST_LSADDR);
    }
    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    float* addr = (float*)_barr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_barr->tag, tag,lsaddr);
    mfc_put(lsaddr.p, (unsigned long)addr, size, tag,0,0);
    return;
}
void store_BArrD(BlockDistArray* _barr, Addr64 lsaddr,int lbound,int ubound){
    int local_lbound = global_local_index_b(_barr, lbound);
    int local_ubound = global_local_index_b(_barr, ubound);

    check_dma("store_BArrD",local_lbound,local_ubound,lsaddr.p,_barr->base.p,DVAR,TEST_BOUNDS);
    if(ubound-lbound){
	check_dma("store_BArrD",local_lbound,local_ubound,lsaddr.p,_barr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("store_BArrD",local_lbound,local_ubound,lsaddr.p,_barr->base.p,DVAR,TEST_SIZE);
	check_dma("store_BArrD",local_lbound,local_ubound,lsaddr.p,_barr->base.p,DVAR,TEST_MMADDR);
	check_dma("store_BArrD",0,0,lsaddr.p,_barr->base.p,DVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(double); // the +1 comes from transfer(0,1) = size2
    double* addr = (double*)_barr->base.p + local_lbound;
    int tag = get_available_mfc_chan();
    set_tag(_barr->tag, tag,lsaddr);
    mfc_put(lsaddr.p, (unsigned long)addr, size, tag,0,0);
    return;
}


//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from a replicated array with the userspace-name name from/to the DataManager
void get_RArrI(RepArray* _rarr,Addr64 lsaddr, int lbound, int ubound){
    if(ubound-lbound){
	check_dma("get_RArrI",lbound,ubound,NULL,_rarr->base.p,IVAR,TEST_SIZE_OFFSET);
	check_dma("get_RArrI",lbound,ubound,NULL,_rarr->base.p,IVAR,TEST_SIZE);
	check_dma("get_RArrI",lbound,ubound,NULL,_rarr->base.p,IVAR,TEST_MMADDR);
    }

    //    int* tmp = (int*)malloc((ubound-lbound)*sizeof(int));
    //NOTE: start DMA-transfer from _rarr to tmp
    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    int* addr = (int*)_rarr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_rarr->tag, tag,lsaddr);
    mfc_get(lsaddr.p, (unsigned long)addr, size, tag,0,0);
}
void get_RArrF(RepArray* _rarr,Addr64 lsaddr, int lbound, int ubound){
    if(ubound-lbound){
	check_dma("get_RArrF",lbound,ubound,NULL,_rarr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("get_RArrF",lbound,ubound,NULL,_rarr->base.p,FVAR,TEST_SIZE);
	check_dma("get_RArrF",lbound,ubound,NULL,_rarr->base.p,FVAR,TEST_MMADDR);
    }
    //    float* tmp = (float*)malloc((ubound-lbound)*sizeof(float));
    //NOTE: start DMA-transfer from _rarr to tmp
    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    float* addr = (float*)_rarr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_rarr->tag, tag,lsaddr);
    mfc_get(lsaddr.p, (unsigned long)addr, size, tag,0,0);
}
void get_RArrD(RepArray* _rarr,Addr64 lsaddr, int lbound, int ubound){
    if(ubound-lbound){
	check_dma("get_RArrD",lbound,ubound,NULL,_rarr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("get_RArrD",lbound,ubound,NULL,_rarr->base.p,DVAR,TEST_SIZE);
	check_dma("get_RArrD",lbound,ubound,NULL,_rarr->base.p,DVAR,TEST_MMADDR);
    }

    //    double* tmp = (double*)malloc((ubound-lbound)*sizeof(double));
    //NOTE: start DMA-transfer from _rarr to tmp
    int size = (ubound-lbound + 1)*sizeof(double); // the +1 comes from transfer(0,1) = size2
    double* addr = (double*)_rarr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_rarr->tag, tag,lsaddr);
    mfc_get(lsaddr.p, (unsigned long)addr, size, tag,0,0);
}
void store_RArrI(RepArray* _rarr, Addr64 lsaddr,int lbound,int ubound){
    if(ubound-lbound){
	check_dma("store_RArrI",lbound,ubound,lsaddr.p,_rarr->base.p,IVAR,TEST_SIZE_OFFSET);
	check_dma("store_RArrI",lbound,ubound,lsaddr.p,_rarr->base.p,IVAR,TEST_SIZE);
	check_dma("store_RArrI",lbound,ubound,lsaddr.p,_rarr->base.p,IVAR,TEST_MMADDR);
	check_dma("store_RArrI",0,0,lsaddr.p,_rarr->base.p,IVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(int); // the +1 comes from transfer(0,1) = size2
    int* addr = (int*)_rarr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_rarr->tag, tag,lsaddr);
    mfc_put(lsaddr.p, (unsigned long)addr, size, tag,0,0);
    return;
}
void store_RArrF(RepArray* _rarr, Addr64 lsaddr,int lbound,int ubound){
    if(ubound-lbound){
	check_dma("store_RArrF",lbound,ubound,lsaddr.p,_rarr->base.p,FVAR,TEST_SIZE_OFFSET);
	check_dma("store_RArrF",lbound,ubound,lsaddr.p,_rarr->base.p,FVAR,TEST_SIZE);
	check_dma("store_RArrF",lbound,ubound,lsaddr.p,_rarr->base.p,FVAR,TEST_MMADDR);
	check_dma("store_RArrF",0,0,lsaddr.p,_rarr->base.p,FVAR,TEST_LSADDR);
    }
    int size = (ubound-lbound + 1)*sizeof(float); // the +1 comes from transfer(0,1) = size2
    float* addr = (float*)_rarr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_rarr->tag, tag,lsaddr);
    mfc_put(lsaddr.p, (unsigned long)addr, size, tag,0,0);
    return;
}
void store_RArrD(RepArray* _rarr, Addr64 lsaddr,int lbound,int ubound){
    if(ubound-lbound){
	check_dma("store_RArrD",lbound,ubound,lsaddr.p,_rarr->base.p,DVAR,TEST_SIZE_OFFSET);
	check_dma("store_RArrD",lbound,ubound,lsaddr.p,_rarr->base.p,DVAR,TEST_SIZE);
	check_dma("store_RArrD",lbound,ubound,lsaddr.p,_rarr->base.p,DVAR,TEST_MMADDR);
	check_dma("store_RArrD",0,0,lsaddr.p,_rarr->base.p,DVAR,TEST_LSADDR);
    }

    int size = (ubound-lbound + 1)*sizeof(double); // the +1 comes from transfer(0,1) = size2
    double* addr = (double*)_rarr->base.p + lbound;
    int tag = get_available_mfc_chan();
    set_tag(_rarr->tag, tag,lsaddr);
    mfc_put(lsaddr.p, (unsigned long)addr, size, tag,0,0);
    return;
}

//Barriers for managing DMA transfers to/from the main memory
void wait_PVar_complete(PVar* _pvar, Addr64 lsaddr){
    mfc_write_tag_mask(1 << find_tag(_pvar->tag,lsaddr));
    mfc_read_tag_status_all();
    dataManager.tags[find_tag(_pvar->tag,lsaddr)] = 0;
    remove_tag(_pvar->tag,lsaddr);
    return;
}
void wait_ShVar_complete(ShVar* _shvar, Addr64 lsaddr){
#ifdef DEBUG_SPU_VAR_TRANSFER
    printf("wait_ShVar_complete\n");
#endif// DEBUG_SPU_VAR_TRANSFER
    mfc_write_tag_mask(1 << find_tag(_shvar->tag,lsaddr));
    mfc_read_tag_status_all();
    dataManager.tags[find_tag(_shvar->tag,lsaddr)] = 0;
    remove_tag(_shvar->tag,lsaddr);
    return;
}
void wait_BArr_complete(BlockDistArray* _barr, Addr64 lsaddr){
    int tag = find_tag(_barr->tag,lsaddr);
    mfc_write_tag_mask(1 << tag);
    mfc_read_tag_status_all();
    dataManager.tags[tag] = 0;
    remove_tag(_barr->tag,lsaddr);
    return;
}
void wait_CArr_complete(CyclicDistArray* _carr, Addr64 lsaddr){
    mfc_write_tag_mask(1 << find_tag(_carr->tag,lsaddr));
    mfc_read_tag_status_all();
    dataManager.tags[find_tag(_carr->tag,lsaddr)] = 0;
    remove_tag(_carr->tag,lsaddr);
    return;
}
void wait_PArr_complete(PrivateArray* _parr, Addr64 lsaddr){
    mfc_write_tag_mask(1 << find_tag(_parr->tag,lsaddr));
    mfc_read_tag_status_all();
    dataManager.tags[find_tag(_parr->tag,lsaddr)] = 0;
    remove_tag(_parr->tag,lsaddr);
    return;
}
void wait_RArr_complete(RepArray* _rarr, Addr64 lsaddr){
    mfc_write_tag_mask(1 << find_tag(_rarr->tag,lsaddr));
    mfc_read_tag_status_all();
    dataManager.tags[find_tag(_rarr->tag,lsaddr)] = 0;
    remove_tag(_rarr->tag,lsaddr);
    return;
}

//functions for managing arrays
//these functions set all elements in the arrays to zero.
//TODO: should i implement these here and create an array of zeroes
//and DMA it to main-memory or should I handle these on the ppu?
//NOTE: don't implement them..
void set_RepArray_zero(RepArray *array){
    return;
}
void set_BlockDistArray_zero(BlockDistArray *array){
    return;
}
void set_CyclicDistArray_zero(CyclicDistArray *array){
    return;
}
//TODO: scatter/gather from rep to block/cyclic

//NOTE: read_request_b, read_request_c i.e. mirror()
void read_request_b(PrivateArray* _parr, BlockDistArray* _barr, int lbound, int ubound){
    namecpy(&(dataManager.mboxExtraData.name),&(_barr->name));
    namecpy(&(dataManager.mboxExtraData.prefixname),&(_parr->name));
    dataManager.mboxExtraData.type = BARR;
    dataManager.mboxExtraData.lower = lbound;
    dataManager.mboxExtraData.higher = ubound;
	
	//printf("rq: dest:%lx\n", _parr->base);
	
    write_mboxExtraData();
    spu_write_out_mbox(MIRROR_VARIABLE); //non blocking
    int data = NO_DATA;
    while(TRUE){
	data = spu_read_in_mbox();
#ifdef DEBUG_SPU
	printf("read_request_b after read, got: %i\n",spu_stat_in_mbox());
#endif // DEBUG_SPU
	if(data == MIRROR_VARIABLE){
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	}
    }
}
void read_request_c(PrivateArray* _parr, CyclicDistArray* _carr, int lbound, int ubound){
    namecpy(&(dataManager.mboxExtraData.name),&(_carr->name));
    namecpy(&(dataManager.mboxExtraData.prefixname),&(_parr->name));
    dataManager.mboxExtraData.type = CARR;
    dataManager.mboxExtraData.lower = lbound;
    dataManager.mboxExtraData.higher = ubound;

    write_mboxExtraData();
    spu_write_out_mbox(MIRROR_VARIABLE); //non blocking
    int data = NO_DATA;
    while(TRUE){
	data = spu_read_in_mbox();
#ifdef DEBUG_SPU
	printf("read_request_c after read, got: %i\n",spu_stat_in_mbox());
#endif // DEBUG_SPU
	if(data == MIRROR_VARIABLE){
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	}
    }
}
//NOTE: write_request_b, write_request_c, i.e. update()
void write_request_b(PrivateArray* _parr, BlockDistArray* _barr, int lbound, int ubound){
    namecpy(&(dataManager.mboxExtraData.name),&(_barr->name));
    namecpy(&(dataManager.mboxExtraData.prefixname),&(_parr->name));
    dataManager.mboxExtraData.type = BARR;
    dataManager.mboxExtraData.lower = lbound;
    dataManager.mboxExtraData.higher = ubound;

    write_mboxExtraData();
    spu_write_out_mbox(UPDATE_VARIABLE); //non blocking
    int data = NO_DATA;
    while(TRUE){
	data = spu_read_in_mbox();
#ifdef DEBUG_SPU
	printf("write_request_b after read, got: %i\n",spu_stat_in_mbox());
#endif // DEBUG_SPU
	if(data == UPDATE_VARIABLE){
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	}
    }
}
void write_request_c(PrivateArray* _parr, CyclicDistArray* _carr, int lbound, int ubound){
    namecpy(&(dataManager.mboxExtraData.name),&(_carr->name));
    namecpy(&(dataManager.mboxExtraData.prefixname),&(_parr->name));
    dataManager.mboxExtraData.type = CARR;
    dataManager.mboxExtraData.lower = lbound;
    dataManager.mboxExtraData.higher = ubound;

    write_mboxExtraData();
    spu_write_out_mbox(UPDATE_VARIABLE); //non blocking
    int data = NO_DATA;
    while(TRUE){
	data = spu_read_in_mbox();
#ifdef DEBUG_SPU
	printf("write_request_c after read, got: %i\n",spu_stat_in_mbox());
#endif // DEBUG_SPU
	if(data == UPDATE_VARIABLE){
	    break;
	}else {
	    printf("ERROR: data=%i\n",data);
	}
    }
}


//these functions transform a local index to a global index and
//from global to local respectively.
//returns OUT_OF_BOUNDS if the index is higher than the segment
//and OUT_OF_BOUNDS if the index is lower than the segment
int global_local_index_b(BlockDistArray *array, int index){
    if(index < array->lower){
	return OUT_OF_BOUNDS;
    }
    if(index > array->higher){
	return OUT_OF_BOUNDS;
    }
    return index - array->lower;
}
int local_global_index_b(BlockDistArray *array, int index){
    if(index < 0){
	return OUT_OF_BOUNDS;
    }
    if(index > array->higher - array->lower){
	return OUT_OF_BOUNDS;
    }
    return index + array->lower;

}
int global_local_index_c(CyclicDistArray *array, int index){
    int remainder, ratio;

    if(index < 0){
	return OUT_OF_BOUNDS;
    }
    if(index > array->size){
	return OUT_OF_BOUNDS;
    }
    remainder = index % (array->blocksize * NestStep_get_size());
    ratio = index / (array->blocksize * NestStep_get_size());

    if(remainder >= (NestStep_get_rank() * array->blocksize) &&
       remainder < ((NestStep_get_rank() + 1) * array->blocksize)){

	return ratio * array->blocksize + 
	    remainder - 
	    NestStep_get_rank() * array->blocksize;
    }
    else {
	return OUT_OF_BOUNDS;
    }
}
int local_global_index_c(CyclicDistArray *array, int index){
    int gindex;
    
    //get which block it is in
    gindex = (index / array->blocksize)*array->blocksize*NestStep_get_size();
    //get the rank-offset within the block
    gindex += NestStep_get_rank() * array->blocksize;
    //get the offset within the rank
    gindex += index % array->blocksize;

    if(gindex < 0){
	return OUT_OF_BOUNDS;
    }
    if(gindex > array->size){
	return OUT_OF_BOUNDS;
    }
    //    printf("gindex: %i\n",gindex);
    return gindex;
}
