/* ********************************************** *
 * Dot product test-program, will create arrays   *
 * with increasing numbers and multiply them      *
 * ********************************************** */

#include <stdio.h>
#include <../../NestStep_spulib/neststep_spu.h>

//size:
//1024*1024
//size in memory: 4MiB
//16384 = 16*4KiB
//1048576*4/16384 = 256 chunks = 64 / spu if 4 spus
//ps3, scales to 4spu
//#define N 1048576*4
//cell, scales to 8spu
#define N 1048576*4*4

int MY_RANK;

// Used for transforming clock-ticks to
// seconds, gotten from /proc/cpuinfo
#define TIMEBASE 79800000
//#define TIMEBASE 14318000
// Initializing value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF

// Variables for handling timing.
int ublock1, ublock2, ublocktot;
int twait1, twait2, twaittot;
int ncomb1, ncomb2, ncombtot;
int timetmp1, timetmp2, timetmptot;

//Buffers for datatransfers
float buffer1[4096];
float buffer2[4096];
float buffer3[4096];
float buffer4[4096];
float* _dotP;
ShVar* dotP;
BlockDistArray* x;
BlockDistArray* y;
void calculate_dotproduct(){
    int i, j, no_chunks;
    i = j = 0;
    no_chunks = (N/4096)/NestStep_get_size();

    //initialize variables x and y with data 1 -> 1024*1024 using doublebuffering
    NestStep_step();
    {
	for(j=0;j<4096;j++){
	    buffer1[j] = j + MY_RANK*4096*no_chunks; // 64*4096
	}
	store_BArrF(x,(Addr64)buffer1,
		    MY_RANK*no_chunks*4096,
		    MY_RANK*no_chunks*4096+4096-1);
	store_BArrF(y,(Addr64)buffer1,
		    MY_RANK*no_chunks*4096,
		    MY_RANK*no_chunks*4096+4096-1);

	for (i=1; i<no_chunks; i++){
	    for(j=0;j<4096;j++){
		buffer2[j] = j + MY_RANK*no_chunks*4096+i*4096; // 64*4096
	    }
	    store_BArrF(x,(Addr64)buffer2,
			MY_RANK*no_chunks*4096+i*4096,
			MY_RANK*no_chunks*4096+i*4096+4096-1);
	    store_BArrF(y,(Addr64)buffer2,
			MY_RANK*no_chunks*4096+i*4096,
			MY_RANK*no_chunks*4096+i*4096+4096-1);

	    wait_BArr_complete(x,(Addr64)buffer1);
	    wait_BArr_complete(y,(Addr64)buffer1);
	    i++;
	    if(i >= no_chunks){
		wait_BArr_complete(x,(Addr64)buffer2);
		wait_BArr_complete(y,(Addr64)buffer2);
		break;
	    }

	    for(j=0;j<4096;j++){
		buffer1[j] = j + MY_RANK*no_chunks*4096 + i*4096; // 64*4096
	    }
	    store_BArrF(x,(Addr64)buffer1,
			MY_RANK*no_chunks*4096+i*4096,
			MY_RANK*no_chunks*4096+i*4096+4096-1);
	    store_BArrF(y,(Addr64)buffer1,
			MY_RANK*no_chunks*4096+i*4096,
			MY_RANK*no_chunks*4096+i*4096+4096-1);

	    wait_BArr_complete(x,(Addr64)buffer2);
	    wait_BArr_complete(y,(Addr64)buffer2);
	}
	wait_BArr_complete(x,(Addr64)buffer1);
	wait_BArr_complete(y,(Addr64)buffer1);
    }
    //We don't take creation of the datasets into calculation
    //because this is outside the algorithm beeing checked,
    //data can be generated or imported over network, from disk etc.
    //ncomb1 = spu_read_decrementer();
    NestStep_combine(NULL,NULL);
    //ncomb2 = spu_read_decrementer();
    //ncombtot+=ncomb1-ncomb2;
    NestStep_end_step();

    //do the multiplication and summation, combine to get dotproduct
    //use doublebuffering
    timetmp1 = spu_read_decrementer();
    NestStep_step();
    {
	get_BArrF(x,(Addr64)buffer1,MY_RANK*no_chunks*4096,
		  MY_RANK*no_chunks*4096+4096-1);
	get_BArrF(y,(Addr64)buffer2,MY_RANK*no_chunks*4096,
		  MY_RANK*no_chunks*4096+4096-1);
	*_dotP = 0;
	for(i=1;i<no_chunks;i++){
	    get_BArrF(x,(Addr64)buffer3,MY_RANK*no_chunks*4096+i*4096,
		      MY_RANK*no_chunks*4096+i*4096+4096-1);
	    get_BArrF(y,(Addr64)buffer4,MY_RANK*no_chunks*4096+i*4096,
		      MY_RANK*no_chunks*4096+i*4096+4096-1);
	    twait1 = spu_read_decrementer();
	    wait_BArr_complete(x,(Addr64)buffer1);
	    wait_BArr_complete(y,(Addr64)buffer2);
	    twait2 = spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    ublock1 = spu_read_decrementer();
	    for(j=0;j<4096;j++){
		*_dotP += buffer1[j]*buffer2[j];
	    }
	    ublock2 = spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;
	    i++;
	    if(i >= no_chunks){
		twait1 = spu_read_decrementer();
		wait_BArr_complete(x,(Addr64)buffer3);
		wait_BArr_complete(y,(Addr64)buffer4);
		twait2= spu_read_decrementer();
		twaittot+=twait1-twait2;
		ublock1 = spu_read_decrementer();
		for(j=0;j<4096;j++){
		    *_dotP += buffer3[j]*buffer4[j];
		}
		ublock2 = spu_read_decrementer();
		ublocktot+=ublock1-ublock2;
		break;
	    }
	    get_BArrF(x,(Addr64)buffer1,MY_RANK*no_chunks*4096+i*4096,
		      MY_RANK*no_chunks*4096+i*4096+4096-1);
	    get_BArrF(y,(Addr64)buffer2,MY_RANK*no_chunks*4096+i*4096,
		      MY_RANK*no_chunks*4096+i*4096+4096-1);
	    twait1 = spu_read_decrementer();
	    wait_BArr_complete(x,(Addr64)buffer3);
	    wait_BArr_complete(y,(Addr64)buffer4);
	    twait2 = spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    ublock1 = spu_read_decrementer();
	    for(j=0;j<4096;j++){
		*_dotP += buffer3[j]*buffer4[j];
	    }
	    ublock2 = spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;

	}

	store_ShVarF(dotP,(Addr64)_dotP);
	ublock1 = spu_read_decrementer();
	wait_ShVar_complete(dotP,(Addr64)_dotP);
	ublock2 = spu_read_decrementer();
	ublocktot+=ublock1-ublock2;
	//	printf("Partial dotP: %f\n",*_dotP);
	NestStep_ShVar_attach(dotP, ADD, NULL, NO_PREFIX,NULL, NOT_UPDATED, NULL);
    }
    ncomb1 = spu_read_decrementer();
    NestStep_combine(NULL,NULL);
    ncomb2 = spu_read_decrementer();
    ncombtot+=ncomb1-ncomb2;
    NestStep_end_step();
    
    timetmp2 = spu_read_decrementer();
    timetmptot+= timetmp1-timetmp2;
    if(MY_RANK == 0){
	get_ShVarF(dotP,(Addr64)_dotP);
	wait_ShVar_complete(dotP,(Addr64)_dotP);
	//	printf("Dot product: %lf\n",*_dotP);
    }
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
    //profiling counters
    int tmp_time1, tmp_time2, tot_time;
    ublock1 = ublock2 = ublocktot = 0;
    twait1 = twait2 = twaittot = 0;
    ncomb1 = ncomb2 = ncombtot = 0;
    timetmp1 = timetmp2 =timetmptot=0;
    spu_write_decrementer(DEC_INIT_VAL);
    tot_time = 0;

    int i;

    NestStep_SPU_init(argp);
    MY_RANK = NestStep_get_rank();
    Name name;
    name.procedure = 1;
    name.relative = 0;

    _dotP = malloc(sizeof(float));
    dotP = new_ShVar(name,FVAR,(Addr64)(void*)_dotP);

    name.relative++;
    x = new_BlockDistArray(name,FVAR,N);
    name.relative++;
    y = new_BlockDistArray(name,FVAR,N);

    for(i=0;i<20;i++){
	tmp_time1 = spu_read_decrementer();
	calculate_dotproduct();
	tmp_time2 = spu_read_decrementer();
	tot_time+= tmp_time1-tmp_time2;
    }
    if(MY_RANK == 0){
	printf("Tot programtime: %u, %f\n",tot_time, 
	       ((float)tot_time)/TIMEBASE);
	printf("Tot executiontime: %u, %f\n",timetmptot, 
	       ((float)timetmptot)/TIMEBASE);
	printf("Time spent calculating: %u, %f\n",ublocktot,
	       ((float)ublocktot)/TIMEBASE);
	printf("Time spent waiting: %u, %f\n",twaittot,
	       ((float)twaittot)/TIMEBASE);
	printf("Time spent in combine: %u, %f\n",ncombtot,
	       ((float)ncombtot)/TIMEBASE);
    }
    NestStep_SPU_finalize();
}
