/* ********************************************** *
 * Gauss Elimination test program. This program   *
 * will do a gaussian elimination on a Ax = b     *
 * equation system with A = NxN and b = N.        *
 * ********************************************** */

#include <stdio.h>
#include<math.h>
#include <../../NestStep_spulib/neststep_spu.h>

//#define N 1024

#define THRESHOLD 0.0001

#define TIMEBASE 79800000 // used for transforming clock-ticks to
			  // seconds, gotten from /proc/cpuinfo
#define DEC_INIT_VAL 0XFFFFFFFF

int ublock1, ublock2, ublocktot;
int twait1, twait2, twaittot;
int ncomb1, ncomb2, ncombtot;

int solve(CyclicDistArray* _A, CyclicDistArray* _b, int N){
    int i, j, k, localsize;
    int bufferA=0, bufferb=0;
    double b_elim = 0.0;
    double* elimbase;
    double* tmp_A[2];
    double* tmp_b[2];
    ShVar* _b_elim;
    RepArray* _elim;
    Name name;

    name.procedure = 1;
    name.relative = 0;
    localsize = N/NestStep_get_size();
    _b_elim = new_ShVar(name,DVAR,(Addr64)(void*)NULL);
    name.relative++;
    _elim = new_RepArray(name,DVAR,N,(Addr64)(void*)NULL);
    name.relative++;
    tmp_A[0] = malloc(N*sizeof(double));
    tmp_A[1] = malloc(N*sizeof(double));
    tmp_b[0] = malloc(sizeof(double)*localsize);
    tmp_b[1] = malloc(sizeof(double)*localsize);
    elimbase = malloc(sizeof(double)*N);
    memset(elimbase,0,N*sizeof(double));
    
    for(i=0;i<N;i++){
	NestStep_step();
	{
	    if(global_local_index_c(_A,i*N)!=-1){
		get_CArrD(_A,(Addr64)elimbase,i*N,(i+1)*N-1);
		get_CArrD(_b,(Addr64)&b_elim,i,i);
		wait_CArr_complete(_A,(Addr64)elimbase);
		wait_CArr_complete(_b,(Addr64)&b_elim);
		store_RArrD(_elim,(Addr64)elimbase,0,N-1);
		store_ShVarD(_b_elim,(Addr64)&b_elim);
		wait_RArr_complete(_elim,(Addr64)elimbase);
		wait_ShVar_complete(_b_elim,(Addr64)&b_elim);

		NestStep_RepArray_attach(_elim, UPDATED, NULL, NO_PREFIX, NULL, NestStep_get_rank(), NULL);
		NestStep_ShVar_attach(_b_elim, UPDATED, NULL, NO_PREFIX, NULL, NestStep_get_rank(), NULL);
	    } else {
		NestStep_RepArray_attach(_elim, UPDATED, NULL, NO_PREFIX, NULL, -1, NULL);
		NestStep_ShVar_attach(_b_elim, UPDATED, NULL, NO_PREFIX, NULL, -1, NULL);
	    }
	    

	    get_CArrD(_A,(Addr64)tmp_A[bufferA],(i+NestStep_get_rank()+1)*N,(i+NestStep_get_rank()+1+1)*N-1);
	    get_CArrD(_b,(Addr64)tmp_b[bufferb],i,i);



	    
	}
	NestStep_combine(NULL,NULL);
	NestStep_end_step();

	NestStep_step();
	{

	    wait_CArr_complete(_A,(Addr64)tmp_A[bufferA]);
	    wait_CArr_complete(_b,(Addr64)tmp_b[bufferb]);
	    
	    for(j=i+1;j<N;j++){
		get_CArrD(_A,(Addr64)tmp_A[1-bufferA],i*N,(i+1)*N-1);
		for(k=i+1;k<N;k++){

		}
		
		if(j == i+1){
		    for(j=i+1;j<N;k++){

		    }
		}

	    }


	}
	NestStep_combine(NULL,NULL);
	NestStep_end_step();
    }
}

int solve2(CyclicDistArray * _A, CyclicDistArray * _b, int N)
{
    int _tempc;
        
    int i, j, k, l;
    
    ShVar * _b_elim;
    double b_elim = 0.0;
    RepArray * elim;
    double * elimbase;
    Name name;
    
    double* tmp_A = malloc(N*sizeof(double));
    double* tmp_b = malloc(sizeof(double)*N/NestStep_get_size());

    elimbase = (double *)malloc(sizeof(double) * N);
    memset(elimbase, 0, N * sizeof(double));
    
    name.procedure = 134;
    name.relative = 0;
    elim = new_RepArray(name, DVAR, N, (Addr64)(void*)NULL);
    name.relative++;
    _b_elim = new_ShVar(name, DVAR, (Addr64)(void*)NULL);
    
    // move largest array to elim
    NestStep_step();
    if(global_local_index_c(_A, 0 * N) != -1){
	//k = 0 ?
	get_CArrD(_A,(Addr64)tmp_A,global_local_index_c(_A,0),global_local_index_c(_A,N-1));
	get_CArrD(_b,(Addr64)&(tmp_b[0]),0,0);
	wait_CArr_complete(_A,(Addr64)tmp_A);
	wait_CArr_complete(_b,(Addr64)&(tmp_b[0]));

	memcpy(elimbase,
	       tmp_A,
	       N * sizeof(double));
	//	b_elim = ((double *)(_b->base.p))[global_local_index_c(_b, k)];
	b_elim = tmp_b[0];
	
	store_RArrD(elim,(Addr64)elimbase,0,N-1);
	store_ShVarD(_b_elim,(Addr64)&b_elim);
	wait_RArr_complete(elim,(Addr64)elimbase);
	wait_ShVar_complete(_b_elim,(Addr64)&b_elim);

    }
//    NestStep_RepArray_attach(elim, UPDATED, NULL, NO_PREFIX, NULL, 0, NULL);
//    NestStep_ShVar_attach(_b_elim, UPDATED, NULL, NO_PREFIX, NULL, 0, NULL);
    NestStep_RepArray_attach(elim, ADD, NULL, NO_PREFIX, NULL, 0, NULL);
    NestStep_ShVar_attach(_b_elim, ADD, NULL, NO_PREFIX, NULL, 0, NULL);
    
    // combine
    NestStep_combine(NULL, NULL);
    NestStep_end_step();
    
    for(k = 0; k < (N - 1)/400; k++){
	NestStep_step();
	// eliminate using elim
	l = (k+1) / NestStep_get_size();
	_tempc = (k+1) % NestStep_get_size();
	l += ((_tempc > NestStep_get_rank()) ? 1 : 0);
	for(; local_global_index_c(_A, l * N) != -1; l++){

	    get_RArrD(elim,(Addr64)elimbase,0,N-1);
	    printf("foo: %i, %i\n",l*N,(l+1)*N-1);
	    get_CArrD(_A,(Addr64)tmp_A,(l*N),
		      ((l+1)*N-1));
	    printf("bar\n");
	    wait_RArr_complete(elim,(Addr64)elimbase);
	    wait_CArr_complete(_A,(Addr64)tmp_A);

	    double * elim_tmp = elimbase;
	    

//	    double * row_tmp = (double *)((char *)(_A->base.p) + l * N * sizeof(double));
	    double * row_tmp = tmp_A;
	    double modifier = -(row_tmp[k] / elim_tmp[k]);
	    ((double *)(_b->base.p))[l] += modifier * b_elim;
	    for(j = k; j < N; j++)
		row_tmp[j] += modifier * elim_tmp[j];

	    store_CArrD(_A,(Addr64)tmp_A,(l*N),
			((l+1)*N-1));
	    wait_CArr_complete(_A,(Addr64)tmp_A);
	}
	
	// move largest array to elim
	if(global_local_index_c(_A, (k+1) * N) != -1){

	    get_RArrD(elim,(Addr64)elimbase,k+1,N-(k+1)-1);
	    get_CArrD(_A,(Addr64)tmp_A,(k+1)*N+k+1,
		      (N-(k+1)-1));

	    wait_RArr_complete(elim,(Addr64)elimbase);
	    wait_CArr_complete(_A,(Addr64)tmp_A);


//	    memcpy((char *)(elim->base.p) + (k+1)*sizeof(double),
//		   (char *)(_A->base.p) + global_local_index_c(_A, (k+1) * N) * sizeof(double) + (k+1)*
//		   sizeof(double),
//		   (N - (k+1)) * sizeof(double));
	    memcpy(elimbase,
		   tmp_A,
		   (N - (k+1)) * sizeof(double));

	    //	    b_elim = ((double *)(_b->base.p))[global_local_index_c(_b, k+1)];
	    get_CArrD(_b,(Addr64)&(tmp_b[global_local_index_c(_b,k+1)]),k+1,k+1);
	    wait_CArr_complete(_b,(Addr64)&(tmp_b[global_local_index_c(_b,k+1)]));
	    b_elim = tmp_b[global_local_index_c(_b,k+1)];
	    store_ShVarD(_b_elim,(Addr64)&b_elim);
	    
	}
	NestStep_RepArray_range(elim, k, N - 1);
//	NestStep_RepArray_attach(elim, UPDATED, NULL, NO_PREFIX, NULL, 0, NULL);
//	NestStep_ShVar_attach(_b_elim, UPDATED, NULL, NO_PREFIX, NULL, 0, NULL);
	NestStep_RepArray_attach(elim, ADD, NULL, NO_PREFIX, NULL, 0, NULL);
	NestStep_ShVar_attach(_b_elim, ADD, NULL, NO_PREFIX, NULL, 0, NULL);
	// combine
	NestStep_combine(NULL, NULL);
	NestStep_end_step();
        }
    
    free_RepArray(elim);
    free_ShVar(_b_elim);
    free(elimbase);
}



int main(unsigned long long speid, Addr64 argp, Addr64 envp){
    int no_runs;
    int N;
    int tree;
    int d;
    double fraction;
    int i, j, k, l;
    CyclicDistArray * _b;
    CyclicDistArray * _A;
    
    double before, after, time, total;
    double * times;
    
    spu_write_decrementer(DEC_INIT_VAL);

    NestStep_SPU_init(argp);
    Name name;
    name.procedure = 1;
    name.relative = 0;

    no_runs = 5;
    N = 256;
    N = 512;
    N = 1024;
    times = (double *)malloc((no_runs + 1) * sizeof(double));
    srand(NestStep_get_rank() * spu_read_decrementer());
    
    //b blocksize needs to be a multiple of 16 byte ..
    _A = new_CyclicDistArray(name, DVAR, N * N, N);
    name.relative++;
    _b = new_CyclicDistArray(name, DVAR, N, 1);
    name.relative++;
    
    total = spu_read_decrementer();
    for(k = 0; k <= no_runs; k++){
	// initialize arrays
	double * A = malloc(_A->blocksize*sizeof(double));
	double * b = malloc(sizeof(double)*(N/NestStep_get_size()));
	printf("before create data\n");
	for(i = 0; i < N/(NestStep_get_size()); i++){
	    for(j = 0; j < N; j++){
		//		(A + i * N)[j] = (double)(rand() % 20) + 1.0;
		A[j] = (double)(rand() % 20) + 1.0;
	    }
	    b[i] = (double)(rand() % 20);
	    store_CArrD(_A,(Addr64)A,local_global_index_c(_A,i*N),local_global_index_c(_A,(i+1)*N-1));
	    store_CArrD(_b,(Addr64)&b[i],local_global_index_c(_b,i),local_global_index_c(_b,i));
	    wait_CArr_complete(_A,(Addr64)A);
	    wait_CArr_complete(_b,(Addr64)&b[i]);
	}
	printf("after create data\n");
	//	MPI_Barrier(MPI_COMM_WORLD);
	printf("before solve\n");
	before = spu_read_decrementer();
	solve(_A, _b, N);
	after = spu_read_decrementer();
	printf("after solve\n");
	times[k] = before - after;
	free(A);
	free(b);
    }
    printf("finished all runs\n");
    time = 0.0;
    for(k = 1; k <= no_runs; k++)
	time += times[k];
    time = time / (double)no_runs;
    total-=spu_read_decrementer();

    
    if(NestStep_get_rank() == 0){
	printf("group size: %d, runs: %d el: %d aet: %f total time: %f\n", NestStep_get_size(), 
	       no_runs, N, time/TIMEBASE, total/TIMEBASE);
    }
    
    free_CyclicDistArray(_A);
    free_CyclicDistArray(_b);
    free(times);
    
    NestStep_SPU_finalize();
}

