/* ********************************************** *
 *       Parallell prefix computation             *
 * ********************************************** */

#include <stdio.h>
#include <../../NestStep_spulib/neststep_spu.h>

// Used for transforming clock-ticks to
// seconds, gotten from /proc/cpuinfo
#define TIMEBASE 14318000
// Initialization value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF

// Variables for handling the counters
int ublock1, ublock2, ublocktot;
int twait1, twait2, twaittot;
int ncomb1, ncomb2, ncombtot;
int timetmp1, timetmp2, timetmptot;
int timetmp3, timetmp4, timetmptot2;
//on ps3
//#define N 1048576*4
//on cell
#define N 1048576*8

int MY_RANK;

float *_parPrefix;
PVar  *parPrefix;
float *_parSum;
ShVar *parSum;
float _pre1[4096];
float _pre2[4096];
PrivateArray* pre;
BlockDistArray* x;
float buffer1[4096];
float buffer2[4096];

void calculate_parprefix(){
    int i, j, no_chunks;
    i = j = 0;
    no_chunks = (N/4096)/NestStep_get_size();

    //initialize variables x and with data 0 -> 1024*1024 using doublebuffering
    NestStep_step();
    {
	for(j=0;j<4096;j++){
	    buffer1[j] = j + MY_RANK*4096*no_chunks; // 64*4096
	}
	store_BArrF(x,(Addr64)buffer1,
		    MY_RANK*no_chunks*4096,
		    MY_RANK*no_chunks*4096+4096-1);

	for (i=1; i<no_chunks; i++){
	    for(j=0;j<4096;j++){
		buffer2[j] = j + MY_RANK*no_chunks*4096+i*4096; // 64*4096
	    }
	    store_BArrF(x,(Addr64)buffer2,
			MY_RANK*no_chunks*4096+i*4096,
			MY_RANK*no_chunks*4096+i*4096+4096-1);

	    //twait1 = spu_read_decrementer();
	    wait_BArr_complete(x,(Addr64)buffer1);
	    //twait2 = spu_read_decrementer();
	    //	    twaittot+= twait1-twait2;
	    i++;
	    if(i >= no_chunks){
	        //twait1 = spu_read_decrementer();
		wait_BArr_complete(x,(Addr64)buffer2);
		//twait2 = spu_read_decrementer();
		//		twaittot+=twait1-twait2;
		break;
	    }

	    for(j=0;j<4096;j++){
		buffer1[j] = j + MY_RANK*no_chunks*4096 + i*4096; // 64*4096
	    }
	    store_BArrF(x,(Addr64)buffer1,
			MY_RANK*no_chunks*4096+i*4096,
			MY_RANK*no_chunks*4096+i*4096+4096-1);
	    //	    printf("max local size: %i\n",i*4096+4096-1);
	    //twait1=spu_read_decrementer();
	    wait_BArr_complete(x,(Addr64)buffer2);
	    //twait2=spu_read_decrementer();
	    // twaittot+=twait1-twait2;
	}
	//twait1=spu_read_decrementer();
	wait_BArr_complete(x,(Addr64)buffer1);
	//twait2=spu_read_decrementer();
	//twaittot+=twait1-twait2;
    }
    //ncomb1=spu_read_decrementer();
    NestStep_combine(NULL,NULL);
    //ncomb2=spu_read_decrementer();
    // ncombtot+=ncomb1-ncomb2;
    NestStep_end_step();

    //do prefix-summation
    timetmp3=spu_read_decrementer();
    NestStep_step();
    {
	get_BArrF(x,(Addr64)buffer1,MY_RANK*no_chunks*4096,
		  MY_RANK*no_chunks*4096+4096-1);
	*_parSum = 0;
	for(i=0;i<no_chunks;i++){
	    get_BArrF(x,(Addr64)buffer2,MY_RANK*no_chunks*4096+i*4096,
		      MY_RANK*no_chunks*4096+i*4096+4096-1);
	    twait1=spu_read_decrementer();
	    wait_PArr_complete(pre,(Addr64)_pre1);
	    wait_BArr_complete(x,(Addr64)buffer1);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    ublock1 = spu_read_decrementer();
	    for(j=0;j<4096;j++){
		*_parSum += buffer1[j];
		_pre1[j] = *_parSum;
	    }
	    ublock2 = spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;
	    store_PArrF(pre,(Addr64)_pre1,i*4096,i*4096+4096-1);
	    i++;
	    if(i >= no_chunks){

		twait1 = spu_read_decrementer();
		wait_BArr_complete(x,(Addr64)buffer2);
		wait_PArr_complete(pre,(Addr64)_pre2);
		twait2 = spu_read_decrementer();
		twaittot+=twait1-twait2;
		ublock1 = spu_read_decrementer();
		for(j=0;j<4096;j++){
		    *_parSum += buffer2[j];
		    _pre2[j] = *_parSum;
		}
		ublock2 = spu_read_decrementer();
		ublocktot+=ublock1-ublock2;
		store_PArrF(pre,(Addr64)_pre2,i*4096,i*4096+4096-1);
		twait1 = spu_read_decrementer();
		wait_PArr_complete(pre,(Addr64)_pre1);
		twait2 = spu_read_decrementer();
		twaittot+=twait1-twait2;
		break;
	    }
	    get_BArrF(x,(Addr64)buffer1,MY_RANK*no_chunks*4096+i*4096,
		      MY_RANK*no_chunks*4096+i*4096+4096-1);
	    twait1=spu_read_decrementer();
	    wait_BArr_complete(x,(Addr64)buffer2);
	    wait_PArr_complete(pre,(Addr64)_pre2);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    ublock1=spu_read_decrementer();
	    for(j=0;j<4096;j++){
		*_parSum += buffer2[j];
		_pre2[j] = *_parSum;
	    }
	    ublock2=spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;
	    store_PArrF(pre,(Addr64)_pre2,i*4096,i*4096+4096-1);
	}

	store_ShVarF(parSum,(Addr64)_parSum);
	twait1=spu_read_decrementer();
	wait_ShVar_complete(parSum,(Addr64)_parSum);
	wait_PArr_complete(pre,(Addr64)_pre2);
	twait2=spu_read_decrementer();
	twaittot+=twait1-twait2;

	NestStep_ShVar_attach(parSum, ADD, NULL, PREFIX,parPrefix, NOT_UPDATED, NULL);
    }
    ncomb1=spu_read_decrementer();
    NestStep_combine(NULL,NULL);
    ncomb2=spu_read_decrementer();
    ncombtot+=ncomb1-ncomb2;
    NestStep_end_step();
    
    //    parallel prefix is inside parPrefix, dma it back here,
    //    add to pre for this range.
    //    ...need tripplebuffering here.
    
    get_PVarF(parPrefix,(Addr64)_parPrefix);
    twait1=spu_read_decrementer();
    wait_PVar_complete(parPrefix,(Addr64)_parPrefix);
    twait2=spu_read_decrementer();
    twaittot+=twait1-twait2;
    NestStep_step();
    {
	get_PArrF(pre,(Addr64)buffer1,0, 4096-1);
	get_PArrF(pre,(Addr64)buffer2,4096, 2*4096-1);
	for(i=0;i<no_chunks;){
	    twait1=spu_read_decrementer();
	    wait_PArr_complete(pre,(Addr64)buffer1);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;

	    ublock1 = spu_read_decrementer();
	    for(j=0;j<4096;j++){
		buffer1[j] += *_parPrefix;
	    }
	    ublock2 = spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;

	    store_PArrF(pre,(Addr64)buffer1,i*4096,i*4096+4096-1);
	    i++;

	    if(i>=no_chunks-1){
		twait1=spu_read_decrementer();
		wait_PArr_complete(pre,(Addr64)buffer2);
		twait2=spu_read_decrementer();
		twaittot+=twait1-twait2;

		ublock1 = spu_read_decrementer();
		for(j=0;j<4096;j++){
		    buffer2[j] += *_parPrefix;
		}
		ublock2 = spu_read_decrementer();
		ublocktot+=ublock1-ublock2;

		store_PArrF(pre,(Addr64)buffer2,i*4096,i*4096+4096-1);
		break;
	    }
	    twait1=spu_read_decrementer();
	    wait_PArr_complete(pre,(Addr64)_pre1);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    get_PArrF(pre,(Addr64)_pre1,(i+1)*4096,(i+1)*4096+4096-1);
	    twait1=spu_read_decrementer();
	    wait_PArr_complete(pre,(Addr64)buffer2);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    ublock1=spu_read_decrementer();
	    for(j=0;j<4096;j++){
		buffer2[j] += *_parPrefix;
	    }
	    ublock2=spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;
	    store_PArrF(pre,(Addr64)buffer2,i*4096,i*4096+4096-1);
	    i++;
	    if(i>=no_chunks-1){
		twait1=spu_read_decrementer();
		wait_PArr_complete(pre,(Addr64)_pre1);
		twait2=spu_read_decrementer();
		twaittot+=twait1-twait2;
		ublock1 = spu_read_decrementer();
		for(j=0;j<4096;j++){
		    _pre1[j] += *_parPrefix;
		}
		ublock2 = spu_read_decrementer();
		ublocktot+=ublock1-ublock2;
		store_PArrF(pre,(Addr64)_pre1,i*4096,i*4096+4096-1);
		break;
	    }
	    twait1=spu_read_decrementer();
	    wait_PArr_complete(pre,(Addr64)buffer1);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    get_PArrF(pre,(Addr64)buffer1,(i+1)*4096,(i+1)*4096+4096-1);
	    twait1=spu_read_decrementer();
	    wait_PArr_complete(pre,(Addr64)_pre1);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    ublock1=spu_read_decrementer();
	    for(j=0;j<4096;j++){
		_pre1[j] += *_parPrefix;
	    }
	    ublock2=spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;
	    store_PArrF(pre,(Addr64)_pre1,i*4096,i*4096+4096-1);
	    
	    i++;
	    if(i>=no_chunks-1){
		twait1=spu_read_decrementer();
		wait_PArr_complete(pre,(Addr64)buffer1);
		twait2=spu_read_decrementer();
		twaittot+=twait1-twait2;
		ublock1 = spu_read_decrementer();
		for(j=0;j<4096;j++){
		    buffer1[j] += *_parPrefix;
		}
		ublock2 = spu_read_decrementer();
		ublocktot+=ublock1-ublock2;
		store_PArrF(pre,(Addr64)buffer1,i*4096,i*4096+4096-1);
		break;
	    }
	    twait1=spu_read_decrementer();
	    wait_PArr_complete(pre,(Addr64)buffer2);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    get_PArrF(pre,(Addr64)buffer2,(i+1)*4096,(i+1)*4096+4096-1);
	}
	twait1=spu_read_decrementer();
	wait_PArr_complete(pre,(Addr64)buffer1);
	wait_PArr_complete(pre,(Addr64)buffer2);
	wait_PArr_complete(pre,(Addr64)_pre1);
	twait2=spu_read_decrementer();
	twaittot+=twait1-twait2;
    }
    ncomb1=spu_read_decrementer();
    NestStep_combine(NULL,NULL);
    ncomb2=spu_read_decrementer();
    ncombtot+=ncomb1-ncomb2;
    NestStep_end_step();

    timetmp4=spu_read_decrementer();
    timetmptot2+=timetmp3-timetmp4;
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
    timetmp1 = timetmp2 =timetmptot=0;
    ublock1=ublock2=ublocktot=0;
    twait1=twait2=twaittot=0;
    ncomb1=ncomb2=ncombtot=0;
    spu_write_decrementer(DEC_INIT_VAL);
    int i;

    NestStep_SPU_init(argp);
    MY_RANK = NestStep_get_rank();
    Name name;
    name.procedure = 1;
    name.relative = 0;

    parPrefix = new_PVar(name,FVAR,(Addr64)(void*)NULL);

    name.relative++;

    _parSum = malloc(sizeof(float));
    parSum = new_ShVar(name,FVAR,(Addr64)(void*)NULL);

    name.relative++;

    pre = new_PrivateArray(name,FVAR,N/NestStep_get_size(),
					 (Addr64)(void*)_pre1);

    name.relative++;
    x = new_BlockDistArray(name,FVAR,N);

    timetmp1=spu_read_decrementer();
    for(i=0;i<200;i++){
	calculate_parprefix();
    }
    timetmp2=spu_read_decrementer();
    timetmptot=timetmp1-timetmp2;

    if(MY_RANK==0){
	printf("Tot programtime: %u, %f\n",timetmptot, 
	       ((float)timetmptot)/TIMEBASE);
	printf("Tot executiontime: %u, %f\n",timetmptot2, 
	       ((float)timetmptot2)/TIMEBASE);
	printf("Time spent calculating: %u, %f\n",ublocktot,
	       ((float)ublocktot)/TIMEBASE);
	printf("Time spent waiting: %u, %f\n",twaittot,
	       ((float)twaittot)/TIMEBASE);
	printf("Time spent in combine: %u, %f\n",ncombtot,
	       ((float)ncombtot)/TIMEBASE);
    }

    NestStep_SPU_finalize();

    return 0;
}
