/* ********************************************** *
 *     Pi test-program, uses NestStep-CELL        *
 * ********************************************** */

#include <stdio.h>
#include <../../NestStep_spulib/neststep_spu.h>

#define N 10000000

ShVar * pi;
ShVar *h;

int MY_SIZE;
int MY_RANK;

// Used for transforming clock-ticks to
// seconds, gotten from /proc/cpuinfo
#define TIMEBASE 14318000
// Initializing-value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF

// Timing variables
int ublock1, ublock2, ublocktot;
int twait1, twait2, twaittot;
int ncomb1, ncomb2, ncombtot;
int tot_time, tmptime1, tmptime2;
int tot_time2,tmptime3, tmptime4;

static inline double f( double x )
{
    return  4.0/(1.0 + x*x );
}

void calculate_pi(){
    int i;
    double *_pi;
    double *_h;
    
    NestStep_step();
    {
	_pi = malloc(sizeof(double));
	_h = malloc(sizeof(double));
	*_h = 1.0 / (double)N;
	*_pi = 0;
	ublock1=spu_read_decrementer();
	for (i=MY_RANK; i<N; i+=MY_SIZE){
	    *_pi += f( (*_h)*((double)i - 0.5));
	}
	ublock2 = spu_read_decrementer();
	ublocktot+=ublock1-ublock2;
	store_ShVarD(pi,(Addr64)_pi);
	twait1=spu_read_decrementer();
	wait_ShVar_complete(pi,(Addr64)_pi);
	twait2=spu_read_decrementer();
	twaittot+=twait1-twait2;
	NestStep_ShVar_attach(pi, ADD, NULL, NO_PREFIX,NULL, 
			      NOT_UPDATED, NULL); 
    }
    ncomb1=spu_read_decrementer();
    NestStep_combine(NULL,NULL);
    ncomb2=spu_read_decrementer();
    ncombtot+=ncomb1-ncomb2;
    NestStep_end_step();
    
    NestStep_step();
    {
	get_ShVarD(pi,(Addr64)_pi);
	twait1=spu_read_decrementer();
	wait_ShVar_complete(pi,(Addr64)_pi);
	twait2=spu_read_decrementer();
	twaittot+=twait1-twait2;
	*_pi = (*_pi)*(*_h);
//	if (MY_RANK==0)
//	    printf("PI=%f\n", *_pi );
    }
    ncomb1=spu_read_decrementer();
    NestStep_combine(NULL,NULL);
    ncomb2=spu_read_decrementer();
    ncombtot+=ncomb1-ncomb2;
    NestStep_end_step();

    free(_pi);
    free(_h);

}

int main(unsigned long long speid, Addr64 argp, Addr64 envp){
    int i;
    ublock1=ublock2=ublocktot=0;
    twait1=twait2=twaittot=0;
    ncomb1=ncomb2=ncombtot=0;
    tot_time=tmptime1=tmptime2=0;
    tot_time2=tmptime3=tmptime4=0;


    NestStep_SPU_init(argp);
    Name name;
    name.procedure = 1;
    name.relative = 0;

    MY_SIZE = NestStep_get_size();
    MY_RANK = NestStep_get_rank();

    spu_write_decrementer(DEC_INIT_VAL);

    //    sh float pi;
    pi = new_ShVar(name,DVAR,(Addr64)(void*)NULL);

    //    sh float h<=>;   // shared constant

    name.relative++;
    h = new_ShVar(name,DVAR,(Addr64)(void*)NULL);

    tmptime1 = spu_read_decrementer();
    for(i=0;i<4;i++){
	calculate_pi();
    }
    tmptime2 = spu_read_decrementer();
    tot_time+=tmptime1-tmptime2;
    if(MY_RANK == 0){
	printf("Tot executiontime: %u, %f\n",tot_time, 
	       ((float)tot_time)/TIMEBASE);
	printf("Time spent calculating: %u, %f\n",ublocktot,
	       ((float)ublocktot)/TIMEBASE);
	printf("Time spent waiting: %u, %f\n",twaittot,
	       ((float)twaittot)/TIMEBASE);
	printf("Time spent in combine: %u, %f\n",ncombtot,
	       ((float)ncombtot)/TIMEBASE);
    }

    NestStep_SPU_finalize();
}
