#include "reduce.h"
#include "spu_time.h"
#define CHUNKSIZE 4096 


int calc_chunk(int left, int chunksize)
{
	return 	(chunksize < left ? chunksize : left);
}

int calc_chunk_dma(int left, int chunksize)
{
	int chunk =	(chunksize < left ? chunksize : left);
	if(chunk%4)
		chunk += 4-(chunk%4); // dma transfers must be factor of 16B
	return chunk;
}

spu_timer total_time={0,0};
spu_timer exec_time={0,0};
spu_timer dma_time={0,0};
spu_timer comb_time={0,0};
spu_timer calc_time={0,0};



int getBDALocalSize(BlockDistArray* x)
{
	return x->localsize;
}

int getBDALowerBound(BlockDistArray* x)
{
	return x->lower;
}

float distReduce(float (*red_func)(float, float), BlockDistArray* x, int N)
{
	int i;
	float buffer1[4096];
	float buffer2[4096];
	int rank = NestStep_get_rank();
	int groupSize = NestStep_get_size();
	float __sum;
	float *_sum = &__sum;
	Name name;
	name.procedure=NAME_OFFSET+NAME_REDUCE;
	name.relative=0;
    ShVar *sum = new_ShVar(name,FVAR,(Addr64)(void*)NULL);
    name.relative++;
	float _resVec[MAX_SPUS];
	for(i=0;i<groupSize;i++)
		_resVec[i]=0;
	//RepArray* resVec = new_RepArray(name, FVAR, 4*groupSize, (Addr64)_resVec);
	RepArray* resVec = new_RepArray(name, FVAR, MAX_SPUS, (Addr64)_resVec);
//	BlockDistArray *resVec = new_BlockDistArray(name,FVAR,4*groupSize);
	

	int sumFirst=1;
	int part = getBDALocalSize(x);
	int start = getBDALowerBound(x);
	//~ int part = N/groupSize;
	//~ if(part%4) // keep bounderys 16b aligned
		//~ part -= part%4; 
	//~ int start = rank*part;	
	//~ if(rank == groupSize -1) // last in group
		//~ part += N-part*groupSize; // last member get the rest
	
	int noChunks = part/CHUNKSIZE;
	int rest = part - noChunks*CHUNKSIZE;
	
	
	NestStep_step();
    {	
		if(noChunks) // do all full chunks
		{
			get_BArrF(x,(Addr64)buffer1,start,
				start+calc_chunk(part, CHUNKSIZE)-1);
			
			for(i=0;i<noChunks;i++)
			{
				if(i < noChunks-1)
					get_BArrF(x,(Addr64)buffer2,start+(i+1)*CHUNKSIZE,
						start+(i+1)*CHUNKSIZE+CHUNKSIZE-1);
				startTimer(&dma_time);
				wait_BArr_complete(x,(Addr64)buffer1);
				stopTimer(&dma_time);
				startTimer(&calc_time);
				
				if(sumFirst)
				{
					*_sum = localReduce(red_func,buffer1,CHUNKSIZE);
					sumFirst = 0;
				}
				else
					*_sum = red_func(*_sum, localReduce(red_func,buffer1,CHUNKSIZE));
				//for(j=0;j<4096;j++)
				//{
					//tmpSum += buffer1[j];
				//	tmpSum = red_func(tmpSum, buffer1[j]);
				//}
				stopTimer(&calc_time);
				//printf("loop: %d, sum: %f\n", i, tmpSum);
				
				i++;
				if(i >= noChunks)
					break;
				if(i < noChunks-1)
					get_BArrF(x,(Addr64)buffer1,start+(i+1)*CHUNKSIZE,
					  start+(i+1)*CHUNKSIZE+CHUNKSIZE-1);
				startTimer(&dma_time);
				wait_BArr_complete(x,(Addr64)buffer2);
				stopTimer(&dma_time);
				startTimer(&calc_time);
				*_sum = red_func(*_sum, localReduce(red_func,buffer2,CHUNKSIZE));
				//for(j=0;j<4096;j++)
				//{
					//tmpSum += buffer2[j];
				//	tmpSum = red_func(tmpSum, buffer2[j]);
				//}
				stopTimer(&calc_time);
				//printf("loop: %d, sum: %f\n", i, tmpSum);
				
			}
		}
		if(rest)
		{
			get_BArrF(x,(Addr64)buffer1,start+noChunks*CHUNKSIZE,
					start+noChunks*CHUNKSIZE+rest-1);
			startTimer(&dma_time);
			wait_BArr_complete(x,(Addr64)buffer1);
			stopTimer(&dma_time);
			if(sumFirst)
				*_sum = localReduce(red_func,buffer1,rest);
			else
				*_sum = red_func(*_sum, localReduce(red_func,buffer1,rest));
		}
		
	//	printf("foo\n");
//		printf("part result: %f\n", *_sum);

		_resVec[rank] = *_sum;
		store_RArrF(resVec, (Addr64)_resVec,0,MAX_SPUS-1);
		wait_RArr_complete(resVec,(Addr64)_resVec);
		NestStep_RepArray_attach(resVec, ADD, NULL, NO_PREFIX, NULL,NOT_UPDATED, NULL);
//		NestStep_RepArray_range(resVec, rank*4, rank*4+1);
		NestStep_combine(NULL,NULL);
		//		store_ShVarF(sum,(Addr64)_sum);
//		wait_ShVar_complete(sum,(Addr64)_sum);

//		NestStep_ShVar_attach(sum, ADD, NULL, NO_PREFIX,NULL, NOT_UPDATED, NULL);

	}	
	NestStep_end_step();

//	printf("foo\n");
	
	NestStep_step();
	{
		if(rank == 0)
		{
			//float vec[MAX_SPUS*4];
			//float vec2[MAX_SPUS];
			get_RArrF(resVec, (Addr64)_resVec, 0, MAX_SPUS-1);
			wait_RArr_complete(resVec,(Addr64)_resVec);
			//for(i=0;i<groupSize;i++)
			//	vec2[i] = _resVec[i*4];
			float res = localReduce(red_func, _resVec, groupSize);
			store_ShVarF(sum,(Addr64)&res);
			wait_ShVar_complete(sum,(Addr64)&res);
	//		printf("combine result: %f\n", res);
		}
		NestStep_ShVar_attach(sum, LEADER, NULL, NO_PREFIX,NULL, NOT_UPDATED, NULL);
		NestStep_combine(NULL,NULL);
	}
	NestStep_end_step();	
	NestStep_step();
	{
		get_ShVarF(sum,(Addr64)_sum);
		wait_ShVar_complete(sum,(Addr64)_sum);
//		printf("res from %d: %f\n", rank, *_sum);
		NestStep_combine(NULL,NULL);
	}
	NestStep_end_step();
	free_ShVar(sum);
	free_RepArray(resVec);

	return *_sum;
}



float localReduce(float (*red_func)(float, float), float* x, int N)
{
	int i=0;
	float tmp=x[0];
	for(i=1;i<N;i++)
		tmp = red_func(tmp, x[i]);
	return tmp;
}
